/*
 * AbstractStream.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 *
 * This file was modified by Sun Microsystems on June 24th, 2005.
 * Modifications to this file include:
 * - exposure of openOutbound and closeOutbound for non blocking
 * stream open or close.
 * - firePacketTransferred(RECEIVED) is called as soon as a packet is
 * built from the read data, instead of waiting for all read data to
 * be processed.
 * 
 */

package net.outer_planes.jso;


import java.io.IOException;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.log4j.Logger;

import org.jabberstudio.jso.JSOImplementation;
import org.jabberstudio.jso.NSI;
import org.jabberstudio.jso.Packet;
import org.jabberstudio.jso.PacketRouter;
import org.jabberstudio.jso.Stream;
import org.jabberstudio.jso.StreamAttribute;
import org.jabberstudio.jso.StreamContext;
import org.jabberstudio.jso.StreamDataFactory;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.StreamError;
import org.jabberstudio.jso.StreamException;
import org.jabberstudio.jso.StreamFeatureset;
import org.jabberstudio.jso.StreamNamespace;
import org.jabberstudio.jso.StreamNode;
import org.jabberstudio.jso.StreamObject;
import org.jabberstudio.jso.event.PacketEvent;
import org.jabberstudio.jso.event.PacketListener;
import org.jabberstudio.jso.event.StreamFeaturesEvent;
import org.jabberstudio.jso.event.StreamFeaturesListener;
import org.jabberstudio.jso.event.StreamStatusEvent;
import org.jabberstudio.jso.event.StreamStatusListener;
import org.jabberstudio.jso.event.StreamStatusSetListener;
import org.jabberstudio.jso.io.StreamBuilder;
import org.jabberstudio.jso.io.StreamParseState;
import org.jabberstudio.jso.io.StreamSource;
import org.jabberstudio.jso.io.XMLExporter;
import org.jabberstudio.jso.util.ContextDecorator;
import org.jabberstudio.jso.util.Utilities;

import net.outer_planes.jso.ixp.IxpException;
import net.outer_planes.jso.ixp.IxpParser;

/**
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public abstract class AbstractStream extends DataFactory implements Stream {
    //Inner-classes
    private class Input {
        //Inner-classes
        public class BuilderPlace {
            //Fields
            public final NSI            name;
            public final StreamBuilder  builder;
            
            //Constructors
            public BuilderPlace(NSI name, StreamBuilder b) {
                this.name = name;
                this.builder = b;
            }
        }
        //Fields
        private IxpParser   _Parser;
        private List        _BuildStack = new java.util.LinkedList();
        
        private boolean     _Opened = false;

        //Constructors
        public Input(StreamSource in) {
            StreamContext       ctx = AbstractStream.this.getInboundContext();
            
            _Parser = new IxpParser(AbstractStream.this.getInboundContext(), in);
            pushBuilder(ctx.getNSI(), new ContextBuilder(ctx) {
                //Methods
                public void addExtendedData(StreamNode data) {
                    Packet      packet = (Packet)data;
                    String      uri = packet.getNamespaceURI(), ln = packet.getLocalName();
                    Locale      lang, declared;
                    
                    lang = getContext().getLocale();
                    declared = packet.getDeclaredLocale();
                    if (declared == null)
                        packet.setDeclaredLocale(lang);
                    if (data instanceof StreamError)
                        setStreamException(new StreamException((StreamError)data));
                    
                    super.addExtendedData(data);
                    AbstractStream.this.getInboundQueue().add(packet);
		    firePacketTransferredEvent(PacketEvent.RECEIVED);
                }

                public StreamBuilder.Status processElementStart(StreamParseState state) {
                    StreamBuilder.Status    stat = super.processElementStart(state);
                    
                    if ((stat == StreamBuilder.PROCESSING) && (true)) {
                        AbstractStream.this.fireStatusChangedEvent(getContext(), Stream.OPENED, null);
                    }
                    
                    return stat;
                }
                public StreamBuilder.Status processElementEnd(StreamParseState state) {
                    StreamBuilder.Status    stat = super.processElementEnd(state);
                    
                    if ((stat == StreamBuilder.COMPLETE) && (true)) {
                        AbstractStream.this.fireStatusChangedEvent(getContext(), Stream.CLOSED, getStreamException());
                        try {
                            AbstractStream.this.close();
                        } catch (StreamException se) {
                            obtainLogger().info("Somewhat unexpected stream exception thrown", se);
                        }
                        try {
                            AbstractStream.this.disconnect();
                        } catch (StreamException se) {
                            obtainLogger().info("Somewhat unexpected stream exception thrown", se);
                        }
                    }
                    
                    return stat;
                }
            });
        }
        
        //Methods
        private final boolean isOpened() {
            return _Opened;
        }
        private final void setOpened(boolean open) {
            _Opened = open;
        }
        
        public synchronized void process() throws IOException, IxpException {
            Logger              log = obtainLogger();
            BuilderPlace        place = topBuilder();
            StreamBuilder       b = (place != null) ? place.builder : null;
            StreamParseState    ps;
            NSI                 name = (place != null) ? place.name : null;
            boolean             opened = isOpened();

            while ((b != null) && (name != null) && ((ps = _Parser.process()) != null)) {
                if          (ps.getType() == StreamParseState.ELEMENT_START) {
                    if (log.isDebugEnabled())
                        log.debug("processing start of element (" + ps.getName() + ")");
                    while (b.processElementStart(ps) == StreamBuilder.EXTENDED_DATA) {
                        b = b.buildExtendedData();
                        name = ps.getName();
                        if (log.isDebugEnabled()) 
			    log.debug("pushing extended data " + name);
                        pushBuilder(name, b);
                    }

                    setOpened(true);
                    if (!opened) {
			log.info("stream was not opened.  ignoring data");
                        break;
		    }
                } else if   (ps.getType() == StreamParseState.ELEMENT_END) {
                    boolean complete = false;
                    
                    if (log.isDebugEnabled())
                        log.debug("processing end of element (" + ps.getName() + ")");
                    do {
                        complete = (b.processElementEnd(ps) == StreamBuilder.COMPLETE);
                        if (complete) {
                            StreamNode      node = b.getNode();

                            log.debug("popping extended data");
                            popBuilder();
                            place = topBuilder();
                            if (place != null) {
                                b = place.builder;
                                name = place.name;
                                b.addExtendedData(node);
                            } else {
                                name = null;
                            }
                        }
                    } while (complete && (ps.getName().equals(name)));
                } else if   (ps.getType() == StreamParseState.TEXT) {
                    if (log.isDebugEnabled())
                        log.debug("processing text (" + ps.getText() + ")");
                    b.processText(ps);
                }
            }
            
            if (_Parser.isEOD()) {
                log.debug("end of document...");
                AbstractStream.this.fireStatusChangedEvent(AbstractStream.this.getInboundContext(), Stream.DISCONNECTED, null);
            }
        }

        private BuilderPlace topBuilder() {
            return (BuilderPlace)(_BuildStack.isEmpty() ? null : _BuildStack.get(0));
        }
        private void pushBuilder(NSI name, StreamBuilder b) {
            _BuildStack.add(0, new BuilderPlace(name, b));
        }
        private BuilderPlace popBuilder() {
            return (BuilderPlace)(_BuildStack.isEmpty() ? null : _BuildStack.remove(0));
        }
    }
    private class Output {
        //Fields
        private StreamSource    _Writer;
        private XMLExporter     _Exporter;
        
        //Constructors
        public Output(StreamSource out) {
            StreamContext   ctx = AbstractStream.this.getOutboundContext();
            
            _Writer = out;
            _Exporter = getJSO().createXMLExporter(ctx);
        }
        
        //Methods
        public void sendOpen() throws IOException {
            StreamContext   ctx = AbstractStream.this.getOutboundContext();
            StringBuffer    buffer = new StringBuffer();
            Map             attrs = ctx.getAttributes();
            boolean         procDefaultNS = false, procThisNS = false;
            Iterator        itr;
            byte[]          output;
            String          charset = ctx.obtainCharset().name();
            
            //Start with tag name (qualified)
            buffer.append("<").append(ctx.getQualifiedName());
            
            //Setup namespaces
            itr = ctx.getDeclaredNamespaces().iterator();
            while (itr.hasNext()) {
                StreamNamespace ns = (StreamNamespace)itr.next();
                String          uri = ns.getURI();
                String          prefix = ns.getPrefix();

                uri = encodeAttributeValue(uri);
                
                buffer.append(" ");
                buffer.append("xmlns");
                if (Utilities.isValidString(prefix))
                    buffer.append(":").append(prefix);
                else
                    procDefaultNS = true;
                buffer.append("=").append("'").append(uri).append("'");

                if (Utilities.equateStrings(uri, ctx.getNamespaceURI()))
                    procThisNS = true;
            }
            if (!procDefaultNS) {
                String  uri = encodeAttributeValue(AbstractStream.this.getDefaultNamespace());

                buffer.append(" ");
                buffer.append("xmlns");
                buffer.append("=").append("'").append(uri).append("'");
            }
            if (!procThisNS) {
                String  uri = encodeAttributeValue(ctx.getNamespaceURI());
                String  prefix = ctx.getNamespacePrefix();
                
                if (prefix == null)
                    prefix = "stream";
                
                buffer.append(" ");
                buffer.append("xmlns:").append(prefix);
                buffer.append("=").append("'").append(uri).append("'");
            }
            
            //Setup attributes
            itr = attrs.keySet().iterator();
            while (itr.hasNext()) {
                NSI             name = (NSI)itr.next();
                StreamAttribute attr = (StreamAttribute)attrs.get(name);
                String          val = encodeAttributeValue(attr.getValue());
                
                buffer.append(" ");
                buffer.append(attr.getQualifiedName()).append("=").append("'").append(val).append("'");
            }
            buffer.append(">");
            
            output = buffer.toString().getBytes(charset);
            _Writer.write(output, 0, output.length);
        }
        public void sendPacket(Packet p) throws IOException {
            StreamContext   ctx = AbstractStream.this.getOutboundContext();
            String          buffer = " ";
            String          charset = ctx.obtainCharset().name();
            byte[]          output;
            
            if (p != null) {
                _Exporter.getContext().add(p);
                buffer = _Exporter.write(p);
                _Exporter.getContext().remove(p);
            }
            
            output = buffer.getBytes(charset);
            synchronized(this) {
                _Writer.write(output, 0, output.length);
            }
            
            if (p != null) {
                AbstractStream.this.getOutboundQueue().add(p);
            }
        }
        public void sendClose(StreamException se) throws IOException {
            StreamContext       ctx = AbstractStream.this.getOutboundContext();
            StringBuffer        buffer = new StringBuffer();
            byte[]              output;
            
            if (se != null) {
                buffer.append(_Exporter.write(se.asNode(ctx)).toString());
            }
            buffer.append("</").append(ctx.getQualifiedName()).append(">");
            
            output = buffer.toString().getBytes("UTF-8");
            _Writer.write(output, 0, output.length);
        }
        
        private String encodeAttributeValue(String attr) {
            StringBuffer    buffer = new StringBuffer();
            int             idx;
            
            for (idx = 0; idx < attr.length(); idx++) {
                char    ch = attr.charAt(idx);
                
                switch(ch) {
                    case '&':   buffer.append("&amp;");     break;
                    case '<':   buffer.append("&lt;");      break;
                    case '>':   buffer.append("&gt;");      break;
                    case '\'':  buffer.append("&apos;");    break;
                    case '\"':  buffer.append("&quot;");    break;
                    default:    buffer.append(ch);          break;
                }
            }
            
            return buffer.toString();
        }
    }
    
    private class Context extends ContextDecorator {
        //Fields
        private StreamContext   _Base;
        private boolean         _Inbound;
        
        //Constructors
        public Context(StreamContext base, boolean inbound) {
            setContext(base);
            _Inbound = inbound;
        }
        
        //Methods
        protected StreamContext getDecoratedContext() {
            return _Base;
        }
        protected void setContext(StreamContext ctx) throws IllegalArgumentException {
            if (ctx == null)
                throw new IllegalArgumentException("StreamContext cannot be null");
            _Base = ctx;
        }
        
        public boolean isInbound() {
            return _Inbound;
        }
        public boolean isOutbound() {
            return !_Inbound;
        }
        
        public Stream getStream() {
            return AbstractStream.this;
        }
        public StreamDataFactory getFactory() {
            return AbstractStream.this.getDataFactory();
        }
        
        public StreamDataFactory getDataFactory() {
            return AbstractStream.this.getDataFactory();
        }
        public PacketRouter getRouter() {
            return AbstractStream.this;
        }
        public Stream.Status getCurrentStatus() {
            return (_Inbound) ? AbstractStream.this.getInboundStatus() : AbstractStream.this.getOutboundStatus();
        }
        
        public StreamObject copy(StreamElement p) {
            return new Context(getDecoratedContext(), isInbound());
        }
    }

    //Fields
    private JSO                 _JSO;
    
    private String              _NS;
    
    private StreamSource        _Src;

    private Input               _Inbound;
    private StreamContext       _InCtx;
    private Stream.Status       _InStat;
    private List                _InPkts = Collections.EMPTY_LIST;
    private List                _InFeats = Collections.EMPTY_LIST;
    private List                _InQueue;

    private Output              _Outbound;
    private StreamContext       _OutCtx;
    private Stream.Status       _OutStat;
    private List                _OutPkts = Collections.EMPTY_LIST;
    private List                _OutFeats = Collections.EMPTY_LIST;
    private List                _OutQueue;
    
    private List                _StatListens = Collections.EMPTY_LIST;
    
    private StreamException     _Except;
    
    private StreamDataFactory   _sdf;
    
    //Constructors
    public AbstractStream(JSO jso, DataFactory df, String ns, StreamContext inCtx, StreamContext outCtx) {
        super(df);
        _sdf = df;
        setupInbound(inCtx);
        setupOutbound(outCtx);
        
        setDefaultNamespace(ns);
        setJSO(jso);
    }
    
    //Methods
    protected Logger obtainLogger() {
        return Logger.getLogger(getClass());
    }
    
    public StreamContext getInboundContext() {
        return _InCtx;
    }
    public synchronized Stream.Status getInboundStatus() {
        return _InStat;
    }
    private synchronized void setInboundStatus(Stream.Status next) {
        _InStat = next;
    }
    private void setupInbound(StreamContext ctx) throws IllegalArgumentException {
        _InCtx = new Context(ctx, true);
        _InStat = Stream.DISCONNECTED;
    }
    
    public StreamContext getOutboundContext() {
        return _OutCtx;
    }
    public synchronized Stream.Status getOutboundStatus() {
        return _OutStat;
    }
    private synchronized void setOutboundStatus(Stream.Status next) {
        _OutStat = next;
    }
    private void setupOutbound(StreamContext ctx) throws IllegalArgumentException {
        _OutCtx = new Context(ctx, false);
        _OutStat = Stream.DISCONNECTED;
    }
    
    public Stream.Status getCurrentStatus() {
        return Stream.Status.getAggregateStatus(getInboundStatus(), getOutboundStatus());
    }
    public StreamSource getSource() {
        return _Src;
    }

    public JSOImplementation getJSO() {
        return _JSO;
    }
    private void setJSO(JSO jso) throws IllegalArgumentException {
        if (jso == null)
            throw new IllegalArgumentException("JSOImplementation cannot be null");
        _JSO = jso;
    }
    
    public StreamDataFactory getDataFactory() {
        //return this;
        return _sdf;
    }
    
    public String getDefaultNamespace() {
        return _NS;
    }
    private void setDefaultNamespace(String ns) throws IllegalArgumentException {
        if (!Utilities.isValidString(ns))
            throw new IllegalArgumentException("Default namespace cannot be null or \"\"");
        _NS = ns;
        getInboundContext().addNamespace("", ns);
        getOutboundContext().addNamespace("", ns);
    }
    
    protected StreamException getStreamException() {
        return _Except;
    }
    protected synchronized void setStreamException(StreamException except) {
        _Except = except;
    }
    
    public synchronized void addStreamStatusListener(final StreamStatusListener l) {
        if (l != null) {
            _StatListens = new java.util.ArrayList(_StatListens);
            
            _StatListens.add(l);
        }
    }
    public synchronized void removeStreamStatusListener(StreamStatusListener l) {
        if (l != null) {
            _StatListens = new java.util.ArrayList(_StatListens);
            
            _StatListens.remove(l);
        }
    }
    protected void fireStatusChangedEvent(StreamContext ctx, Stream.Status next, Exception e) {
        StreamStatusEvent       evt;
        Stream.Status           prev;
        List                    listeners;
        Iterator                itr;
        
        //Determine "direction"
        if (ctx == getInboundContext()) {
            prev = getInboundStatus();
            setInboundStatus(next);
        } else if   (ctx == getOutboundContext()) {
            prev = getOutboundStatus();
            setOutboundStatus(next);
        } else {
            obtainLogger().warn("Attempt to fire events for unknown context");
            return;
        }
        
        if (prev != next) {
            //Fire event!
            synchronized (this) {
                //Grab caught exception
                if (e == null)
                    e = getStreamException();
                
                itr = _StatListens.iterator();
                evt = new StreamStatusEvent(this, ctx, prev, next, e);
            }

            while (itr.hasNext()) {
                try {
                    ((StreamStatusListener)itr.next()).statusChanged(evt);
                } catch (Exception e2) {
                    obtainLogger().warn("StreamStatusListener exception thrown", e2);
                }
            }
        } else {
            //Fire event!
            synchronized (this) {
                //Grab caught exception
                if (e == null)
                    e = getStreamException();

                itr = _StatListens.iterator();
                evt = new StreamStatusEvent(this, ctx, prev, next, e);
            }

            while (itr.hasNext()) {
                try {
                    Object obj = itr.next();
                    if (obj instanceof StreamStatusSetListener){
                        ((StreamStatusSetListener)obj).statusSet(evt);
                    }
                } catch (Exception e2) {
                    obtainLogger().warn("StreamStatusListener exception thrown", e2
);
                }
	    }
	}
    }
    
    public void addPacketListener(PacketListener l) {
        addPacketListener(PacketEvent.RECEIVED, l);
        addPacketListener(PacketEvent.SENT, l);
    }
    public void removePacketListener(PacketListener l) {
        removePacketListener(PacketEvent.RECEIVED, l);
        removePacketListener(PacketEvent.SENT, l);
    }
    
    public synchronized void addPacketListener(PacketEvent.Type type, PacketListener l) {
        if (l != null) {
            List        listeners = null;
            
            if      (type == PacketEvent.RECEIVED)
                listeners = _InPkts = new java.util.ArrayList(_InPkts);
            else if (type == PacketEvent.SENT)
                listeners = _OutPkts = new java.util.ArrayList(_OutPkts);

            if (listeners != null)
                listeners.add(l);
        }
    }
    public synchronized void removePacketListener(PacketEvent.Type type, PacketListener l) {
        if (l != null) {
            List        listeners = null;
            
            if      (type == PacketEvent.RECEIVED)
                listeners = _InPkts = new java.util.ArrayList(_InPkts);
            else if (type == PacketEvent.SENT)
                listeners = _OutPkts = new java.util.ArrayList(_OutPkts);
            
            if (listeners != null)
                listeners.remove(l);
        }
    }
    protected void firePacketTransferredEvent(PacketEvent.Type type) {
        List                queue, listeners;
        StreamContext       ctx;
        Iterator            pItr;
        
        //Determine "direction"
        synchronized(this) {
            if (type == PacketEvent.RECEIVED) {
                ctx = getInboundContext();
                queue = getInboundQueue();
                listeners = _InPkts;
                pItr = new java.util.ArrayList(queue).iterator();
            } else if (type == PacketEvent.SENT) {
                ctx = getOutboundContext();
                queue = getOutboundQueue();
                listeners = _OutPkts;
                pItr = new java.util.ArrayList(queue).iterator();
            } else {
                obtainLogger().warn("Attempt to process events for non-existent type");
                return;
            }
            
            queue.clear();
        }
        
        //Process the packet queue
        while (pItr.hasNext()) {
            Packet      p = (Packet)pItr.next();
            String      uri = p.getNamespaceURI(), ln = p.getLocalName();
            PacketEvent evt = new PacketEvent(ctx, p, type);
            Iterator    lItr = listeners.iterator();

            try {
                if (p instanceof StreamFeatureset)
                    fireFeaturesReported(type, (StreamFeatureset)p);
            } catch (Throwable t) {
                    obtainLogger().warn("StreamFeatures exception thrown", t);
            }
            
            while (lItr.hasNext()) {
                try {
                    ((PacketListener)lItr.next()).packetTransferred(evt);
                } catch (Throwable t) {
                    obtainLogger().warn("PacketListener exception thrown", t);
                }
            }
            
            if (p instanceof StreamError)
                setStreamException(new StreamException((StreamError)p));
            p.detach();
        }
    }

    public void addStreamFeaturesListener(StreamFeaturesListener l) {
        addStreamFeaturesListener(PacketEvent.RECEIVED, l);
        addStreamFeaturesListener(PacketEvent.SENT, l);
    }
    public void removeStreamFeaturesListener(StreamFeaturesListener l) {
        removeStreamFeaturesListener(PacketEvent.RECEIVED, l);
        removeStreamFeaturesListener(PacketEvent.SENT, l);
    }
    
    public synchronized void addStreamFeaturesListener(PacketEvent.Type type, StreamFeaturesListener l) {
        if (l != null) {
            List    listeners = null;
            
            if      (type == PacketEvent.RECEIVED)
                listeners = _InFeats = new java.util.ArrayList(_InFeats);
            else if (type == PacketEvent.SENT)
                listeners = _OutFeats = new java.util.ArrayList(_OutFeats);
            
            if (listeners != null)
                listeners.add(l);
        }
    }
    public synchronized void removeStreamFeaturesListener(PacketEvent.Type type, StreamFeaturesListener l) {
        if (l != null) {
            List    listeners = null;
            
            if      (type == PacketEvent.RECEIVED)
                listeners = _InFeats = new java.util.ArrayList(_InFeats);
            else if (type == PacketEvent.SENT)
                listeners = _OutFeats = new java.util.ArrayList(_OutFeats);
            
            if (listeners != null)
                listeners.remove(l);
        }
    }
    protected void fireFeaturesReported(PacketEvent.Type type, StreamFeatureset feats) {
        Iterator            itr = null;
        StreamFeaturesEvent sfe;
        StreamContext       ctx;
        
        //Determine "direction"
        synchronized(this) {
            if          (type == PacketEvent.RECEIVED) {
                ctx = getInboundContext();
                itr = new java.util.ArrayList(_InFeats).iterator();
            } else if   (type == PacketEvent.SENT) {
                ctx = getOutboundContext();
                itr = new java.util.ArrayList(_OutFeats).iterator();
            } else {
                obtainLogger().warn("Attempt to process stream features events for non-existent type");
                return;
            }
        }

        //Process the event already!
        sfe = new StreamFeaturesEvent(ctx, feats, type);
        while (itr.hasNext()) {
            ((StreamFeaturesListener)itr.next()).featuresReported(sfe);
        }
    }
    
    protected synchronized List getInboundQueue() {
        if (_InQueue == null)
            _InQueue = Collections.synchronizedList(new java.util.LinkedList());
        return _InQueue;
    }
    protected synchronized List getOutboundQueue() {
        if (_OutQueue == null)
            _OutQueue = Collections.synchronizedList(new java.util.LinkedList());
        return _OutQueue;
    }
    
    /** <p>
     * Connects this <tt>Stream</tt> to the <tt>StreamSource</tt>. This
     * method does not return until {@link #getCurrentStatus} returns {@link
     * #CONNECTED} or an exception occurs.  If this <tt>Stream</tt> is
     * already in this state (or "higher"), this method does nothing.</p>
     *
     * <p>
     * Any exception (other than <tt>IllegalArgumentException</tt> and
     * <tt>StreamException</tt>) thrown is caught and re-thrown as a
     * <tt>StreamException</tt>.</p>
     *
     * <p>
     * The value of <tt>src</tt> cannot be <tt>null</tt> or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param src The <tt>StreamSource</tt>.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @throws StreamException If any other errors occur.
     */
    public synchronized void connect(StreamSource src) throws IllegalArgumentException, StreamException {
        Logger      log = obtainLogger();
        
        //Validate State (harmless)
        if (getCurrentStatus().isConnected())
            return;
        
        //Validate StreamSource
        if (src == null)
            throw new IllegalArgumentException("StreamSource cannot be null");
        
        try {
            //Reset exceptions
            setStreamException(null);
            
            //Connect to StreamSource
            src.connect(this);
            _Src = src;
            
            //Setup I/O
            _Outbound = new Output(getSource());
            _Inbound = new Input(getSource());

            //Fire events
            fireStatusChangedEvent(getOutboundContext(), Stream.CONNECTED, null);
            fireStatusChangedEvent(getInboundContext(), Stream.CONNECTED, null);
        } catch (Exception e) {
            StreamException se = new StreamException(getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION), e);
            
            obtainLogger().debug("throwing StreamException for connect", se);
            throw se;
        }
    }
    /** <p>
     * Disconnects this <tt>Stream</tt>.  This method does not return until
     * {@link #getCurrentStatus} returns {@link #DISCONNECTED} or an exception
     * occurs.  If this <tt>Stream</tt> is already is in this state, this
     * method does nothing.</p>
     *
     * <p>
     * Any exception thrown is caught and re-thrown as a
     * <tt>StreamException</tt>.</p>
     *
     * @throws StreamException If any errors occur.
     */
    public synchronized void disconnect() throws StreamException {
        Logger          log = obtainLogger();
        StreamException except = getStreamException();
        
        //Validate State (harmless)
        if (getCurrentStatus().isDisconnected())
            return;
        
        log.debug("disconnecting stream...");
        try {
            //Disconnect from StreamSource
            _Src.disconnect(this);

            //Reset I/O
            _Inbound = null;
            _Outbound = null;
            _Src = null;
        } catch (IOException ioe) {
            except = new StreamException(getDataFactory().createStreamError(StreamError.REMOTE_CONNECTION_FAILED_CONDITION), ioe);

            obtainLogger().debug("throwing (I/O) StreamException for disconnect", except);
            throw except;
        } catch (Exception e) {
            except = new StreamException(getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION), e);
            
            obtainLogger().debug("throwing (general) StreamException for disconnect", except);
            throw except;
        } finally {
            if (getStreamException() != null)
                except = getStreamException();
            
            fireStatusChangedEvent(getOutboundContext(), Stream.DISCONNECTED, except);
            fireStatusChangedEvent(getInboundContext(), Stream.DISCONNECTED, except);
        }
    }
    
    public synchronized void open() throws StreamException {
        open(0);
    }
    public void open(long timeout) throws StreamException {
        Logger              log = obtainLogger();
        StreamContext       out;
        StreamException     except = null;
        
        //Validate status (partially and harmlessly)
        if (!getCurrentStatus().isConnected())
            return;

        try {
            //Setup for Open
            setupOpen();
            openOutbound();
            openInbound(timeout);
        } catch (StreamException se) {
            except = se;
        } catch (Exception e) {
            except = new StreamException(getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION), e);
        }
        
        if (except != null) {
            try {
                close();
            } catch (Exception e) {
                obtainLogger().info("Exception thrown on open->close", e);
            }
            try {
                disconnect();
            } catch (Exception e) {
                obtainLogger().info("Exception thrown on open->disconnect", e);
            }

            obtainLogger().debug("throwing StreamException for open", except);
            throw except;
        }
    }
    protected abstract void setupOpen() throws Exception;
    public void openOutbound() throws StreamException {
        StreamContext       out = getOutboundContext();
        
        try {
            //Validate state (harmless)
            if (out.getCurrentStatus().isOpened())
                return;
            
            //Setup for Open
            if (!Utilities.isValidString(out.getNamespaceURI("")))
                out.addNamespace("", getDefaultNamespace());
            if (!Utilities.isValidString(out.getNamespacePrefix(out.getNamespaceURI())))
                out.addNamespace("stream", out.getNamespaceURI());
            
            //Try and send the opening
            _Outbound.sendOpen();
            fireStatusChangedEvent(out, Stream.OPENED, null);
        } catch (IOException ioe) {
            StreamError     err = getDataFactory().createStreamError(StreamError.REMOTE_CONNECTION_FAILED_CONDITION);
            StreamException se = new StreamException(err, ioe);
            
            obtainLogger().debug("throwing (I/O) StreamException for open(outbound)", se);
            throw se;
        } catch (Exception e) {
            StreamError     err = getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION);
            StreamException se = new StreamException(err, e);
            
            obtainLogger().debug("throwing (general) StreamException for open(outbound)", se);
            throw se;
        }
    }
    protected void openInbound(long timeout) throws StreamException {
        long        start, end;
        
        start = System.currentTimeMillis();
        while (getInboundStatus().isConnected() && !getInboundStatus().isOpened()) {
            process();
            
            end = System.currentTimeMillis();
            if ((timeout > 0) && ((end - start) > timeout)) {
                obtainLogger().warn("timeout reached on open");
                throw new StreamException(getDataFactory().createStreamError(StreamError.REMOTE_CONNECTION_FAILED_CONDITION));
            }
        }
    }

    
    public void dropInbound(boolean newSource) throws StreamException {        
        try {
            if (newSource) {
                _Inbound = new Input(getSource());
            }

            fireStatusChangedEvent(getInboundContext(),
                                   Stream.CONNECTED, null);

        } catch (Exception e) {
            StreamException se = new StreamException(getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION), e);
            
            obtainLogger().info("throwing StreamException for drop", se);
            throw se;
        }
    }
    
    public void dropOutbound(boolean newSource) throws StreamException {
        try {
            if (newSource) {
                _Outbound = new Output(getSource());
            }

            fireStatusChangedEvent(getOutboundContext(),
                                   Stream.CONNECTED, null);

        } catch (Exception e) {
            StreamException se = new StreamException(getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION), e);
            
            obtainLogger().info("throwing StreamException for drop", se);
            throw se;
        }
    }
    
    public void drop() throws StreamException {
        Logger      log = obtainLogger();
        
        try {
            //Setup I/O
            _Outbound = new Output(getSource());
            _Inbound = new Input(getSource());

            //Fire events
            fireStatusChangedEvent(getInboundContext(), Stream.CONNECTED, null);
            fireStatusChangedEvent(getOutboundContext(), Stream.CONNECTED, null);
        } catch (Exception e) {
            StreamException se = new StreamException(getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION), e);
            
            log.info("throwing StreamException for drop", se);
            throw se;
        }
    }
    
    /** <p>
     * Closes this <tt>Stream</tt>.  This method sends the outbound
     * &lt;/stream:stream&gt; and waits to receive the inbound
     * &lt;/stream:stream&gt; (if not already received).  This method does
     * not return until both closing &lt;/stream:stream&gt; tags have been
     * transferred.</p>
     *
     * <p>
     * Any exception thrown is caught and re-thrown as a
     * <tt>StreamException</tt>.</p>
     *
     * @throws StreamException If any errors are encountered.
     */
    public void close() throws StreamException {
        close(null, 0);
    }
    public void close(long timeout) throws StreamException {
        close(null, timeout);
    }
    public synchronized void close(StreamException se) throws StreamException {
    }
    public void close(StreamException se, long timeout) throws StreamException {
        Logger      log = obtainLogger();
        long        start, end;
        
        //Validate State (harmless)
        if (!getInboundStatus().isOpened() && !getOutboundStatus().isOpened())
            return;

        setStreamException(se);
            
        closeOutbound(se);
    
        try {
            //Prep the timeout
            start = System.currentTimeMillis();
            
            while (getInboundStatus().isOpened()) {
                process();
                end = System.currentTimeMillis();
                if ((timeout > 0) && ((end - start) > timeout))
                    throw new StreamException(getDataFactory().createStreamError(StreamError.REMOTE_CONNECTION_FAILED_CONDITION));
            }
        } catch (Exception e) {
            obtainLogger().info("exception thrown while closing", e);
            if (se == null) se = wrapException(e);
        }
        
        if (se != null) {
            obtainLogger().debug("throwing exception for close", se);
            throw se;
        }
    }
    public void closeOutbound() throws StreamException {
        closeOutbound(null);
    }

    public void closeOutbound(StreamException se) throws StreamException {
        if (!getOutboundStatus().isOpened())
            return;
        try {
            //Try and send the closing
            if (getOutboundStatus().isOpened()) {
                _Outbound.sendClose(se);
                fireStatusChangedEvent(getOutboundContext(), Stream.CLOSED, se);
            }
        } catch (Exception e) {
            obtainLogger().info("exception thrown while sending close", e);
            if (se == null) se = wrapException(e);
            throw se;
        }
    }

    private StreamException wrapException(Exception e) {
        StreamException se = null;
        if (e instanceof StreamException) {
            se = (StreamException)e;
        } else {
            StreamError err = getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION);
            err.setText(e.getMessage());
            se = new StreamException(err);
        }
        return se;
    }
    
    /** <p>
     * Send a <tt>Packet</tt> across this <tt>Stream</tt>.</p>
     *
     * <p>
     * If the value of <tt>p</tt> is <tt>null</tt>, this method sends
     * an empty string.</p>
     *
     * <p>
     * Any exception thrown (for reasons other than the above) is caught and
     * re-thrown as a <tt>StreamException</tt>.  Also, this
     * <tt>Stream</tt> will progress through {@link #close} and {@link
     * #disconnect}.</p>
     *
     * @param p The <tt>Packet</tt> to send.
     * @throws StreamException If any errors are encountered.
     */
    public void send(Packet p) throws StreamException {
        StreamError         err = null;
        StreamException     except = null;

        //Validate State (harmless)
        if (!getOutboundStatus().isOpened())
            return;

            try {
                //Send output Packet
                _Outbound.sendPacket(p);
            } catch (IOException ioe) {
                err = getDataFactory().createStreamError(StreamError.REMOTE_CONNECTION_FAILED_CONDITION);
                except = new StreamException(err, ioe);
            } catch (Exception e) {
                err = getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION);
                except = new StreamException(err, e);
            }
        
        //fire sent message
        firePacketTransferredEvent(PacketEvent.SENT);
        if (except != null) {
            //Close and Disconnect, then throw
            try {
                close(except);
            } catch (StreamException se) {
                obtainLogger().info("stream exception thrown on send->close", se);
            }
            try {
                disconnect();
            } catch (StreamException se) {
                obtainLogger().info("stream exception thrown on send->disconnect", se);
            }
            
            obtainLogger().debug("throwing exception for send", except);
            throw except;
        }
    }
    /** <p>
     * Processes <tt>Packet</tt> data coming into this
     * <tt>Stream</tt>.</p>
     *
     * <p>
     * Any exception thrown is caught and re-thrown as a
     * <tt>StreamException</tt>.  Also, this <tt>Stream</tt> will
     * progress through {@link #close} and {@link #disconnect}.</p>
     *
     * @throws StreamException If any errors are encountered.
     */
    public void process() throws StreamException {
        Logger              log = obtainLogger();
        StreamError         err = null;
        StreamException     except = null, caught = null;

        //Validate State (harmless)
        if (getInboundStatus().isConnected()) {
                try {
                    //Process inbound XML
                    _Inbound.process();
                } catch (IOException ioe) {
                    err = getDataFactory().createStreamError(StreamError.REMOTE_CONNECTION_FAILED_CONDITION);
                    caught = new StreamException(err, ioe);
                } catch (IxpException ixpe) {
                    err = getDataFactory().createStreamError(StreamError.INVALID_XML_CONDITION);
                    caught = new StreamException(err, ixpe);
                } catch (RuntimeException re) {
                    err = getDataFactory().createStreamError(StreamError.UNDEFINED_CONDITION);
                    caught = new StreamException(err, re);
                }
                except = getStreamException();
        }

        //firePacketTransferredEvent(PacketEvent.RECEIVED);
        if (except == null) {
            except = caught;
        }
        if (except == null) {
            except = getStreamException();
        }
        
        if (except != null) {
            Stream.Status   outStat = getOutboundStatus();
            
            //Close and Disconnect, then throw
            if ((outStat != CLOSED) && !outStat.isOpened()) {
                try {
                    openOutbound();
                } catch (Exception e) {
                    obtainLogger().error("failed to open outbound stream", e);
                }
            }
            try {
                if (getOutboundStatus().isOpened())
                    close((caught == getStreamException()) ? null : caught);
            } catch (StreamException se) {
                obtainLogger().info("stream exception thrown on process->close", se);
            }
            try {
                disconnect();
            } catch (StreamException se) {
                obtainLogger().info("stream exception thrown on process->disconnect", se);
            }
            
            obtainLogger().debug("throwing exception for process", except);
            throw except;
        }
    }
}
