/*
 * DataFactory.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package net.outer_planes.jso;

import java.util.Collections;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jaxen.XPath;

import org.saxpath.SAXPathException;

import org.jabberstudio.jso.JID;
import org.jabberstudio.jso.NSI;
import org.jabberstudio.jso.Extension;
import org.jabberstudio.jso.Packet;
import org.jabberstudio.jso.PacketError;
import org.jabberstudio.jso.StreamDataFactory;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.StreamElementFactory;
import org.jabberstudio.jso.StreamError;
import org.jabberstudio.jso.StreamText;
import org.jabberstudio.jso.SerializedNodes;
import org.jabberstudio.jso.io.StreamBuilder;
import org.jabberstudio.jso.util.Utilities;

import net.outer_planes.jso.xpath.JSOXPath;

/**
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public class DataFactory implements StreamDataFactory {
    //Fields
    private String      _DefaultNS;
    private DataFactory _BaseDF;
    private Map         _Registered;
    private Map         _NamespaceMappings;
    
    //Constructors
    public DataFactory() {
        this((String)null);
    }
    public DataFactory(String ns) {
        setDefaultNamespace(ns);
    }
    public DataFactory(DataFactory base) {
        this(base.getDefaultNamespace(), base);
    }
    public DataFactory(String ns, DataFactory base) {
        _BaseDF = base;
        setDefaultNamespace(ns);
    }
    
    //Methods
    public String getDefaultNamespace() {
        return _DefaultNS;
    }
    private void setDefaultNamespace(String ns) {
        _DefaultNS = (ns == null) ? "" : ns;
    }
    
    public Map getNamespaceMappings() {
        Map     baseMap = null;
        
        if(_BaseDF != null)
            baseMap = _BaseDF.getNamespaceMappings();
        
        if          ((baseMap == null) && (_NamespaceMappings == null)) {
             baseMap = new java.util.TreeMap();
        } else if   ((baseMap == null) && (_NamespaceMappings != null)) {
            baseMap = new java.util.TreeMap(_NamespaceMappings);
        } else if   ((baseMap != null) && (_NamespaceMappings != null)) {
            baseMap.putAll(_NamespaceMappings);
        }

        return baseMap;
    }
    public void setNamespaceMappings(Map nss) {
        Iterator    itr;
        Object      key, value;
        String      prefix, uri;
        
        if (nss == null)
            nss = java.util.Collections.EMPTY_MAP;
        
        //Out with the Old...
        if(_NamespaceMappings != null) {
            _NamespaceMappings.clear();
        }
        
        //...and in with the New
        itr = nss.keySet().iterator();
        while (itr.hasNext()) {
            key = itr.next();
            value = nss.get(key);
            
            prefix = (String)((key == null) ? null : key.toString());
            uri = ((value == null) ? null : value.toString());
            putNamespaceMapping(prefix, uri);
        }
    }
    public void putNamespaceMapping(String prefix, String uri) {
        if (prefix == null)
            prefix = "";
        if (uri == null)
            uri = "";
        
        if(_NamespaceMappings == null) {
             _NamespaceMappings = new java.util.TreeMap();
        }
        _NamespaceMappings.put(prefix, uri);
    }
    public void removeNamespaceMapping(String prefix) {
        if(_NamespaceMappings != null) {
            if (prefix == null)
                prefix = "";
            _NamespaceMappings.remove(prefix);
        }
    }
    
    public XPath createXPath(String expr) throws SAXPathException {
        return createXPath(expr, null);
    }
    public XPath createXPath(String expr, Map nss) throws SAXPathException {
        XPath                   xpath;
        Map                     mappings = getNamespaceMappings();
        
        //Setup mappings
        if (nss == null)
            nss = java.util.Collections.EMPTY_MAP;
        mappings.putAll(nss);
        
        xpath = new JSOXPath(expr);
        Utilities.setupNamespaces(xpath, mappings);
        
        return xpath;
    }
    
    public NSI createNSI(String name, String uri) throws IllegalArgumentException {
        return new NSI(name, uri);
    }

    public StreamElement createElement(String name) throws IllegalArgumentException {
        if (!Utilities.isValidString(name))
            throw new IllegalArgumentException("Element local name cannot be null or \"\"");
        return createElement(createNSI(name, null));
    }
    public StreamElement createElement(String name, String uri) throws IllegalArgumentException {
        if (!Utilities.isValidString(name))
            throw new IllegalArgumentException("Element local name cannot be null or \"\"");
        return createElement(createNSI(name, uri));
    }
    public StreamElement createElement(NSI name) throws IllegalArgumentException {
        return createElementNode(name, null);
    }
    
    public StreamText createText(String txt) {
        return new TextNode(this, txt);
    }

    public SerializedNodes createSerializedNodes(String txt){
        return new SerializedNodesImpl(this,txt);
    }
    
    public JID createJID(Object spec) throws IllegalArgumentException {
        return new JID(spec);
    }
    public JID createJID(String node, String domain, String resource) throws IllegalArgumentException {
        return new JID(node, domain, resource);
    }
    
    protected StreamElementFactory getElementFactory(NSI name, Class iface) {
        StreamElementFactory    matched = null;
        boolean                 hasLN, hasNS;
        
        if (name == null)
            name = new NSI(null, null);
        hasLN = Utilities.isValidString(name.getLocalName());
        hasNS = Utilities.isValidString(name.getNamespaceURI());
        
        if (hasNS)
            matched = locateElementFactory(name, iface);
        if ((matched == null) && hasNS)
            matched = locateElementFactory(new NSI(null, name.getNamespaceURI()), iface);
        if ((matched == null) && hasLN)
            matched = locateElementFactory(new NSI(name.getLocalName(), null), iface);
        if (matched == null)
            matched = locateElementFactory(new NSI(null, null), iface);
        
        return matched;
    }
    private final StreamElementFactory locateElementFactory(NSI name, Class iface) {
        List                    possibles = null;
        Iterator                itr;
        StreamElementFactory    test, matched = null;
        
        if(_Registered != null) {
            possibles = (List)_Registered.get(name);
            if (possibles != null) {
                itr = possibles.iterator();
                while ((matched == null) && itr.hasNext()) {
                    test = (StreamElementFactory)itr.next();
                    matched = (test.isSupportedElement(name, iface)) ? test : null;
                }
            }
        } 
        
        if(matched == null) {
            if(_BaseDF != null){
                matched = _BaseDF.locateElementFactory(name, iface);
             }
         }
        
        return matched;
        
    }
    
    public void registerElementFactory(StreamElementFactory sef) {
        Iterator    itr;
        
        if (sef != null) {
            itr = sef.getSupportedElements().iterator();
            
            while (itr.hasNext()) {
                registerElementFactoryDirect((NSI)itr.next(), sef);
            }
        }
    }
    public void registerElementFactory(NSI name, StreamElementFactory sef) {
        List        factories;
        
        if ((name != null) && (sef != null) && sef.isSupportedElement(name, null)) {
            registerElementFactoryDirect(name, sef);
        }
    }
    protected void registerElementFactoryDirect(NSI name, StreamElementFactory sef) {
        List        factories;
        
        if (_Registered == null)
            _Registered = new java.util.TreeMap(NSI.STRICT_COMPARATOR);
        
        factories = (List)_Registered.get(name);
        if (factories == null)
            _Registered.put(name, factories = new java.util.ArrayList());
        factories.add(0, sef);
    }
    
    public void unregisterElementFactory(StreamElementFactory sef) {
        Iterator    itr;
        
        if (sef != null) {
            itr = sef.getSupportedElements().iterator();
            
            while (itr.hasNext()) {
                unregisterElementFactoryDirect((NSI)itr.next(), sef);
            }
        }
    }
    public void unregisterElementFactory(NSI name, StreamElementFactory sef) {
        List        factories;
        
        if ((name != null) && (sef != null) && sef.isSupportedElement(name, null)) {
            unregisterElementFactoryDirect(name, sef);
        }
    }
    protected void unregisterElementFactoryDirect(NSI name, StreamElementFactory sef) {
        if(_Registered != null) {
            List        factories = (List)_Registered.get(name);

            if (factories != null) {
                factories.remove(sef);
                if (factories.isEmpty())
                    _Registered.remove(name);
            }
        }
    }
    
    public Set getSupportedElements() {
        Set     regSet = null;
        Set     baseSet = null;
        Set     result;
        
        if (_Registered != null)
            regSet = _Registered.keySet();
        if (_BaseDF != null) {
            baseSet = _BaseDF.getSupportedElements();
        }
        
        if          ((baseSet == null) && (regSet == null)) {
            result = java.util.Collections.EMPTY_SET;
        } else if   ((baseSet != null) && (regSet == null)) {
            result = baseSet;
        } else if   ((baseSet == null) && (regSet != null)) {
             result = java.util.Collections.unmodifiableSet(regSet);
        } else {
            //combine both
            result = new java.util.TreeSet();
            result.addAll(regSet);
            result.addAll(baseSet);
        }
        
        return result;
    }
    public boolean isSupportedElement(NSI name, Class iface) {
        return getElementFactory(name, iface) != null;
    }
    
    public StreamBuilder createElementBuilder(NSI name) throws IllegalArgumentException {
        return createElementBuilder(name, null);
    }
    public StreamBuilder createElementBuilder(NSI name, Class iface) throws IllegalArgumentException {
        StreamElementFactory    factory = getElementFactory(name, iface);
        
        if (factory == null)
            throw new IllegalArgumentException("No StreamElementFactory for the given NSI and interface");
        return factory.createElementBuilder(name, iface, this);
    }

    public StreamElement createElementNode(NSI name) throws IllegalArgumentException {
        return createElementNode(name, null);
    }
    public StreamElement createElementNode(NSI name, Class iface) throws IllegalArgumentException {
        StreamElementFactory    factory = getElementFactory(name, iface);
        
        if (factory == null)
            throw new IllegalArgumentException("No StreamElementFactory for the given NSI and interface");
        return factory.createElementNode(name, iface, this);
    }
    
    public StreamError createStreamError(String cond) throws IllegalArgumentException {
        StreamError     err;
        
        if (!Utilities.isValidString(cond))
            cond = StreamError.UNDEFINED_CONDITION;
        
        err = (StreamError)createElementNode(createNSI("error", Utilities.STREAM_NAMESPACE), StreamError.class);
        err.setDefinedCondition(cond);
        
        return err;
    }
    public PacketError createPacketError(PacketError.Type type, String cond) throws IllegalArgumentException {
        PacketError     err;
        
        if (type == null)
            throw new IllegalArgumentException("type cannot be null");
        if (!Utilities.isValidString(cond))
            cond = PacketError.UNDEFINED_CONDITION;
        
        err = (PacketError)createElementNode(createNSI("error", null), PacketError.class);
        err.setType(type);
        err.setDefinedCondition(cond);
        
        return err;
    }
    public PacketError createError(int code) throws IllegalArgumentException {
        return createError(code, null);
    }
    public PacketError createError(int code, String msg) throws IllegalArgumentException {
        PacketError     err = (PacketError)createElementNode(new NSI("error", null), PacketError.class);
        
        err.setCode(code);
        err.setText(msg);
        
        return err;
    }
    
    public Packet createPacketNode(NSI name) throws IllegalArgumentException {
        return (Packet)createElementNode(name, Packet.class);
    }
    public Packet createPacketNode(NSI name, Class iface) throws IllegalArgumentException {
        if (iface == null)
            iface = Packet.class;
        if (!Packet.class.isAssignableFrom(iface))
            throw new IllegalArgumentException("Interface " + iface.getName() + " is not derived from Packet");

        return (Packet)createElementNode(name, iface);
    }
    public StreamBuilder createPacketBuilder(NSI name) throws IllegalArgumentException {
        return createElementBuilder(name, Packet.class);
    }
    public StreamBuilder createPacketBuilder(NSI name, Class iface) throws IllegalArgumentException {
        if (iface == null)
            iface = Packet.class;
        if (!Packet.class.isAssignableFrom(iface))
            throw new IllegalArgumentException("Interface " + iface.getName() + " is not derived from Packet");

        return createElementBuilder(name, iface);
    }
    
    public Extension createExtensionNode(NSI name) throws IllegalArgumentException {
        return (Extension)createElementNode(name, Extension.class);
    }
    public Extension createExtensionNode(NSI name, Class iface) throws IllegalArgumentException {
        if (iface == null)
            iface = Extension.class;
        if (!Extension.class.isAssignableFrom(iface))
            throw new IllegalArgumentException("Interface " + iface.getName() + " is not derived from Extension");

        return (Extension)createElementNode(name, iface);
    }
    public StreamBuilder createExtensionBuilder(NSI name) throws IllegalArgumentException {
        return createElementBuilder(name, Extension.class);
    }
    public StreamBuilder createExtensionBuilder(NSI name, Class iface) throws IllegalArgumentException {
        if (iface == null)
            iface = Extension.class;
        if (!Extension.class.isAssignableFrom(iface))
            throw new IllegalArgumentException("Interface " + iface.getName() + " is not derived from Extension");

        return createElementBuilder(name, iface);
    }
    
    public Object clone() {
        return new DataFactory(this);
    }
}
