/*
 * ElementNode.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package net.outer_planes.jso;

import java.util.Iterator;
import java.util.List;

import org.jabberstudio.jso.NSI;
import org.jabberstudio.jso.StreamDataFactory;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.StreamNode;
import org.jabberstudio.jso.StreamObject;
import org.jabberstudio.jso.io.StreamBuilder;
import org.jabberstudio.jso.util.Utilities;

/**
 * <p>
 * Basic implementation for the <code>StreamElement</code> interface.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 *
 * Modified by Sun Microsystems on Thu, Nov 03, 2005
 *    added synchronization around _Nodes.
 */
public class ElementNode extends AbstractElement {
    //Fields
    private StreamElement       _Parent;

    private NSI                 _Name;
    private List                _Nodes = new java.util.LinkedList();

    //Constructors
    public ElementNode(StreamDataFactory sdf, NSI name) {
        super(sdf);
        setNSI(name);
    }
    protected ElementNode(StreamElement p, ElementNode cp) {
        super(p, cp);
        
        Iterator        itr;
        StreamNode      node;

        //Copy child nodes, properly setting parentage
        itr = cp.getNodes().iterator();
        while (itr.hasNext()) {
            node = (StreamNode)itr.next();
             node = (StreamNode)node.copy(this);
            _Nodes.add(node);
        }
    }

    //Methods
    public StreamBuilder createBuilder() {
        return new ElementBuilder(this);
    }
    
    public NSI getNSI() {
        return _Name;
    }
    public void setNSI(NSI name) throws IllegalArgumentException {
        if (name == null)
            throw new IllegalArgumentException("Name cannot be null");
        if (_Name != null)
            checkNSI(name);
        _Name = name;
    }

    /** <p>
     * Retrieves the parent <code>StreamElement</code> for this
     * <code>StreamNode</code>.</p>
     *
     * <p>
     * The value returned by this method may be <code>null</code>, if the
     * parent has not yet been defined.</p>
     *
     * @return The <code>StreamElement</code> parent.
     */
    public StreamElement getParent() {
        return _Parent;
    }
    /** <p>
     * Sets the parent <code>StreamElement</code> for this
     * <code>StreamNode</code>.</p>
     *
     * <p>
     * The value of <code>p</code> may be <code>null</code>, in which case
     * the current parent (if any)for this <code>StreamElement</code> is
     * detached.  If the parent cannot be changed, this method throws an
     * <code>IllegalStateException</code>.</p>
     *
     * @param p the <code>StreamElement</code> parent.
     * @throws IllegalStateException If the parent cannot be changed.
     */
    public void setParent(StreamElement p) throws IllegalStateException {
        detach();
        _Parent = p;
    }

    /** <p>
     * Lists the <code>StreamNode</code> objects that are contained by this
     * <code>StreamNode</code>.  This list may or may not be modifiable.</p>
     *
     * <p>
     * The value returned by this method is never <code>null</code>.</p>
     *
     * @return The list of contained <code>StreamNode</code> objects.
     */
    public List listNodes() {
	synchronized(_Nodes) {
	    return new java.util.ArrayList(getNodes());
	}
    }
    private List getNodes() {
        return _Nodes;
    }
    
    /** <p>
     * Creates an "almost-exact" duplicate of this <code>StreamNode</code>,
     * using the given parent <code>StreamElement</code>.  By "almost-exact",
     * all information for this <code>StreamNode</code>, and its contained
     * objects (obtained via <code>{@link #listNodes}</code>) are duplicated.
     * The parent for this <code>StreamNode</code> is one specified by
     * <code>p</code>, rather than thatreferenced by this
     * <code>StreamNode</code>.</p>
     *
     * <p>
     * The value of <code>p</code> may be <code>null</code>, in which case no
     * parent is attached to the newly created <code>StreamNode</code>.</p>
     *
     * @return the "almost-exact" duplicate of this <code>StreamNode</code>.
     */
    public StreamObject copy(StreamElement p) {
        return new ElementNode(p, this);
    }
    
    /** <p>
     * Adds the given <code>StreamNode</code> as a child to this
     * <code>StremElement</code>.</p>
     *
     * @param elem The <code>StreamNode</code> to add.
     * @throws UnsupportedOperationException If the change is not permitted.
     */
    public void add(StreamNode node) throws UnsupportedOperationException {
        if (node != null) {
            node.setParent(this);
	    synchronized(_Nodes) {
                _Nodes.add(node);
	    }
        }
    }
    /**
     * <p>
     * Inserts the given <code>StreamNode</code> at the given index, as a
     * child to this <code>StreamElement</code>.</p>
     *
     * <p>
     * If the value of <code>idx</code> is less than 0, either it will be
     * considered 0 or an <code>IllegalArgumentException</code> may be thrown.
     * If the value of <code>idx</idx> is greater than or equal to the number
     * of child nodes already contained, this method performs the same
     * operation that <code>{@link #add}</code> would.</p>
     *
     * <p>
     * If the value of <code>node</code> is <code>null</code>, no actions are
     * performed.  If <code>node</code> already has another parent, and that
     * parent is not this <code>StreamElement</code>, this method throws an
     * <code>IllegalArgumentException</code>.</p>
     *
     * @param idx  The index to insert the <code>StreamElement</code>.
     * @param node The <code>StreamNode</code> to insert.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws UnsupportedOperationException If the change is not permitted.
     */
    public void insert(int idx, StreamNode node) throws IllegalArgumentException, UnsupportedOperationException {
        if (node != null) {
	    synchronized(_Nodes) {
		_Nodes.add(idx, node);
	    }
            node.setParent(this);
        }
    }
    /** <p>
     * Removed the given <code>StreamNode</code> from this
     * <code>StremElement</code>.</p>
     *
     * @param node The <code>StreamNode</code> to remove.
     */
    public void remove(StreamNode node) throws UnsupportedOperationException {
        if (node != null) {
	    boolean b = false;
	    synchronized(_Nodes) {
		b = _Nodes.remove(node);
	    }
            if (b) node.setParent(null);
        }
    }
    
    public void clear() {
	synchronized(_Nodes) {
	    _Nodes.clear();
	}
    }
    
    protected void checkNSI(NSI check) throws IllegalArgumentException {
        if (!Utilities.isValidString(check.getLocalName()))
            throw new IllegalArgumentException("Local name cannot be null or \"\"");
    }
    protected void checkElement(StreamElement check) throws IllegalArgumentException {
    }
}
