/*
 * PresencePacket.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package net.outer_planes.jso;


import java.util.Iterator;
import java.util.Locale;

import org.jabberstudio.jso.NSI;
import org.jabberstudio.jso.Packet;
import org.jabberstudio.jso.Presence;
import org.jabberstudio.jso.StreamDataFactory;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.StreamObject;
import org.jabberstudio.jso.util.Utilities;

/**
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public class PresencePacket extends BasicPacketNode implements Presence {
    //"Constants"
    public static final NSI NAME = new NSI("presence", null);
    
    public static final NSI NAME_PRIORITY = new NSI("priority", null);
    public static final NSI NAME_SHOW = new NSI("show", null);
    public static final NSI NAME_STATUS = new NSI("status", null);
    
    //Constructors
    /** Creates a new instance of PresencePacket */
    public PresencePacket(StreamDataFactory sdf) {
        this(sdf, NAME);
    }
    public PresencePacket(StreamDataFactory sdf, NSI name) {
        super(sdf, name);
    }
    public PresencePacket(StreamElement parent, PresencePacket base) {
        super(parent, base);
    }
    
    //Methods
    public void checkNSI(NSI check) throws IllegalArgumentException {
        if (NSI.NS_LOOSE_COMPARATOR.compare(check, NAME) != 0)
            throw new IllegalArgumentException("Name must be 'presence'");
    }
    
    public StreamObject copy(StreamElement p) {
        return new PresencePacket(p, this);
    }
    
    public void setType(Packet.Type type) throws IllegalArgumentException {
        if ((type == null) || (type == ERROR) || (type instanceof Presence.Type))
            super.setType(type);
        else
            throw new IllegalArgumentException("Only ERROR or presence types permitted");
    }
    
    protected Packet.Type determineType(String name) {
        Packet.Type     type;
        
        if          (Utilities.equateStrings(INVISIBLE.toString(), name)) {
            type = INVISIBLE;
        } else if   (Utilities.equateStrings(PROBE.toString(), name)) {
            type = PROBE;
        } else if   (Utilities.equateStrings(SUBSCRIBE.toString(), name)) {
            type = SUBSCRIBE;
        } else if   (Utilities.equateStrings(SUBSCRIBED.toString(), name)) {
            type = SUBSCRIBED;
        } else if   (Utilities.equateStrings(UNAVAILABLE.toString(), name)) {
            type = UNAVAILABLE;
        } else if   (Utilities.equateStrings(UNSUBSCRIBE.toString(), name)) {
            type = UNSUBSCRIBE;
        } else if   (Utilities.equateStrings(UNSUBSCRIBED.toString(), name)) {
            type = UNSUBSCRIBED;
        } else {
            type = super.determineType(name);
        }
        
        return type;
    }

    protected StreamElement getPriorityElement() {
        return getFirstElement(NAME_PRIORITY.getLocalName());
    }
    public boolean hasPriority() {
        return (getPriorityElement() != null);
    }
    public int getPriority() {
        StreamElement   elem = getPriorityElement();
        int             pri = 0;

        if (elem != null) {
            try { pri = Integer.parseInt(elem.normalizeTrimText()); } catch (NumberFormatException nfe) {}
        }
        
        return pri;
    }
    public void setPriority(int p) {
        StreamElement   elem;
        
        if ((p < Byte.MIN_VALUE) || (p > Byte.MAX_VALUE))
            throw new IllegalArgumentException("priority must be between -127 and 128 (inclusive)");
        
        elem = getPriorityElement();
        if (elem == null)
            elem = addElement(NAME_PRIORITY.getLocalName());
        elem.clearText();
        elem.addText(String.valueOf(p));
    }
    public void removePriority() {
        clearElements(NAME_PRIORITY.getLocalName());
    }
    
    public String getShow() {
        StreamElement   elem = getFirstElement(NAME_SHOW.getLocalName(), getNamespaceURI());
        String          show = null;

        if (elem != null) {
            show = elem.normalizeTrimText();
        }
        
        return (show != null) ? show : "";
    }
    public void setShow(String s) {
        clearElements(NAME_SHOW.getLocalName(), getNamespaceURI());
        
        if (s != null) {
            addElement(NAME_SHOW.getLocalName(), getNamespaceURI()).addText(s);
        }
    }

    public boolean hasStatus() {
        return hasStatus(null);
    }
    public boolean hasStatus(Locale l) {
        return !findElementsForLocale(NAME_STATUS.getLocalName(), getNamespaceURI(), l).isEmpty();
    }
    public String getStatus() {
        return getStatus(null);
    }
    public String getStatus(Locale l) {
        Iterator        itr = findElementsForLocale(NAME_STATUS.getLocalName(), getNamespaceURI(), l).iterator();
        String          status = "";
        
        if (itr.hasNext()) {
            status = ((StreamElement)itr.next()).normalizeTrimText();
        }
        
        return status;
    }
    public void setStatus(String s) {
        setStatus(null, s);
    }
    public void setStatus(Locale l, String s) {
        Iterator        itr = findElementsForLocale(NAME_STATUS.getLocalName(), getNamespaceURI(), l).iterator();
        StreamElement   elem;
        
        while (itr.hasNext())
            ((StreamElement)itr.next()).detach();
        
        if (s != null) {
            elem = addElement(NAME_STATUS.getLocalName(), getNamespaceURI());
            elem.addText(s);
            if (l != null)
                elem.setDeclaredLocale(l);
        }
    }
}