/*
 * ServerPLAINMechanism.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package net.outer_planes.jso.sasl.mechanisms;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.List;
import java.util.Map;

import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.UnsupportedCallbackException;

import org.jabberstudio.jso.JID;
import org.jabberstudio.jso.sasl.SASLAuthPacket;
import org.jabberstudio.jso.sasl.SASLFailedException;
import org.jabberstudio.jso.sasl.SASLFailurePacket;
import org.jabberstudio.jso.sasl.SASLPacket;
import org.jabberstudio.jso.sasl.SASLServerInfo;
import org.jabberstudio.jso.sasl.callback.ConfirmAuthorizationCallback;

import net.outer_planes.jso.sasl.ServerMechanism;

/**
 *
 * @author  linuxwolf
 * @version 0.13-DEV-20051010073318
 * @since   JSO-0.12
 */
public class ServerPLAINMechanism extends ServerMechanism {
    //Constructors
    public ServerPLAINMechanism(SASLServerInfo si) {
        super(si);
    }

    //Methods
    public String getMechanismName() {
        return "PLAIN";
    }
    
    protected SASLPacket evaluateAuth(SASLAuthPacket auth) throws IllegalStateException, SASLFailedException {
        SASLPacket  output = super.evaluateAuth(auth);
        byte[]      data = auth.getData();
        
        if (data.length > 0)
            output = evaluatePlain(data);
        
        return output;
    }
    protected SASLPacket evaluateResponse(SASLPacket resp) throws IllegalStateException, SASLFailedException {
        return evaluatePlain(resp.getData());
    }
    
    private final SASLPacket evaluatePlain(byte[] data) throws IllegalStateException, SASLFailedException {
        SASLPacket          output = null;
        Map                 params = readParameters(data);
        List                calls = new java.util.ArrayList();
        Object              user, pass, authz, authn;
        CallbackHandler     cbh = getCallbackHandler();
        
        if (cbh != null) {
            NameCallback                    nc = null;
            PasswordCallback                pc = null;
            ConfirmAuthorizationCallback    cac = null;
            String                          actualPass, expectedPass;
            
            user = params.get("javax.security.sasl.username");
            pass = params.get("javax.security.sasl.password");
            authz = params.get("javax.security.sasl.authzid");
            authn = null;

            try {
                calls.add(nc = new NameCallback("username", (String)user));
                calls.add(pc = new PasswordCallback("password", false));
                if (authz != null)
                    calls.add(new ConfirmAuthorizationCallback((JID)authn, (JID)authz));
                
                cbh.handle((Callback[])calls.toArray(new Callback[calls.size()]));
            } catch (ClassCastException cce) {
                throw completeFailure(SASLFailurePacket.INVALID_AUTHZID_CONDITION);
            } catch (Exception e) {
                throw completeFailure(SASLFailurePacket.TEMPORARY_AUTH_FAILURE_CONDITION);
            }
            
            actualPass = ((pass != null) ? pass.toString() : "");
            expectedPass = ((pc.getPassword() != null) ? new String(pc.getPassword()) : "");
            
            if (!actualPass.equals(expectedPass))
                throw completeFailure(SASLFailurePacket.NOT_AUTHORIZED_CONDITION);
            completeSuccess((SASLPacket)getDataFactory().createPacketNode(SASLPacket.NAME_SUCCESS));
        } else {
            throw completeFailure(SASLFailurePacket.TEMPORARY_AUTH_FAILURE_CONDITION);
        }
        
        return output;
    }
    private final Map readParameters(byte[] input) throws SASLFailedException {
        String[]    data;
        int         idx;
        
        try {
            data = new String(input, "UTF-8").split("\0");
            
            for (idx = 0; idx < data.length; idx++) {
                switch(idx) {
                    case 0:     //Authzid
                        setProperty("javax.security.sasl.authzid", data[idx]);
                        break;
                    case 1:
                        setProperty("javax.security.sasl.username", data[idx]);
                        break;
                    case 2:
                        setProperty("javax.security.sasl.password", data[idx]);
                        break;
                }
            }
        } catch (UnsupportedEncodingException uee) {
            throw completeFailure(SASLFailurePacket.INCORRECT_ENCODING_CONDITION);
        }
        
        return getProperties();
    }
}