/*
 * JID.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso;


import java.util.Comparator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.jabberstudio.jso.util.Utilities;
import org.jabberstudio.jso.util.stringprep.DomainprepProfile;
import org.jabberstudio.jso.util.stringprep.NodeprepProfile;
import org.jabberstudio.jso.util.stringprep.ResourceprepProfile;
import org.jabberstudio.jso.util.stringprep.StringprepException;

/**
 * <p>
 * Object representation of a Jabber Identifier (JID).  This
 * class performs the necessary validation upon construction.  This
 * class also provides both a strict equality test and loose match test,
 * and a "natural ordering" based on the pattern
 * (domain =&gt; node =&gt; resource).</p>
 *
 * <p>
 * The JID uses "node" to refer to the "username" portion, "domain" to
 * refer to the "hostname" portion, and "resource" for the JID's
 * resource.  The JID allows "empty" values for <tt>{@link #getNode}</tt>
 * and <tt>{@link #getResource}</tt>.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public final class JID implements Comparable, java.io.Serializable {

    private static NodeprepProfile _Nodeprep = NodeprepProfile.getInstance();
    private static ResourceprepProfile _Resourceprep = ResourceprepProfile.getInstance();
    private static DomainprepProfile _Domainprep = DomainprepProfile.getInstance();

    //"Constants"
    /** <p>
     * <tt>Comparator</tt> for "strict" <tt>JID</tt> comparisons. This is
     * also the "natural ordering" algorithm for the <tt>JID</tt> class.</p>
     *
     * @since   JSO-0.12
     */
    public static final Comparator  STRICT_COMPARATOR = new Comparator() {
        public int compare(Object obj1, Object obj2) throws NullPointerException, JIDFormatException {
            JID     jid1, jid2;
            String  test1, test2;
            int     code = 0;
            
            //setup JIDs
            jid1 = ((obj1 != null) ? ((obj1 instanceof JID) ? (JID)obj1 : valueOf(obj1)) : null);
            jid2 = ((obj2 != null) ? ((obj2 instanceof JID) ? (JID)obj2 : valueOf(obj2)) : null);
            
            if ((jid1 == null) || (jid2 == null))
                throw new NullPointerException("comparison objects cannot be null");
            
            //Check domain
            test1 = jid1.getDomain();
            test2 = jid2.getDomain();
            code = test1.compareToIgnoreCase(test2);

            if (code == 0) {
                //Check node
                test1 = jid1.getNode();
                test2 = jid2.getNode();
                code = test1.compareToIgnoreCase(test2);
            }

            if (code == 0) {
                //Check resource
                test1 = jid1.getResource();
                test2 = jid2.getResource();
                code = test1.compareTo(test2);
            }

            return code;
        }
    };
    /** <p>
     * <tt>Comparator</tt> for "loose" <tt>JID</tt> comparisons. This
     * algorithm only considers the node and resource if both objects
     * contain a non-empty node and resource.</p>
     *
     * @since   JSO-0.12
     */
    public static final Comparator  LOOSE_COMPARATOR = new Comparator() {
        public int compare(Object obj1, Object obj2) {
            JID     jid1, jid2;
            String  test1, test2;
            int     code = 0;
            
            //setup JIDs
            jid1 = ((obj1 != null) ? ((obj1 instanceof JID) ? (JID)obj1 : valueOf(obj1)) : null);
            jid2 = ((obj2 != null) ? ((obj2 instanceof JID) ? (JID)obj2 : valueOf(obj2)) : null);
            
            if ((jid1 == null) || (jid2 == null))
                throw new NullPointerException("comparison objects cannot be null");
            
            //Check domain
            test1 = jid1.getDomain();
            test2 = jid2.getDomain();
            code = test1.compareToIgnoreCase(test2);

            if (code == 0) {
                //Check node
                test1 = jid1.getNode();
                test2 = jid2.getNode();
                code = ((Utilities.isValidString(test1) && Utilities.isValidString(test2)) ? test1.compareToIgnoreCase(test2) : 0);
            }

            if (code == 0) {
                //Check resource
                test1 = jid1.getResource();
                test2 = jid2.getResource();
                code = ((Utilities.isValidString(test1) && Utilities.isValidString(test2)) ? test1.compareToIgnoreCase(test2) : 0);
            }

            return code;
        }
    };
    
    private static final int NODE_LENGTH        = 1024;
    private static final int DOMAIN_LENGTH      = 1024;
    private static final int RESOURCE_LENGTH    = 1024;
    
    private static final Pattern    RE_PATTERN      = Pattern.compile("(([^@]+)@)?([^/]+)(/(.*))?", Pattern.DOTALL);
    
    private static final int        POS_NODE        = 2;
    private static final int        POS_DOMAIN      = 3;
    private static final int        POS_RESOURCE    = 5;
    
    //Fields
    private String          _Node;
    private String          _Domain;
    private String          _Resource;
    
    private String          _Spec;

    //Constructors
    /**
     * <p>
     * Creates a new <tt>JID</tt> with the given parameters.
     * The values for <tt>node</tt> and/or <tt>resource</tt>
     * may be <tt>null</tt>, but <tt>domain</tt> may not be.</p>
     *
     * <p>
     * Note that this constructor throws <tt>JIDFormatException</tt>
     * if <tt>node</tt> and/or <tt>domain</tt> do not meet the
     * 'stringprep' resources outlined in the XMPP specifications.</p>
     *
     * @param node The node for this <tt>JID</tt>.
     * @param domain The domain for this <tt>JID</tt>.
     * @param resource The resource for this <tt>JID</tt>.
     * @throws JIDFormatException If <tt>node</tt>,
     * <tt>domain</tt>, or <tt>resource</tt> do not meet RFC
     * requirements.
     * @since   JSO-0.2
     */    
    public JID(String node, String domain, String resource) throws IllegalArgumentException {
        setDomain(domain);
        setNode(node);
        setResource(resource);
    }
    /**
     * <p>
     * Creates a new <tt>JID</tt> with the given string. The string must
     * be parsable from the form "[node@]domain[/resource]".</p>
     *
     * <p>
     * Note that this constructor throws <tt>JIDFormatException</tt>
     * if <tt>node</tt> and/or <tt>domain</tt> do not meet the
     * 'stringprep' resources outlined in the XMPP specifications.</p>
     *
     * @param spec The fully-qualified string to parse.
     * @throws JIDFormatException If <tt>spec</tt> is invalid.
     * @since   JSO-0.12
     */    
    public JID(String spec) throws JIDFormatException {
        Matcher     match = RE_PATTERN.matcher((spec != null) ? spec : "");

        if (!match.matches())
            throw new JIDFormatException("object does not represent a valid JID");

        setDomain(match.group(POS_DOMAIN));
        setNode(match.group(POS_NODE));
        setResource(match.group(POS_RESOURCE));
    }
    /**
     * <p>
     * Creates a new <tt>JID</tt> with the given object. The string
     * representation of the given object must be parsable from the form
     * "[node@]domain[/resource]".</p>
     *
     * <p>
     * Note that this constructor throws <tt>JIDFormatException</tt>
     * if <tt>domain</tt>, <tt>node</tt>, and/or <tt>resource</tt> do not meet
     * the 'stringprep' resources outlined in the XMPP specifications.</p>
     *
     * @param spec The fully-qualified string to parse.
     * @throws JIDFormatException If <tt>spec</tt> is invalid.
     * @since   JSO-0.8.2
     * @deprecated As of JSO-0.12, this constructor is replaced by {@link
     * #JID(String)}.
     */    
    public JID(Object spec) throws JIDFormatException {
        this((spec != null) ? spec.toString() : "");
    }

    private JID() { }

    //Methods
     private final String getSpecification() {
         return _Spec;
     }
     private final void setSpecification(String spec) {
         _Spec = spec;
     }

    /**
     * <p>
     * Retrieves the node for this <tt>JID</tt>.  The node is
     * synonymous with "username" for user JID's.</p>
     *
     * <p>
     * This method never returns <tt>null</tt>, although ""
     * may be returned if the JID has no node.</p>
     *
     * @return The node for this <tt>JID</tt>.
     * @since   JSO-0.2
     */     
    public final String getNode() {
        return _Node;
    }
    /**
     * <p>
     * Determines if this <tt>JID</tt> contains a valid node.</p>
     *
     * @return <tt>true</tt> if the node is not <tt>null</tt> or "".
     * @since   JSO-0.12
     */
    public final boolean hasNode() {
        return Utilities.isValidString(_Node);
    }
    private final void setNode(String node) throws JIDFormatException {
        try {
            _Node = (Utilities.isValidString(node) ?
                _Nodeprep.prepare(node) :
                "");
        } catch (StringprepException spe) {
            throw new JIDFormatException("node is not valid");
        }
    }
    /**
     * <p>
     * Determines if the given string is a valid JID node. This method
     * performs the necessary checks for length and pattern matching.</p>
     *
     * <p>
     * The value of <tt>node</tt> may be <tt>null</tt> or "", in which case
     * this method returns <tt>false</tt>.</p>
     *
     * @param node The string to validate.
     * @return <tt>true</tt> If <tt>node</tt> is a valid JID node.
     * @since   JSO-0.12
     */
    public static final boolean isValidNode(String node) {
        return (Utilities.isValidString(node) ?
            _Nodeprep.validate(node) :
            true);
    }

    /**
     * <p>
     * Retrieves the domain for this <tt>JID</tt>.  The domain
     * is synonymous with "hostname" for user JID's.</p>
     *
     * <p>
     * This method never returns <tt>null</tt> or "".  Also,
     * the value returned is a Fully-qualified Domain Name (FQDN).</p>
     *
     * @return The domain for this <tt>JID</tt>.
     * @since   JSO-0.2
     */    
    public final String getDomain() {
        return _Domain;
    }
    private final void setDomain(String domain) throws JIDFormatException {
        try {
            _Domain = _Domainprep.prepare(domain);
        } catch (StringprepException spe) {
	    spe.printStackTrace(); // debug
            throw new JIDFormatException(domain + " is not a valid FQDN or IP address: " + spe.toString());
        }
    }
    /**
     * <p>
     * Determines if the given string is a valid JID domain. This method
     * performs the necessary checks for length and pattern matching.</p>
     *
     * <p>
     * The value of <tt>domain</tt> may be <tt>null</tt> or "", in which case
     * this method returns <tt>false</tt>.</p>
     *
     * @param domain The string to validate.
     * @return <tt>true</tt> If <tt>domain</tt> is a valid JID domain.
     * @since   JSO-0.12
     */
    public static final boolean isValidDomain(String domain) {
        return (Utilities.isValidString(domain) ?
            _Domainprep.validate(domain) :
            false);
    }

    /**
     * <p>
     * Retrieves the resource for this <tt>JID</tt>.  The
     * resource is used to further qualify a node at a domain.</p>
     *
     * <p>
     * This method never returns <tt>null</tt>, although it may
     * return "" if there is no resource defined.  Also, unlike the
     * node and domain, this value is case-sensitive.</p>
     *
     * @return The resource for this <tt>JID</tt>.
     * @since   JSO-0.2
     */    
    public final String getResource() {
        return _Resource;
    }
    /**
     * <p>
     * Determines if this <tt>JID</tt> contains a valid resource.</p>
     *
     * @return <tt>true</tt> if the resource is not <tt>null</tt> or "".
     * @since   JSO-0.12
     */
    public final boolean hasResource() {
        return Utilities.isValidString(_Resource);
    }
    private final void setResource(String res) throws JIDFormatException {
        try {
            _Resource = (Utilities.isValidString(res) ?
                _Resourceprep.prepare(res) :
                "");
        } catch (StringprepException spe) {
            throw new JIDFormatException("resource is not valid");
        }
    }
    /**
     * <p>
     * Determines if the given string is a valid JID resource. This method
     * performs the necessary checks for length and pattern matching.</p>
     *
     * <p>
     * The value of <tt>res</tt> may be <tt>null</tt> or "", in which case
     * this method returns <tt>true</tt>.</p>
     *
     * @param res The string to validate.
     * @return <tt>true</tt> If <tt>res</tt> is a valid JID resource.
     * @since   JSO-0.12
     */
    public static final boolean isValidResource(String res) {
        return (Utilities.isValidString(res) ?
            _Resourceprep.validate(res) :
            true);
    }
    
    /**
     * <p>
     * Generates the hashcode for this <tt>JID</tt>.</p>
     *
     * @return The hashcode for this <tt>JID</tt>.
     * @since   JSO-0.8.5
     */
    public final int hashCode() {
        int     results = 23;
        
        results = 41 * results + getDomain().toLowerCase().hashCode();
        if (hasNode())
            results = 41 * results + getNode().toLowerCase().hashCode();
        if (hasResource())
            results = 41 * results + getResource().hashCode();
        
        return results;
    }
    
    /**
     * <p>
     * Tests this <tt>JID</tt> for equality to the given
     * object.  This test is "strict", meaning that this <tt>JID</tt>
     * must match <tt>cmp</tt> exactly in node, domain, and resource.  The
     * test of node and domain are case-insensitive, while the test for
     * resource is case-sensitive.</p>
     *
     * <p>
     * If <tt>cmp</tt> is <tt>null</tt>, or is not a <tt>JID</tt>
     * and the results for <tt>toString()</tt> do not result in a valid
     * <tt>JID</tt>, this method returns <tt>false</tt>.</p>
     *
     * @param cmp The Object to compare to.
     * @return <tt>true</tt> if this <tt>JID</tt> is strictly
     * equal to <tt>cmp</tt>.
     * @since   JSO-0.2
     */    
    public final boolean equals(Object cmp) {
        JID     test = null;

        if (cmp instanceof JID) {
            test = (JID)cmp;
        } else if (isValid(cmp)) {
            test = valueOf(cmp);
        }
        
        return ((test != null) ? equals(test) : false);
    }
    /**
     * <p>
     * Tests this <tt>JID</tt> for equality to the given
     * <tt>JID</tt>.  This test is"strict" test, meaning that this
     * <tt>JID</tt> must match <tt>cmp</tt> exactly in node, domain,
     * and resource.  The test of node and domain are case-insensitive, while
     * the test for resource is case-sensitive.</p>
     *
     * <p>
     * If <tt>cmp</tt> is <tt>null</tt>, this method returns
     * <tt>false</tt>.</p>
     *
     * @param cmp The <tt>JID</tt> to compare to.
     * @return <tt>true</tt> if this <tt>JID</tt> is strictly
     * equal to <tt>cmp</tt>.
     * @since   JSO-0.2
     */    
    public final boolean equals(JID cmp) {
        boolean     result = cmp != null;

        if (result) {
            //Check domain
            result = getDomain().equalsIgnoreCase(cmp.getDomain());
            if (result) {
                //Check node
                result = getNode().equalsIgnoreCase(cmp.getNode());
            }
            if (result) {
                //Check resource
                result = getResource().equals(cmp.getResource());
            }
        }

        return result;
    }

    /**
     * <p>
     * Compares this <tt>JID</tt> to the given object.  This
     * comparison generates a "natural ordering" for <tt>JID</tt>
     * objects.  This ordering uses the following pattern:
     * <ol>
     * <li>compare domains (case insensitive)</li>
     * <li>compare nodes (case insensitive)</li>
     * <li>compare resources (case sensitive)</li>
     * </ol>
     * </p>
     *
     * <p>
     * This method returns a value less than 0 if this <tt>JID</tt>
     * comes before <tt>cmp</tt>, greater than 0 if this
     * <tt>JID</tt> comes after, or 0 if this <tt>JID</tt> and
     * <tt>cmp</tt> are positionally equal.  Note that this
     * method does not return 0 if the method
     * <tt>{@link #equals(java.lang.Object)}</tt> returns
     * <tt>false</tt>.</p>
     *
     * <p>
     * The value of <tt>cmp</tt> must be convertable to a
     * <tt>JID</tt> (either by casing to a <tt>JID</tt>,
     * or evaluating the results of <tt>toString()</tt>), or
     * an <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param cmp The object to compare to.
     * @return The results of the comparison.
     * @since   JSO-0.2
     */    
    public final int compareTo(Object cmp) {
        return STRICT_COMPARATOR.compare(this, cmp);
    }

    /**
     * <p>
     * Tests this <tt>JID</tt> for "matching" against the given object.
     * The test is a "cascade" equality test, which means this <tt>JID</tt>
     * must match <tt>cmp</tt> in domain, but node only needs to match if
     * this <tt>JID</tt> defines a node, and resource only needs to match
     * if this <tt>JID</tt> defines a resource. The tests for node and domain
     * are case-insensitive, while the test for resource is case-sensitive.</p>
     *
     * <p>
     * If <tt>cmp</tt> is <tt>null</tt>, or is not a <tt>JID</tt>
     * and the results for <tt>toString()</tt> do not result in a valid
     * <tt>JID</tt>, this method returns <tt>false</tt>.</p>
     *
     * @param cmp The Object to compare to.
     * @return <tt>true</tt> if this <tt>JID</tt> is loosely
     * equal to <tt>cmp</tt>.
     * @since   JSO-0.2
     */    
    public final boolean match(Object cmp) {
        JID     test = null;

        if (cmp instanceof JID) {
            test = (JID)cmp;
        } else {
            try {
                test = JID.valueOf(cmp.toString());
            } catch (Exception e) {
                return false;
            }
        }
        return match(test);
    }
    /**
     * <p>
     * Tests this <tt>JID</tt> for "matching" against the given object.
     * The test is a "cascade" equality test, which means this <tt>JID</tt>
     * must match <tt>cmp</tt> in domain, but node only needs to match if
     * this <tt>JID</tt> defines a node, and resource only needs to match
     * if this <tt>JID</tt> defines a resource. The tests for node and domain
     * are case-insensitive, while the test for resource is case-sensitive.</p>
     *
     * <p>
     * If <tt>cmp</tt> is <tt>null</tt>, this method returns
     * <tt>false</tt>.</p>
     *
     * @param cmp The <tt>JID</tt> to compare to.
     * @return <tt>true</tt> if this <tt>JID</tt> is loosely
     * equal to <tt>cmp</tt>.
     * @since   JSO-0.2
     */
    public final boolean match(JID cmp) {
        String      test1, test2;
        boolean     result = (cmp != null);

        if (result) {
            //Check domain
            result = getDomain().equalsIgnoreCase(cmp.getDomain());
        }
        
        if (result && Utilities.isValidString(getNode())) {
            //Check node (if present)
            result = getNode().equalsIgnoreCase(cmp.getNode());
        }

        if (result && Utilities.isValidString(getResource())) {
            //Check resource (if present)
            result = getResource().equals(cmp.getResource());
        }

        return result;
    }

    /**
     * <p>
     * Generates a string representation of this <tt>JID</tt>.
     * This method generates a string in the form "[node@]domain[/resource]",
     * where [node@] and [/resource] are absent if this <tt>JID</tt>
     * does not define them.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The string representing this <tt>JID</tt>.
     * @since   JSO-0.2
     */    
    public final String toString() {
        return toString(false);
    }
    /**
     * <p>
     * Generates a string representation of this <tt>JID</tt> in either
     * "standard" or "normalized" format. This method generates a string in the
     * form "[node@]domain[/resource]", where [node@] and [/resource] are
     * absent this <tt>JID</tt> does not define them.</p>
     *
     * <p>
     * The difference between "standard" and "normalized" representations is
     * with case-sensitivity. The "standard" representation preserves case,
     * while the "normalized" representation uses the lower-case form for
     * {@link #getNode} and {@link #getDomain}, but preserves the case of
     * {@link #getResource}.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The string representing this <tt>JID</tt>
     * @param normalize <tt>true</tt> to generate the "normalized" format
     * for this <tt>JID</tt>.
     * @since   JSO-0.8
     * @deprecated As of JSO-0.12.3, this constructor is replaced by {@link
     * #JID(String)}.
     */
    public final String toString(boolean normalize) {
        String  spec = getSpecification();
        
        if (spec == null) {
            StringBuffer    buffer = new StringBuffer();
            String          node = getNode(), domain = getDomain(), res = getResource();
            
            if (Utilities.isValidString(node))
                buffer.append(node).append('@');
            buffer.append(domain);
            if (Utilities.isValidString(res))
                buffer.append('/').append(res);
            
            setSpecification(spec = buffer.toString());
        }
        
        return spec;
    }
    /**
     * <p>
     * Generates a "bare" JID from this <tt>JID</tt>. A bare JID includes only
     * the node and domain, not the resource.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The "bare" JID for this <tt>JID</tt>
     * @since   JSO-0.10
     */
    public final JID toBareJID() {
        JID jid = new JID();
        jid._Domain = _Domain;
        jid._Node = _Node;
        jid._Resource = "";
        return jid;
    }

    /**
     * <p>
     * Static method to determine if the given object represents a valid
     * <tt>JID</tt>.</p>
     *
     * <p>
     * The value of <tt>spec</tt> may be <tt>null</tt>, in which case this
     * method always returns <tt>false</tt>.</p>
     *
     * @param spec The object to validate.
     * @return <tt>true</tt> if <tt>spec</tt> is an instance of <tt>JID</tt>,
     * or can successfully be converted to a <tt>JID</tt>.
     * @since   JSO-0.12
     */
    public static final boolean isValid(Object spec) {
        boolean     result = (spec != null);
        Matcher     match = (result ? RE_PATTERN.matcher(spec.toString()) : null);
        
        if ((match != null) && match.matches()) {
            result = result && isValidDomain(match.group(POS_DOMAIN));
            result = result && isValidNode(match.group(POS_NODE));
            result = result && isValidResource(match.group(POS_RESOURCE));
        } else {
            result = false;
        }
        
        return result;
    }
    
    /**
     * <p>
     * Static method to generate a <tt>JID</tt> from the given object. This
     * method attempts to use the results of calling <tt>toString()</tt> on
     * <tt>spec</tt>.</p>
     *
     * <p>
     * This method never returns <tt>null</tt>. If <tt>spec</tt> cannot be
     * converted to a <tt>JID</tt>, then a <tt>JIDFormatException</tt>
     * is thrown. If <tt>spec</tt> is an instance of <tt>JID</tt>, then it is
     * returned directly.</p>
     *
     * @param spec The object to convert.
     * @throws IllegalArgumentException If a valid JID cannot be created from
     * <tt>spec</tt>.
     * @return The <tt>JID</tt> represented by <tt>spec</tt>.
     * @throws JIDFormatException If <tt>spec</tt> cannot be parsed into a valid
     * <tt>JID</tt>.
     * @since   JSO-0.8.2
     */    
    public final static JID valueOf(Object spec) throws JIDFormatException {
        return ((spec instanceof JID) ? (JID)spec : new JID((spec != null) ? spec.toString() : ""));
    }
}
