/*
 * Message.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso;


import java.util.Locale;

/**
 * <p>
 * Encapsulates a message element.  The message is used to send or receive
 * data, without need of an immediate response.</p>
 *
 * <p>
 * The message can have a subject, thread, and/or body (along with a possible
 * "error" sub-element). The possible types available to messages are:
 * <ul>
 * <li>{@link #NORMAL}:    A normal message (this is the default
 * type).</li>
 * <li>{@link #CHAT}:      A "one-on-one" chat message.</li>
 * <li>{@link #GROUPCHAT}: A "group chat" message
 * (to/from a "chat" or "conference" room)</li>
 * <li>{@link #HEADLINE}:  A "headline" message, usually
 * containing links to news (as extensions).</li>
 * <li>{@link #ERROR}:     The failed result of sending a
 * message of any other type.</li>
 * </ul></p>
 *
 * @author linuxwolf
 * @version $VERSION$
 * @since JSO-0.2
 */
public interface Message extends Packet {
    //Inner-classes
    /**
     * <p>
     * Enumerates the possible types for <tt>Message</tt> objects.
     * In general, only the "static final" instances of this class should be
     * used.  "Custom" types can be defined by extending this class.</p>
     *
     * @author  linuxwolf
     * @version $VERSION$
     * @since   JSO-0.2
     */
    public static class Type extends Packet.Type {
        //Constructors
        /**
         * <p>
         * Constructs a <tt>Type</tt> for the given a type name.</p>
         *
         * <p>
         * The value of <tt>type</tt> cannot be <tt>null</tt> or "", or
         * an <tt>IllegalArgumentException</tt> is thrown.</p>
         *
         * @param type Name of the type.
         * @since   JSO-0.2
         */
        protected Type(String type) {
            super(type);
        }
    }
    
    //"Constants"
    /**
     * Represents the "normal" type for <tt>Message</tt>.  This is the
     * default type.
     * @since   JSO-0.2
     */
    public static final Type    NORMAL = new Type("normal");
    /** Represents the "chat" type for <tt>Message</tt>.
     * @since   JSO-0.2
     */
    public static final Type    CHAT = new Type("chat");
    /** Represents the "groupchat" type for <tt>Message</tt>.
     * @since   JSO-0.2
     */
    public static final Type    GROUPCHAT   = new Type("groupchat");
    /** Represents the "headline" type for <tt>Message</tt>.
     * @since   JSO-0.2
     */
    public static final Type    HEADLINE    = new Type("headline");
    
    //Methods
    /**
     * <p>
     * Retrieves the thread for this <tt>Message</tt>.  The
     * thread is used to help relate a series of messages.</p>
     *
     * <p>
     * This method returns <tt>null</tt> if the thread has
     * not be set.</p>
     *
     * @return The thread for this <tt>Message</tt>.
     * @since   JSO-0.2
     */    
    public String getThread();
    /**
     * <p>
     * Sets the thread for this <tt>Message</tt>.</p>
     *
     * <p>
     * If the value of <tt>t</tt> is <tt>null</tt>, the
     * thread is removed.</p>
     *
     * @param t The new thread for this <tt>Message</tt>.
     * @since   JSO-0.2
     */    
    public void setThread(String t);
    
    /**
     * <p>
     * Determines if this <tt>Message</tt> includes a subject. This method
     * attempts to find the first &lt;subject/&gt; element for this
     * <tt>Message</tt>'s locale (as specified by {@link
     * StreamNode#getLocale}).</p>
     *
     * <p>
     * This method first uses the specific locale (using exactly what (@link
     * StreamNode#getLocale} returns), then uses a language-only variant.</p>
     *
     * @return <tt>true</tt> if an appropriate subject is available.
     * @since   JSO-0.12
     */
    public boolean hasSubject();
    /**
     * <p>
     * Determines if this <tt>Message</tt> includes a subject for the given
     * locale.</p>
     *
     * <p>
     * If the value of <tt>l</tt> is <tt>null</tt>, the result from {@link
     * StreamNode#getLocale} for this <tt>Message</tt> is used.</p>
     *
     * @param l The locale to search by.
     * @return <tt>true</tt> if an appropriate subject is available.
     * @since   JSO-0.12
     */
    public boolean hasSubject(Locale l);
    /**
     * <p>
     * Retrieves the subject for this <tt>Message</tt>. This method attempts to
     * find the first &lt;subject/&gt; element for this <tt>Message</tt>'s
     * locale (as specified by {@link StreamNode#getLocale}.</p>
     *
     * <p>
     * This method never returns <tt>null</tt>. If an appropriate subject is
     * not available, this method returns "".</p>
     *
     * @return The subject of this <tt>Message</tt>.
     * @since   JSO-0.2
     */    
    public String getSubject();
    /**
     * <p>
     * Retrieves the subject for this <tt>Message</tt> for the given locale.
     * This method attempts to find the first &lt;subject/&gt; element for the
     * given locale.</p>
     *
     * <p>
     * This method never returns <tt>null</tt>. If an appropriate subject is
     * not available, this method returns "".</p>
     *
     * <p>
     * If the value of <tt>l</tt> is <tt>null</tt>, the result from {@link
     * StreamNode#getLocale} for this <tt>Message</tt> is used.</p>
     *
     * @param l The locale to search by.
     * @return The subject of this <tt>Message</tt>.
     * @since   JSO-0.12
     */
    public String getSubject(Locale l);
    /**
     * <p>
     * Sets the subject for this <tt>Message</tt>. This method first removes
     * all &lt;subject/&gt; elements in the <tt>Message</tt>'s locale, then
     * adds a &lt;subject/&gt; without a declared locale.</p>
     *
     * <p>
     * If the value of <tt>s</tt> is <tt>null</tt>, the "locale default"
     * subject is removed.</p>
     *
     * @param s The new subject for this <tt>Message</tt>.
     * @since   JSO-0.2
     */    
    public void setSubject(String s);
    /**
     * <p>
     * Sets the subject for this <tt>Message</tt> for the given locale.</p>
     *
     * <p>
     * If the value of <tt>l</tt> is <tt>null</tt>, the result from {@link
     * StreamNode#getLocale} for this <tt>Message</tt> is used.</p>
     *
     * <p>
     * If the value of <tt>s</tt> is <tt>null</tt>, the "locale default"
     * subject is removed.</p>
     *
     * @param l The locale to search by.
     * @param s The new subject for this <tt>Message</tt>.
     * @since   JSO-0.2
     */    
    public void setSubject(Locale l, String s);

    /**
     * <p>
     * Determines if this <tt>Message</tt> includes a body. This method
     * attempts to find the first &lt;body/&gt; element for this
     * <tt>Message</tt>'s locale (as specified by {@link
     * StreamNode#getLocale}).</p>
     *
     * <p>
     * This method first use the specific locale (using exactly what (@link
     * StreamNode#getLocale} returns), then uses a language-only variant.</p>
     *
     * @return <tt>true</tt> if an appropriate body is available.
     * @since   JSO-0.12
     */
    public boolean hasBody();
    /**
     * <p>
     * Determines if this <tt>Message</tt> includes a body for the given
     * locale.</p>
     *
     * <p>
     * If the value of <tt>l</tt> is <tt>null</tt>, the result from {@link
     * StreamNode#getLocale} for this <tt>Message</tt> is used.</p>
     *
     * @param l The locale to search by.
     * @return <tt>true</tt> if an appropriate body is available.
     * @since   JSO-0.12
     */
    public boolean hasBody(Locale l);
    /**
     * <p>
     * Retrieves the body for this <tt>Message</tt>. This method attempts to
     * find the first &lt;body/&gt; element for this <tt>Message</tt>'s
     * locale (as specified by {@link StreamElement#getLocale}.</p>
     *
     * <p>
     * This method never returns <tt>null</tt>. If an appropriate body is
     * not available, this method returns "".</p>
     *
     * @return The body of this <tt>Message</tt>.
     * @since   JSO-0.2
     */    
    public String getBody();
    /**
     * <p>
     * Retrieves the subject for this <tt>Message</tt> for the given locale.
     * This method attempts to find the first &lt;body/&gt; element for the
     * given locale.</p>
     *
     * <p>
     * This method never returns <tt>null</tt>. If an appropriate body is
     * not available, this method returns "".</p>
     *
     * <p>
     * If the value of <tt>l</tt> is <tt>null</tt>, the result from {@link
     * StreamNode#getLocale} for this <tt>Message</tt> is used.</p>
     *
     * @param l The locale to search by.
     * @return The body of this <tt>Message</tt>.
     * @since   JSO-0.12
     */
    public String getBody(Locale l);
    /**
     * <p>
     * Sets the body for this <tt>Message</tt>. This method first removes
     * all &lt;body/&gt; elements in the <tt>Message</tt>'s locale, then
     * adds a &lt;body/&gt; without a declared locale.</p>
     *
     * <p>
     * If the value of <tt>b</tt> is <tt>null</tt>, the "locale default"
     * body is removed.</p>
     *
     * @param b The new body for this <tt>Message</tt>.
     * @since   JSO-0.2
     */    
    public void setBody(String b);
    /**
     * <p>
     * Sets the body for this <tt>Message</tt> for the given locale.</p>
     *
     * <p>
     * If the value of <tt>l</tt> is <tt>null</tt>, the result from {@link
     * StreamNode#getLocale} for this <tt>Message</tt> is used.</p>
     *
     * <p>
     * If the value of <tt>b</tt> is <tt>null</tt>, the "locale default"
     * body is removed.</p>
     *
     * @param l The locale to search by.
     * @param b The new body for this <tt>Message</tt>.
     * @since   JSO-0.2
     */    
    public void setBody(Locale l, String b);
}