/*
 * PacketEvent.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.event;


import org.jabberstudio.jso.Packet;
import org.jabberstudio.jso.StreamContext;
import org.jabberstudio.jso.util.Enumerator;

/**
 * <p>
 * Event for received or sent <tt>Packet</tt> objects.  Instances of this
 * class are used in conjunction with <tt>PacketListener</tt> to report on
 * transmissions across the stream.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public class PacketEvent extends java.util.EventObject implements Cloneable {
    //Inner-Classes
    /**
     * <P>
     * Enumerates the possible types for <tt>PacketEvent</tt> objects.</p>
     *
     * @author  linuxwolf
     * @version $VERSION$
     * @since  JSO-0.2
     */
    public static final class Type extends Enumerator {
        //Constructors
        private Type(String name) {
            super(name);
        }
    }

    //"Constants"
    /** Represents a "received" <tt>PacketEvent</tt>.
     * @since   JSO-0.2
     */
    public static final Type    RECEIVED = new Type("received");
    /** Represents a "sent" <tt>PacketEvent</tt>.
     * @since   JSO-0.2
     */
    public static final Type    SENT = new Type("sent");
    
    //Fields
    private Packet          _Data;
    private Type            _Type;
    private boolean         _Handled = false;

    //Construstors
    /**
     * <p>
     * Creates a new <tt>PacketEvent</tt> with the given
     * <tt>StreamContext</tt>, <tt>Packet</tt>, and <tt>PacketEvent.Type</tt>.
     * The given <tt>StreamContext</tt> is the source (as returned by
     * <tt>getSource()</tt>), and <tt>data</tt> is the <tt>Packet</tt> object
     * received or sent.</p>
     *
     * @param src The <tt>StreamContext</tt> event source.
     * @param data The <tt>Packet</tt> data received/sent.
     * @param t The <tt>PacketEvent.Type</tt>.
     * @since   JSO-0.10
     */
    public PacketEvent(StreamContext src, Packet data, Type t) {
        super(src);
        _Data = data;
        _Type = t;
    }
    /**
     * <p>
     * Creates a new <tt>PacketEvent</tt> that is a duplicate of the given
     * <tt>PacketEvent</tt>. The references that <tt>og</tt> contains are
     * referenced by this new <tt>PacketEvent</tt>, not copied.</p>
     *
     * @param og The <tt>PacketEvent</tt> to copy from.
     * @since   JSO-0.8
     */
    protected PacketEvent(PacketEvent og) {
        this (og.getContext(), og.getData(), og.getType());
    }

    //Methods
    /**
     * <p>
     * Retreives the <tt>StreamContext</tt> for this <tt>PacketEvent</tt>.
     *  This is a convenience method over <tt>getSource()</tt>, automatically
     * performing the cast to <tt>StreamContext</tt>.</p>
     *
     * @return The source <tt>StreamContext</tt>.
     * @since   JSO-0.10
     */
    public StreamContext getContext() {
        return (StreamContext)getSource();
    }
    
    /**
     * <p>
     * Retrieves the <tt>Type</tt> for this <tt>PacketEvent</tt>.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The type for this <tt>PacketEvent</tt>.
     * @since   JSO-0.2
     */
    public Type getType() {
        return _Type;
    }
    /**
     * <p>
     * Retrieves the <tt>Packet</tt> that was received or sent.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The <tt>Packet</tt> data received or sent.
     * @since   JSO-0.2
     */    
    public Packet getData() {
        return _Data;
    }
    
    /**
     * <p>
     * Determines if this <tt>PacketEvent</tt> has been handled. This
     * method, along with {@link #setHandled}, is used for flagging events that
     * have been addressed by an application. A common usage is for the last
     * <tt>PacketListener</tt> to send a 501 (NOT IMPLEMENTED) error for
     * any packets that have not been handled.</p>
     *
     * @return <tt>true</tt> if this <tt>
     * @since   JSO-0.8
     */
    public boolean isHandled() {
        return _Handled;
    }
    /**
     * <p>
     * Sets the "handled" flag for this <tt>PacketEvent</tt>. This method
     * directly impacts the value returned by {@link #isHandled}.</p>
     *
     * @param flag The new "handled" value.
     * @since   JSO-0.8
     */
    public void setHandled(boolean flag) {
        _Handled = flag;
    }
    
    /**
     * <p>
     * Clones this <tt>PacketEvent</tt> object. This operation performs a
     * "shallow copy" of the <tt>PacketEvent</tt>. It also retains the
     * current values for {@link #getContext}, {@link #getType}, {@link
     * #getData}, and {@link #isHandled}.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The clone of this <tt>PacketEvent</tt>.
     * @since   JSO-0.8
     */
    public Object clone() {
        return new PacketEvent(this);
    }
}