/*
 * XPathMonitorEvent.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.event;

import org.jabberstudio.jso.Packet;
import org.jabberstudio.jso.StreamException;

import org.jaxen.XPath;

import org.jabberstudio.jso.util.XPathMonitor;

/**
 * <p>
 * Class for monitoring events specific to <tt>XPathMonitor</tt>.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public class XPathMonitorEvent extends MonitorEvent {
    //Fields
    private XPath       _XPath;
    
    //Constructors
    /**
     * <p>
     * Creates a new <tt>XPathMonitorEvent</tt> with the given source,
     * evaluated XPath, actual <tt>Packet</tt>, reason, and cause.</p>
     *
     * <p>
     * This constructor does not attempt to validate any of the arguments.</p>
     *
     * @param src The source of the event.
     * @param xpath The <tt>XPath</tt> evaluated against.
     * @param actual The <tt>Packet</tt> actually received.
     * @param reason The reason for the event.
     * @param cause The cause of a failed event.
     * @since   JSO-0.12
     */
    protected XPathMonitorEvent(XPathMonitor src, XPath xpath, Packet actual, MonitorEvent.Reason reason, StreamException cause) {
        super(src, actual, reason, cause);
        
        _XPath = xpath;
    }

    //Methods
    /**
     * <p>
     * Retrieves the <tt>XPathMonitor</tt> triggering this
     * <tt>XPathMonitorEvent</tt>. This method is a convenience over
     * <tt>getSource()</tt> to automatically cast the result to a
     * <tt>XPathMonitor</tt>.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The <tt>XPathMonitor</tt> source of this
     * <tt>XPathMonitorEvent</tt>.
     * @since   JSO-0.12
     */
    public XPathMonitor getXPathMonitor() {
        return (XPathMonitor)getSource();
    }
    /**
     * <p>
     * Retrieves the <tt>XPath</tt> used by the <tt>XPathMonitor</tt> for
     * evaluation.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The <tt>XPath</tt> evaluated.
     * @since   JSO-0.12
     */
    public XPath getXPath() {
        return _XPath;
    }
    
    /**
     * <p>
     * Creates a new <tt>XPathMonitorEvent</tt> for the {@link #FOUND}
     * trigger.</p>
     *
     * <p>
     * The value of <tt>mon</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>xpath</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>actual</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param mon The source <tt>XPathMonitor</tt> of the event.
     * @param xpath The <tt>XPath</tt> evaluated against.
     * @param actual The <tt>Packet</tt> actually received.
     * @return The <tt>XPathMonitorEvent</tt>.
     * @throws IllegalArgumentException If any paramters are invalid.
     * @since   JSO-0.12
     */
    public static XPathMonitorEvent createFoundEvent(XPathMonitor mon, XPath xpath, Packet actual) {
        //Validate input
        if (mon == null)
            throw new IllegalArgumentException("XPathMonitor cannot be null");
        if (xpath == null)
            throw new IllegalArgumentException("XPath cannot be null");
        if (actual == null)
            throw new IllegalArgumentException("Actual Packet cannot be null");
        
        return new XPathMonitorEvent(mon, xpath, actual, FOUND, null);
    }
    /**
     * <p>
     * Creates a new <tt>XPathMonitorEvent</tt> for the {@link #TIMEOUT}
     * trigger.</p>
     *
     * <p>
     * The value of <tt>mon</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>expect</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param mon The source <tt>XPathMonitor</tt> of the event.
     * @return The <tt>XPathMonitorEvent</tt>.
     * @throws IllegalArgumentException If any paramters are invalid.
     * @since   JSO-0.12
     */
    public static XPathMonitorEvent createTimeoutEvent(XPathMonitor mon, XPath xpath) {
        //Validate input
        if (mon == null)
            throw new IllegalArgumentException("XPathMonitor cannot be null");
        if (xpath == null)
            throw new IllegalArgumentException("XPath cannot be null");
        
        return new XPathMonitorEvent(mon, xpath, null, TIMEOUT, null);
    }
    /**
     * <p>
     * Creates a new <tt>XPathMonitorEvent</tt> for the {@link #FAILED}
     * trigger.</p>
     *
     * <p>
     * The value of <tt>mon</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>expect</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>cause</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param mon The source <tt>XPathMonitor</tt> of the event.
     * @param cause The cause of a failed event.
     * @return The <tt>XPathMonitorEvent</tt>.
     * @throws IllegalArgumentException If any paramters are invalid.
     * @since   JSO-0.12
     */
    public static XPathMonitorEvent createFailedEvent(XPathMonitor mon, XPath xpath, StreamException cause) {
        //Validate input
        if (mon == null)
            throw new IllegalArgumentException("XPathMonitor cannot be null");
        if (xpath == null)
            throw new IllegalArgumentException("XPath cannot be null");
        if (cause == null)
            throw new IllegalArgumentException("Cause cannot be null");
        
        return new XPathMonitorEvent(mon, xpath, null, FAILED, cause);
    }
}