/*
 * ContinuingFeatureConsumer.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.features;


import org.jabberstudio.jso.StreamFeature;
import org.jabberstudio.jso.StreamFeatureset;

/**
 * <p>
 * Abstract class for implementing <tt>FeatureConsumer</tt>s. This
 * class assumes the <tt>FeatureConsumer</tt> requires additional protocol
 * exchanges before completing.</p>
 *
 * <p>
 * Subclasses must implement {@link #findFeature} to identify the
 * <tt>StreamFeature</tt> to be consumed, and {@link #startConsumption} to
 * begin the process (which includes registering any necessary
 * <tt>PacketListener</tt>s). They may also override {@link
 * #completeFailedConsumption} and/or {@link #completeSucceededConsumption} to
 * perform specific cleanup routines.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public abstract class AbstractFeatureConsumer implements FeatureConsumer {
    //Fields
    private FeatureContext      _Ctx;
    
    //Methods
    /**
     * <p>
     * Retrieves the active <tt>FeatureContext</tt> for this
     * <tt>AbstractFeatureConsumer</tt>.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if it has not
     * been initialized.</p>
     *
     * @return The active <tt>FeatureContext</tt>.
     * @since   JSO-0.12
     */
    protected FeatureContext getFeatureContext() {
        return _Ctx;
    }
    /**
     * <p>
     * Sets the active <tt>FeatureContext</tt> for this
     * <tt>AbstractFeatureConsumer</tt>.</p>
     *
     * <p>
     * The value of <tt>ctx</tt> may be <tt>null</tt>, in which case any
     * previous value is removed.</p>
     *
     * @param ctx The new <tt>FeatureContext</tt>.
     * @since   JSO-0.12
     */
    protected void setFeatureContext(FeatureContext ctx) {
        _Ctx = ctx;
    }
    
    /**
     * <p>
     * Consumes the given <tt>FeatureContext</tt>. This implementation stores
     * <tt>ctx</tt>, then calls {@link #startConsumption}. If anything is
     * thrown by that method, it is caught and {@link
     * #completeFailedConsumption} is called.</p>
     *
     * <p>
     * The value of <tt>ctx</tt> will never be <tt>null</tt>.</p>
     *
     * @param ctx The <tt>FeatureContext</tt> to consume.
     * @throws Exception if there is any error consuming.
     * @since   JSO-0.12
     */
    public void consume(FeatureContext ctx) throws Exception {
        setFeatureContext(ctx);
        try {
            startConsumption();
        } catch (Throwable t) {
            completeFailedConsumption(t);
        }
    }
    /**
     * <p>
     * Resets this <tt>FeatureConsumer</tt>. This implementation clears any
     * active <tt>FeatureContext</tt>.</p>
     *
     * @since   JSO-0.12
     */
    public void reset() {
        setFeatureContext(null);
    }

    /**
     * <p>
     * Called by {@link #consume} to start the consumption for this
     * <tt>AbstractFeatureConsumer</tt>..</p>
     *
     * @throws Throwable If any failures occur starting consumption.
     * @since   JSO-0.12
     */
    protected abstract void startConsumption() throws Throwable;
    /**
     * <p>
     * Completes consumption in a successful manner. This implementation
     * retrieves the active <tt>FeatureContext</tt>, calls {@link
     * FeatureContext#completeSucceeded} with the value of <tt>restart</tt>,
     * then removes the <tt>FeatureContext</tt>.</p>
     *
     * <p>
     * Subclasses may override this method to perform any cleanup activities
     * necessary, but should call
     * <tt>super.completeSucceededConsumption(restart)</tt> at some point in
     * their override.</p>
     *
     * @param restart <tt>true</tt> if the underlying <tt>Stream</tt> needs to
     * be dropped and reopened.
     * @throws IllegalStateException If {@link #getFeatureContext} returns
     * <tt>null</tt>.
     * @since   JSO-0.12
     */
    protected void completeSucceededConsumption(boolean restart) throws IllegalStateException {
        FeatureContext      ctx = getFeatureContext();
        
        if (ctx == null)
            throw new IllegalStateException("FeatureContext has not been initialized");
        
        ctx.completeSucceeded(restart);
        setFeatureContext(null);
    }
    /**
     * <p>
     * Completes consumption in a failed manner. This implementation retrieves
     * the active <tt>FeatureContext</tt>, calls {@link
     * FeatureContext#completeFailed} with the value of <tt>t</tt>, then
     * removes the <tt>FeatureContext</tt>.</p>
     *
     * <p>
     * Subclasses may override this method to perform any cleanup activities
     * necessary, but should call
     * <tt>super.completeFailedConsumption(t)</tt> at some point in
     * their override.</p>
     *
     * @param t The cause of the failure.
     * @throws IllegalStateException If {@link #getFeatureContext} returns
     * <tt>null</tt>.
     * @since   JSO-0.12
     */
    protected void completeFailedConsumption(Throwable t) {
        FeatureContext      ctx = getFeatureContext();
        
        if (ctx == null)
            throw new IllegalStateException("FeatureContext has not been initialized");

        ctx.completeFailed(t);
        setFeatureContext(null);
    }
}