/*
 * JIDFormat.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.format;

import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.text.FieldPosition;
import java.text.Format;
import java.text.ParseException;
import java.text.ParsePosition;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.jabberstudio.jso.JID;
import org.jabberstudio.jso.util.Utilities;

/**
 * <p>
 * Formats and parses <tt>JID</tt> objects to and from the escaped/unescaped
 * format. This class supports JID-Escaping, as defined in JEP-0106.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.14
 * @see     <a href='http://www.jabber.org/jeps/jep-0106.html'>JEP-0106: JID Escaping</a>
 */
public class JIDFormat extends Format {
    private static class Field extends Format.Field {
        //Constructors
        public Field(String name) {
            super(name);
        }
    }
    
    //"Constants"
    /** Constant for "node" field alignment. Used in <tt>FieldPosition</tt> in
     * formatting.</p>
     *
     * @since   JSO-0.14
     */
    public static final Field       NODE_FIELD      = new Field("node");
    /** Constant for "domain" field alignment. Used in <tt>FieldPosition</tt>
     * in formatting.</p>
     *
     * @since   JSO-0.14
     */
    public static final Field       DOMAIN_FIELD    = new Field("domain");
    /** Constant for "resource" field alignment. Used in <tt>FieldPosition</tt>
     * in formatting.</p>
     *
     * @since   JSO-0.14
     */
    public static final Field       RESOURCE_FIELD  = new Field("resource");
    
    private static final Pattern    RE_PATTERN      = Pattern.compile("(([^@]+)@)?([^/]+)(/(.*))?", Pattern.DOTALL);
    private static final Pattern    RE_JIDENCODE    = Pattern.compile("( )|(\")|(&)|(\')|(/)|(\\:)|(<)|(>)|(@)|(\\\\)", Pattern.DOTALL);
    private static final Pattern    RE_JIDDECODE    = Pattern.compile("(\\\\20)|(\\\\22)|(\\\\26)|(\\\\27)|(\\\\2f)|(\\\\3a)|(\\\\3c)|(\\\\3e)|(\\\\40)|(\\\\5c)", Pattern.DOTALL);
    
    private static final int        POS_NODE        = 2;
    private static final int        POS_DOMAIN      = 3;
    private static final int        POS_RESOURCE    = 5;
    
    private boolean     _Esc;
    
    private static final JIDFormat[] _Inst = new JIDFormat[] {
        new JIDFormat(false), new JIDFormat(true)
    };
    
    //Constructors
    private JIDFormat(boolean esc) {
        _Esc = esc;
    }
    
    //Methods
    /**
     * <p>
     * Determines if formatting should be escaped.</p>
     *
     * @return <tt>true</tt> if nodeprep-reserved characters are escaped.
     * @since   JSO-0.14
     */
    public final boolean isEscaped() {
        return _Esc;
    }
    
    /**
     * <p>
     * Formats the given <tt>JID</tt> into a string. If {@link #isEscaped}
     * returns <tt>false</tt>, any escaped values in the node are unescaped.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt> or "".</p>
     *
     * <p>
     * The value of <tt>j</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param j The JID to format.
     * @return The string.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.14
     */
    public String format(JID j) throws IllegalArgumentException {
        return format((JID)j);
    }
    /**
     * <p>
     * Formats the object as a string. If {@link #isEscaped} returns
     * <tt>false</tt>, any escaped values in the node are unescaped.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * <p>
     * The value of <tt>obj</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown. Also, if <tt>obj</tt> is
     * not an instance of <tt>JID</tt>, an <tt>IllegalArgumentException</tt>
     * is thrown.</p>
     *
     * <p>
     * The value of <tt>toAppendTo</tt> cannot be <tt>null</tt>, or a
     * <tt>NullPointerException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>pos</tt> may be <tt>null</tt>, in which case it is
     * ignored.</p>
     *
     * @param obj The object to format.
     * @param toAppendTo The string buffer to append to.
     * @param pos A <tt>FieldPosition</tt> identifying a field in the formatted
     * text.
     * @return The string buffer passed in as <tt>toAppendTo</tt>, with the
     * formatted text appended.
     * @throws NullPointerException If <tt>toAppendTo</tt> is <tt>null</tt>.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.14
     */
    public StringBuffer format(Object obj, StringBuffer toAppendTo, FieldPosition pos) {
        AttributedCharacterIterator itr = formatToCharacterIterator(obj, toAppendTo);
        
        if (pos != null) {
            if (pos.getFieldAttribute() == NODE_FIELD) {
                pos.setBeginIndex(itr.getRunStart(NODE_FIELD));
                pos.setEndIndex(itr.getRunLimit(NODE_FIELD));
            }
            itr.setIndex(itr.getRunStart(NODE_FIELD));
            
            if (pos.getFieldAttribute() == DOMAIN_FIELD) {
                pos.setBeginIndex(itr.getRunStart(DOMAIN_FIELD));
                pos.setEndIndex(itr.getRunLimit(DOMAIN_FIELD));
            }
            itr.setIndex(itr.getRunStart(DOMAIN_FIELD));
            
            if (pos.getFieldAttribute() == RESOURCE_FIELD) {
                pos.setBeginIndex(itr.getRunStart(RESOURCE_FIELD));
                pos.setEndIndex(itr.getRunLimit(RESOURCE_FIELD));
            }
            itr.setIndex(itr.getRunStart(RESOURCE_FIELD));
        }

        return toAppendTo;
    }
    
    /**
     * <p>
     * Parses text from a JID string into a <tt>JID</tt>.</p>
     *
     * <p>
     * If parsing fails, then a <tt>ParseException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>source</tt> may be <tt>null</tt>, in which case parsing
     * always fails.</p>
     *
     * @param source The string to convert.
     * @return The JID parsed from <tt>source</tt>.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.14
     */
    public JID parse(String source) throws ParseException {
        return (JID)parseObject(source);
    }
    /**
     * <p>
     * Parses text from a string into a <tt>JID</tt>. <b>NOTE:</b> this method
     * always escapes the node portion (if present).</p>
     *
     * <p>
     * If parsing fails, then the error index of <tt>pos</tt> is set to the
     * initial parsing index, and <tt>null</tt> is returned.</p>
     *
     * <p>
     * The value of <tt>source</tt> may be <tt>null</tt>, in which case parsing
     * always fails.</p>
     *
     * <p>
     * The value of <tt>pos</tt> cannot be <tt>null</tt>, or a
     * <tt>NullPointerException</tt> is thrown.</p>
     *
     * @param source The string to convert.
     * @param pos A <tt>ParsePosition</tt> object with the index and error
     * index information.
     * @return The Locale parsed from <tt>source</tt>.
     * @throws NullPointerException If <tt>pos</tt> is <tt>null</tt>.
     * @since   JSO-0.12
     */
    public Object parseObject(String source, ParsePosition pos) throws NullPointerException {
        JID     dst = null;
        Matcher m;
        
        try {
            source = source.substring(pos.getIndex());
            m = RE_PATTERN.matcher(source);
            if (m.matches()) {
                String      node = m.group(POS_NODE);
                String      domain = m.group(POS_DOMAIN);
                String      resource = m.group(POS_RESOURCE);

                if (Utilities.isValidString(node))
                    node = escapeNode(node);

                dst = new JID(node, domain, resource);
                pos.setIndex(m.end() + pos.getIndex());
            } else {
                throw new IllegalArgumentException("string is not a valid JID");
            }
        } catch (Exception e) {
            pos.setErrorIndex(pos.getIndex());
        }
        
        return dst;
    }

    /**
     * <p>
     * Formats a JID into an <tt>AttributedStringIterator</tt> as a JID string.
     * The returned iterator can be used to build the resulting string, as well
     * as to determine information about the resulting string.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each attribute
     * in the returned iterator is an instance of type
     * <tt>Format.Field</tt>. The order to examine the attributes is:</p>
     * <ol>
     * <li>{@link #NODE_FIELD}</li>
     * <li>{@link #DOMAIN_FIELD}</li>
     * <li>{@link #RESOURCE_FIELD}</li>
     * </ol>
     *
     * <p>
     * The value of <tt>obj</tt> must be a non-<tt>null</tt> instance of
     * <tt>JID</tt>, or an <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param obj The JID to format.
     * @return The <tt>AttributedCharacterIterator</tt> describing the resulting
     * string.
     * @throws IllegalArgumentException If the paramter is invalid.
     * @since   JSO-0.14
     */
    public AttributedCharacterIterator formatToCharacterIterator(Object obj) throws IllegalArgumentException {
        return formatToCharacterIterator(obj, new StringBuffer());
    }
    private AttributedCharacterIterator formatToCharacterIterator(Object obj, StringBuffer toAppendTo) throws IllegalArgumentException {
        Map                 attrMap;
        AttributedString    attrStr;
        StringBuffer        buffer = new StringBuffer();
        JID                 j;
        String              n, d, r;
        int                 start = toAppendTo.length(), finish;
        int[]               begin = new int[3], end = new int[3];

        //Validate
        if (obj == null)
            throw new IllegalArgumentException("object cannot be null");
        if (!(obj instanceof JID))
            throw new IllegalArgumentException("object must be an instance of JID");
        j = (JID)obj;
        attrMap = new java.util.HashMap();

        n = j.getNode();
        d = j.getDomain();
        r = j.getResource();

        //append node
        if (Utilities.isValidString(n)) {
            if (!isEscaped())
                n = unescapeNode(n);
            begin[0] = start + buffer.length();
            buffer.append(n).append("@");
            end[0] = start + buffer.length();
            attrMap.put(NODE_FIELD, n);
        }
        
        //append domain
        begin[1] = start + buffer.length();
        buffer.append(d);
        end[1] = start + buffer.length();
        attrMap.put(DOMAIN_FIELD, d);
        
        //append resource
        if (Utilities.isValidString(r)) {
            begin[0] = start + buffer.length();
            buffer.append("/").append(r);
            end[0] = start + buffer.length();
            attrMap.put(RESOURCE_FIELD, r);
        }
        
        finish = start + buffer.length();
        toAppendTo.append(buffer);
        
        attrStr = new AttributedString(buffer.toString(), attrMap);
        if (Utilities.isValidString(n))
            attrStr.addAttribute(NODE_FIELD, n, begin[0], end[0]);
        if (Utilities.isValidString(d))
            attrStr.addAttribute(DOMAIN_FIELD, d, begin[1], end[1]);
        if (Utilities.isValidString(r))
            attrStr.addAttribute(RESOURCE_FIELD, r, begin[1], end[1]);

        return attrStr.getIterator();
    }
    
    /**
     * <p>
     * Retrieves the default <tt>JIDFormat</tt> instance. This method is
     * equivalent to calling {@link #getInstance} with <tt>false</tt>.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The default <tt>JIDFormat</tt>.
     */
    public static JIDFormat getInstance() {
        return getInstance(false);
    }
    /**
     * <p>
     * Retrieves a <tt>JIDFormat</tt> instance.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @param esc <tt>true</tt> if formatted results are escaped.
     * @return The <tt>JIDFormat</tt>.
     * @since   JSO-0.14
     */
    public static JIDFormat getInstance(boolean esc) {
        return _Inst[esc ? 1 : 0];
    }
    /**
     * <p>
     * Generates an escaped form of the given node.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * <p>
     * If <tt>input</tt> is "" or <tt>null</tt>, "" is returned.</p>
     *
     * @param input The node to escape.
     * @return the escaped node.
     * @since   JSO-0.14
     * @see     <a href='http://www.jabber.org/jeps/jep-0106.html'>JEP-0106: JID Escaping</a>
     */
    public static final String escapeNode(String input) {
        StringBuffer    output = new StringBuffer();
        Matcher         m;
        String          grp;

        if (Utilities.isValidString(input)) {
            m = RE_JIDENCODE.matcher(input);
            
            while (m.find()) {
                grp = m.group();
                
                if      (grp.equals(" "))
                    grp = "\\20";
                else if (grp.equals("\""))
                    grp = "\\22";
                else if (grp.equals("&"))
                    grp = "\\26";
                else if (grp.equals("\'"))
                    grp = "\\27";
                else if (grp.equals("/"))
                    grp = "\\2f";
                else if (grp.equals(":"))
                    grp = "\\3a";
                else if (grp.equals("<"))
                    grp = "\\3c";
                else if (grp.equals(">"))
                    grp = "\\3e";
                else if (grp.equals("@"))
                    grp = "\\40";
                else if (grp.equals("\\"))
                    grp = "\\5c";
                
                m.appendReplacement(output, "");
                output.append(grp);
            }
            
            m.appendTail(output);
        }
        
        return output.toString();
    }
    
    /**
     * <p>
     * Generates an unescaped form of the given node.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * <p>
     * If <tt>input</tt> is "" or <tt>null</tt>, "" is returned.</p>
     *
     * @param input The node to unescape.
     * @return the unescaped node.
     * @since   JSO-0.14
     * @see     <a href='http://www.jabber.org/jeps/jep-0106.html'>JEP-0106: JID Escaping</a>
     */
    public static final String unescapeNode(String input) {
        StringBuffer    output = new StringBuffer();
        Matcher         m;
        String          grp;
        
        if (Utilities.isValidString(input)) {
            m = RE_JIDDECODE.matcher(input);
            
            while (m.find()) {
                grp = m.group();
                
                if      (grp.equals("\\20"))
                    grp = " ";
                else if (grp.equals("\\22"))
                    grp = "\"";
                else if (grp.equals("\\26"))
                    grp = "&";
                else if (grp.equals("\\27"))
                    grp = "\'";
                else if (grp.equals("\\2f"))
                    grp = "/";
                else if (grp.equals("\\3a"))
                    grp = ":";
                else if (grp.equals("\\3c"))
                    grp = "<";
                else if (grp.equals("\\3e"))
                    grp = ">";
                else if (grp.equals("\\40"))
                    grp = "@";
                else if (grp.equals("\\5c"))
                    grp = "\\";
                
                m.appendReplacement(output, "");
                output.append(grp);
            }
            
            m.appendTail(output);
        }
        
        return output.toString();
    }
}