/*
 * Importer.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.io;

import org.jabberstudio.jso.StreamContext;
import org.jabberstudio.jso.StreamElement;

/**
 * <p>
 * General-purpose interface for generating imports of <tt>StreamElement</tt>
 * objects.</p>
 *
 * <p>
 * This interface represents a "service". Implementations of this interface
 * can be automatically registered at class-loading time via the SPI
 * process, using the service name "org.jabberstudio.jso.io.Importer".</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.10
 * @see <a href='http://java.sun.com/j2se/1.4.2/docs/guide/jar/jar.html#Service%20Provider'>JAR File Specification :: Service Provider</a>
 */
public interface Importer {
    //Methods
    /**
     * <p>
     * Retrieves the <tt>StreamElement</tt> for this <tt>Importer</tt>. It is
     * used as a "fallback" for namespace information.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The <tt>StreamElement</tt>.
     * @since   JSO-0.10
     */
    public StreamElement getContext();
    
    /**
     * <p>
     * Configures this <tt>Importer</tt> with the given <tt>StreamElement</tt>.
     * This method is called by {@link
     * org.jabberstudio.jso.JSOImplementation#createImporter}.</p>
     *
     * <p>
     * This method can only be called once. Subsequent calls will throw an
     * <tt>IllegalStateException</tt>.</p>
     *
     * <p>
     * The value of <tt>ctx</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param ctx The <tt>StreamElement</tt> to configure to.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @throws IllegalStateException If this <tt>Exporter</tt> is already
     * configured.
     * @throws RuntimeException If there are any other errors while configuring
     * this <tt>Importer</tt>.
     * @since   JSO-0.10
     */
    public void configure(StreamElement ctx) throws IllegalArgumentException, IllegalStateException, RuntimeException;
    /**
     * <p>
     * Executes the import for the given <tt>Object</tt>. The import
     * uses the <tt>StreamContext</tt> from {@link #getContext} for any
     * unknown namespace declarations. The specific class of object provided
     * is implementation-dependent.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @param in The <tt>Object</tt> to generate from.
     * @return The <tt>StreamElement</tt> imported from <tt>in</tt>
     * @throws RuntimeException If any errors occur while executing.
     * @since   JSO-0.10
     */
    public StreamElement execute(Object in) throws RuntimeException;
}