/*
 * StreamBuilder.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.io;

import org.jabberstudio.jso.StreamNode;

/**
 * <p>
 * Builds <code>StreamNode</code> objects using a SAX-like interface.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public interface StreamBuilder {
    //Inner-classes
    /**
     * <p>
     * Enumerates the current state of this <code>StreamBuilder</code>.</p>
     *
     * @author  linuxwolf
     * @version $VERSION$
     * @since   JSO-0.2
     */
    public static final class Status {
        private final String    _Name;
        
        //Constructors
        private Status(String name) {
            _Name = name;
        }
        
        //Methods
        /**
         * <p>
         * Retrieves a string representation of this <Code>Status</code>.</p>
         *
         * <p>
         * The value returned by this method is never <code>null</code> or
         * "".</p>
         *
         * @return The string representing this <code>Status</code>.
         * @since   JSO-0.2
         */
        public String toString() {
            return _Name;
        }
        
        /**
         * <p>
         * Compares this <Code>Status</code> to the given object for
         * equality.  This method compares the object references to determine
         * if both reference the same object.</p>
         *
         * @param cmp The object to compare to.
         * @return <code>true</code> if the references are to the same object.
         * @since   JSO-0.2
         */
        public boolean equals(Object cmp) {
            return (this == cmp);
        }
    }
    //"Constants"
    /** Flags that the <code>StreamBuilder</code> is in an invalid state.
     * @since   JSO-0.2
     */
    public static final Status INVALID = new Status("invalid");
    /** Flags that the <code>StreamBuilder</code> is ready to process.
     * @since   JSO-0.2
     */
    public static final Status READY = new Status("ready");
    /** Flags that the <code>StreamBuilder</code> is still processing.
     * @since   JSO-0.2
     */
    public static final Status PROCESSING = new Status("processing");
    /** Flags that the <code>StreamBuilder</code> has extended data available.
     * @since   JSO-0.2
     */
    public static final Status EXTENDED_DATA = new Status("extended data");
    /** Flags that the <code>StreamBuilder</code> is complete.
     * @since   JSO-0.2
     */
    public static final Status COMPLETE = new Status("complete");

    //Methods
    /**
     * <p>
     * Retrieves the current status of this <code>StreamBuilder</code>.  The
     * status is ued to determine when certain methods (such as <code>{@link
     * #getNode}</code>, <code>{@link #buildExtendedData}</code>, and
     * <code>{@link #addExtendedData}</code>) can successfully be called.</p>
     *
     * @return  The current status for this <code>StreamBuilder</code>.
     * @since   JSO-0.2
     */   
    public Status getStatus();

    /**
     * <p>
     * Retrieves the <code>StreamNode</code> this <code>StreamBuilder</code> has
     * been constructing.</p>
     *
     * <p>
     * This method never returns <code>null</code>.</p>
     *
     * @return The new <code>StreamNode</code> that was built.
     * @throws IllegalStateException If the current state is not <code>{@link
     * #COMPLETE}</code>.
     * @since   JSO-0.2
     */
    public StreamNode getNode() throws IllegalStateException;

    /**
     * <p>
     * Resets this <code>StreamBuilder</code>.  By resetting, the status is
     * set to <code>{@link #READY}</code>, and all internal data for building
     * an object is either deleted or set to default values.</p>
     * @since   JSO-0.2
     */
    public void reset();

    /**
     * <p>
     * Retrieves the <code>StreamBuilder</code> from which to create extended
     * data.</p>
     *
     * <p>
     * This method never returns <Code>null</Code>.</p>
     *
     * @return The <code>StreamBuilder</code> for the extended data.
     * @throws IllegalStateException If the current status is not <code>{@link
     * #EXTENDED_DATA}</code>.
     * @since   JSO-0.2
     */
    public StreamBuilder buildExtendedData() throws IllegalStateException;
    /**
     * <p>
     * The the given <code>StreamNode</code> as extended data.</p>
     *
     *
     * @param node The extended data to add.
     * @throws IllegalStateException If the current status is not <code>{@link
     * #EXTENDED_DATA}</code>.
     * @since   JSO-0.2
     */
    public void addExtendedData(StreamNode node) throws IllegalStateException;

    /**
     * <p>
     * Processes "start-tag" data.</p>
     *
     * <p>
     * The value of <code>s</code> is never <code>null</code>.</p>
     *
     * @param s The current <code>StreamParseState</code>.
     * @return The new status for this <Code>StreamBuilder</code>.
     * @since   JSO-0.2
     */
    public Status processElementStart(StreamParseState s);
    /**
     * <p>
     * Process "end-tag" data.</p>
     *
     * <p>
     * The value of <code>s</code> is never <code>null</code>.</p>
     *
     * @param s The current <code>StreamParseState</code>.
     * @return The new status for thsi <code>StreamBuilder</code>.
     * @since   JSO-0.2
     */
    public Status processElementEnd(StreamParseState s);
    /**
     * <p>
     * Processes text data.</p>
     *
     * <p>
     * The value of <code>s</code> is never <code>null</code>.</p>
     *
     * @param s The current <code>StreamParseState</code>.
     * @return The new status for this <code>StreamBuilder</code>.
     * @since   JSO-0.2
     */
    public Status processText(StreamParseState s);
}