/*
 * StreamParseState.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.io;


import java.util.Collection;
import java.util.Set;

import org.jabberstudio.jso.NSI;
import org.jabberstudio.jso.StreamContext;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.util.Enumerator;

/**
 * <p>
 * Encapsulates a state in the processing on incoming XML for a
 * <tt>Stream</tt>.  This provides a similar interface as used by
 * XML "pull-parsers", but only for data retrieval.</p>
 *
 * <p>
 * This is the only interface for describing the current parse state.  As
 * such, certain methods are only valid for certain state "types" (as
 * defined by {@link Type}.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public interface StreamParseState {
    //Inner-classes
    /**
     * <p>
     * Class to enumerate the possible types for a <tt>StreamState</tt>.
     * Instances of this class only exist as "constant" field members of the
     * parent <tt>StreamState</tt> interface.</p>
     *
     * @author  linuxwolf
     * @version $VERSION$
     * @since   JSO-0.2
     */
    public static final class Type extends Enumerator {
        //Constructors
        private Type(String name) {
            super(name);
        }
    }
    
    //"Constants"
    /** Flag that this <tt>StreamParseState</tt> is for character data.
     * @since   JSO-0.2
     */
    public static final Type    TEXT = new Type("text");
    /** Flag that this <tt>StreamParseState</tt> is for the start of an
     * element.
     * @since   JSO-0.2
     */
    public static final Type    ELEMENT_START = new Type("element start");
    /** Flag that this <tt>StreamParseState</tt> is for the end of an
     * element.
     * @since   JSO-0.2
     */
    public static final Type    ELEMENT_END = new Type("element end");
    
    //Methods
    /**
     * <p>
     * Retrieves the <tt>Type</tt> for this <tt>StreamState</tt>.
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The <tt>Type</tt> for this <tt>StreamState</tt>.
     * @since   JSO-0.2
     */
    public Type getType();

    /**
     * <p>
     * Retrieves the current depth for this <tt>StreamState</tt>.  The
     * depth is based on the current "element tree", using the root tag
     * (the &lt;stream:stream&gt;) for position 1.  Matching start and end tags
     * have the same depth.</p>
     *
     * <p>
     * The value returned by this method is never less than 0.</p>
     *
     * @return The current tag depth.
     * @since   JSO-0.2
     */
    public int getDepth();
    
    /**
     * <p>
     * Retrieves the context <tt>StreamElement</tt> for this
     * <tt>StreamParseState</tt>.  The <tt>StreamElement</tt> provides such
     * things as the appropriate <tt>StreamDataFactory</tt>.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The <tt>StreamElement</tt> for this <tt>StreamParseState</tt>.
     * @since   JSO-0.2
     */
    public StreamElement getContext();
    
    /**
     * <p>
     * Retrieves the "raw" XML string for the current <tt>StreamState</tt>.
     * This string is exactly what the XML processor has just processed.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if the
     * underlying XML parser/processor does not provide this information.</p>
     *
     * @return The "raw" XML string.
     * @since   JSO-0.2
     */
    public String getXML();
    
    /**
     * <p>
     * Retrieves the text for this <tt>StreamState</tt>.</p>
     *
     * <p>
     * The value returned by this method is not <tt>null</tt> if the
     * value returned by {@link #getType} is {@link #TEXT}.</p>
     *
     * @return The text string for this <tt>StreamState</tt>.
     * @since   JSO-0.2
     */
    public String getText();

    /**
     * <p>
     * Retrieves the element name for this <tt>StreamState</tt>.  This
     * contains the local name and namespace URI for the element (either
     * start tag or end tag).</p>
     *
     * <p>
     * The value returned by this method is not <tt>null</tt> only if
     * {@link #getType} returns {@link #ELEMENT_START} or {@link
     * #ELEMENT_END}.</p>
     *
     * @return The name (local name and namespace URI) for this
     * <tt>StreamState</tt>
     * @since   JSO-0.2
     */
    public NSI getName();

    /**
     * <p>
     * Retrieves the set of namespace URI's for the given depth. The
     * returned set only contains URI's declared at the given depth.</p>
     *
     * <p> 
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @param depth The depth to retrieve for.
     * @return The set of declared Namespace URI's
     * @since   JSO-0.2
     * @deprecated As of JSO-0.12, this method is replaced with {@link
     * #getNamespacePrefixes(int)}, since prefixes are unique for a given
     * depth, but not URIs.
     */
    public Set getNamespaceURIs(int depth);
    /**
     * <p>
     * Retrieves the set of namespace prefixes for the given depth. The
     * returned set only contains prefixes declared at the given depth.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @param depth The depth to retrieve for.
     * @return The set of declared Namespace prefixes.
     * @since   JSO-0.12
     */
    public Set getNamespacePrefixes(int depth);
    /**
     * <p>
     * Retrieves the namespace prefix for the given URI (if found).  This
     * method searches all depths, starting with the current depth, for
     * the first match to <tt>uri</tt>.</p>
     *
     * <p>
     * The value of <tt>uri</tt> can be <tt>null</tt> or "", in which
     * case it is considered to be "" (or the "non" namespace).</p>
     *
     * <p>
     * This method returns <tt>null</tt> if a match for <tt>uri</tt>
     * cannot be found.  If <tt>uri</tt> is the default namespace, this
     * method returns "".</p>
     *
     * @param uri The namespace URI to search for.
     * @return The prefix matching the given URI.
     * @since   JSO-0.2
     */
    public String getNamespacePrefix(String uri);
    /**
     * <p>
     * Retrieves the namespace URI for the given prefix (if found).  This
     * method searches all depths, starting with the current depth, for the
     * first match to <tt>prefix</tt>.</p>
     *
     * <p>
     * The value of <tt>prefix</tt> can be <tt>null</tt> or "", in
     * which case it is considered to be "" (or the default namespace).</p>
     *
     * <p>
     * This method returns <tt>null</tt> if a match for <tt>prefix</tt>
     * cannot be found.  If <tt>prefix</tt> is for the "non" namespace, this
     * method returns "".</p>
     *
     * @param prefix The namespace prefix to search for.
     * @return The URI matching the given prefix.
     * @since   JSO-0.2
     */
    public String getNamespaceURI(String prefix);
    
    /**
     * <p>
     * Retrieves the set of attribute names (as <tt>NSI</tt> objects).
     * The returned set does not include any namespace declarations.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt> if the
     * value of {@link #getType} is {@link #ELEMENT_START} or {@link
     * #ELEMENT_END}.</p>
     *
     * @return The set of declared attributes.
     * @since   JSO-0.2
     */
    public Set getAttributeNames();
    /**
     * <p>
     * Retrieves the attribute value with the given (local) name.  This method
     * returns the value of the first attribute located, regardless of
     * namespace.  If more than two attributes have the same local name (but
     * different namespaces), the one returned is not defined by position
     * in XML.  Note that this method always returns <tt>null</tt> if the
     * value returned by {@link #getType} is not {@link #ELEMENT_START} or
     * {@link #ELEMENT_END}.</p>
     *
     * <p>
     * The value of <tt>name</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if no
     * attribute with the given name exists.</p>
     *
     * @param name The (local) name of the attribute.
     * @return The value of the attribute.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.2
     */
    public String getAttributeValue(String name) throws IllegalArgumentException;
    /**
     * <p>
     * Retrieves the attribute value with the given name.  This method
     * returns the value of the attribute matching both the "local name" and
     * namespace URI of <tt>name</tt>.  If the namespace URI for
     * <tt>name</tt> is <tt>null</tt>, the bahavior of this method is
     * the same as {@link #getAttributeValue(String)}.  Note that this method
     * always returns <tt>null</tt> if the value returned by {@link
     * #getType} is {@link #ELEMENT_START} or {@link #ELEMENT_END}.</p>
     *
     * <p>
     * The value of <tt>name</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.  Also, the value
     * <tt>name</tt> returns for {@link NSI#getLocalName} cannot be
     * <tt>null</tt> or "", or an <tt>IllegalArgumentException</tt>
     * is thrown.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if no
     * attribute with the given name exists.</p>
     *
     * @param name The full name of the attribute.
     * @return The value of the attribute.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.2
     */
    public String getAttributeValue(NSI name) throws IllegalArgumentException;
}