/*
 * StreamSource.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.io;


import java.io.IOException;

import org.jabberstudio.jso.Stream;

/**
 * <p>
 * Interface defining an end-point of the XML stream. This interface
 * is intended to provide all the necessary information about the
 * "source" of a connection, and primitive control over that source.
 * The <tt>{@link Stream}</tt> uses this class to
 * configure itself, and to send/receive XML data.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public interface StreamSource {
    //Methods
    /**
     * <p>
     * Retrieves the "expected" hostname for this connection.  This
     * value is the name of the "remote" side of the connection, as
     * a resolvable domain name or IP address. It is mostly provided as a
     * fallback when a more suitable "domain" cannot be provided.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>
     * or "".</p>
     *
     * @return The "remote" hostname.
     * @since   JSO-0.2
     */    
    public String getHostname();
    
    /**
     * <p>
     * Connects this <tt>StreamSource</tt> for input and output.
     * Until this method is called, <tt>{@link #read}</tt>  and <tt>{@link
     * #write}</tt> will throw <tt>IOException</tt>.</p>
     *
     * <p>
     * This method takes the <tt>Stream</tt> performing the method call,
     * allowing this <tt>StreamSource</tt> to initialize itself.</p>
     *
     * <p>
     * The value of <tt>s</tt> must not be <tt>null</tt>, or a
     * <tt>NullPointerException</tt> is thrown.</p>
     *
     * <p>
     * If there are any I/O errors in opening this <tt>StreamSource</tt>,
     * this method throws an <tt>IOException</tt>.  All other errors
     * result in some form of <tt>RuntimeException</tt> being thrown.</p>
     *
     * @param s The <tt>Stream</tt> to connect to.
     * @throws IOException If any I/O errors occur trying to open this
     * <tt>StreamSource</tt>.
     * @throws Exception If any other errors occur trying to open this
     * <tt>StreamSource</tt>.
     * @since   JSO-0.10
     */    
    public void connect(Stream s) throws IOException, Exception;
    /**
     * <p>
     * Disconnects this <tt>StreamSource</tt> from input and output. Once this
     * method is called, <tt>{@link #read}</tt> and <tt>{@link #write}</tt>
     * will throw <tt>IOException</tt>.</p>
     *
     * <p>
     * This method takes the <tt>Stream</tt> performing the method call,
     * allowing this <tt>StreamSource</tt> to uninitialize itself.</p>
     *
     * <p>
     * The value of <tt>s</tt> must not be <tt>null</tt>, or a
     * <tt>NullPointerException</tt> is thrown.</p>
     *
     * <p>
     * If there are any I/O errors in closing this <tt>StreamSource</tt>,
     * this method throws an <tt>IOException</tt>.  All other errors
     * result in some form of <tt>RuntimeException</tt> being thrown.</p>
     *
     * @param s The <tt>Stream</tt> to disconnect from.
     * @throws IOException If any I/O errors occur trying to close this
     * <tt>StreamSource</tt>.
     * @throws Exception If any other errors occur trying to close this
     * <tt>StreamSource</tt>.
     * @since   JSO-0.10
     */    
    public void disconnect(Stream s) throws IOException, Exception;
    
    /**
     * <p>
     * Deterines if this <tt>StreamSource</tt> is ready to attempt a read.</p>
     *
     * <p>
     * This method may return <tt>true</tt> if there is no data immediately
     * available, but is not yet closed or has reached the end of file.</p>
     *
     * @return <tt>true</tt> if this <tt>StreamSource</tt> can still be used.
     * @since JSO-0.10
     */
    public boolean ready() throws IOException;
    /**
     * <p>
     * Reads bytes from this <tt>StreamSource</tt> into the specified byte
     * buffer array.</p>
     *
     * <p>
     * This method returns the number of bytes read, or -1 if the end of the
     * input stream is reached.</p>
     *
     * <p>
     * The value of <tt>buffer</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>offset</tt> and <tt>length</tt> must represent a valid
     * range within <tt>buffer</tt>, or an <tt>IllegalArgumentException</tt>
     * is thrown. For <tt>offset</tt> the valid ranges are:</p>
     *
     * <table border='0' cellpadding='0' cellspacing='0'>
     * <tr><th>offset</th><td><tt>0 <= offset < (buffer.length - length)</tt></td></tr>
     * <tr><th>length</th><td><tt>0 <= length < (buffer.length - offset)</tt></td></tr>
     * </table>
     *
     * @param buffer The buffer to read to.
     * @param offset The offset into <tt>buffer</tt> to start at.
     * @param length The length within <tt>buffer</tt> to use.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws IOException If there are any I/O errors.
     * @since JSO-0.10
     */
    public int read(byte[] buffer, int offset, int length) throws IllegalArgumentException, IOException;
    /**
     * <p>
     * Writes bytes to this <tt>StreamSource</tt> from the specified byte
     * buffer array.</p>
     *
     * <p>
     * This method returns the number of bytes written.</p>
     *
     * <p>
     * The value of <tt>buffer</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>offset</tt> and <tt>length</tt> must represent a valid
     * range within <tt>buffer</tt>, or an <tt>IllegalArgumentException</tt>
     * is thrown. For <tt>offset</tt> the valid ranges are:</p>
     *
     * <table border='0' cellpadding='0' cellspacing='0'>
     * <tr><th>offset</th><td><tt>0 <= offset < (buffer.length - length)</tt></td></tr>
     * <tr><th>length</th><td><tt>0 <= length < (buffer.length - offset)</tt></td></tr>
     * </table>
     *
     * @param buffer The buffer to write from.
     * @param offset The offset into <tt>buffer</tt> to start at.
     * @param length The length within <tt>buffer</tt> to use.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws IOException If there are any I/O errors.
     * @since JSO-0.10
     */
    public int write(byte[] buffer, int offset, int length) throws IllegalArgumentException, IOException;
}