/*
 * SocketChannelSource.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.io.src;


import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ByteChannel;
import java.nio.channels.Selector;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.nio.channels.SelectableChannel;
import java.nio.charset.Charset;

import org.jabberstudio.jso.Stream;
import org.jabberstudio.jso.io.StreamSource;
import org.jabberstudio.jso.util.Utilities;

/**
 * <p>
 * Implementation of <code>StreamSource</code> using <code>ByteChannel</code>
 * instances for the source. The <code>ByteChannel</code> should be a
 * <code>SelectableChannel</code>, and configured for non-blocking operation
 * before {@link #connect} is called.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.8
 */
public class ChannelStreamSource implements StreamSource {
    //Fields
    private ByteChannel     _Channel;
    private boolean         _EOF = false;
    
    private Charset         _Charset;

    //Constructors
    /**
     * <p>
     * Creates a new <code>ChannelStreamSource</code> with the given
     * <code>ByteChannel</code>.  Once created, the underlying
     * <code>ByteChannel</code> is assumed to be ready for {@link #connect} to
     * to be called.</p>
     *
     * <p>
     * The value of <code>channel</code> cannot be <code>null</code>, or an
     * <code>IllegalArgumentException</code> is thrown.</p>
     *
     * @param sc The <code>ByteChannel</code> to connect with.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since JSO-0.8
     */
    public ChannelStreamSource(ByteChannel sc) throws IllegalArgumentException {
        setChannel(sc);
    }
    
    //Methods
    /**
     * <p>
     * Creates a <code>ChannelStreamSource</code> that uses a
     * <code>SocketChannel</code> to the given host and port. Before this
     * method returns, the created <code>SocketChannel</code> is connected
     * and configured for non-blocking operation.</p>
     *
     * <p>
     * The value of <code>host</code> cannot be <code>null</code> or "", or an
     * <code>IllegalArgumentException</code> is thrown.</p>
     *
     * <p>
     * The value of <code>port</code> must be a valid port number, or an
     * <code>IllegalArgumentException</code> is thrown.</p>
     *
     * @param host The IP address or FQDN to connect to.
     * @param port The port number to connect to.
     * @return The <code>ChannelStreamSource</code> for the socket parameters.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws IOException If any I/O errors occur.
     */
    public static ChannelStreamSource createSocket(String host, int port) throws IllegalArgumentException, IOException {
        if (!Utilities.isValidString(host))
            throw new IllegalArgumentException("Host cannot be null or \"\"");
        return createSocket(new InetSocketAddress(host, port));
    }
    /**
     * <p>
     * Creates a <code>ChannelStreamSource</code> that uses a
     * <code>SocketChannel</code> to the given internet address and port.
     * Before this method returns, the created <code>SocketChannel</code> is
     * connected and configured for non-blocking operation.</p>
     *
     * <p>
     * The value of <code>addr</code> cannot be <code>null</code>, or an
     * <code>IllegalArgumentException</code> is thrown.</p>
     *
     * <p>
     * The value of <code>port</code> must be a valid port number, or an
     * <code>IllegalArgumentException</code> is thrown.</p>
     *
     * @param addr The <code>InetAddress</code> to connect to.
     * @param port The port number to connect to.
     * @return The <code>ChannelStreamSource</code> for the socket parameters.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws IOException If any I/O errors occur.
     */
    public static ChannelStreamSource createSocket(InetAddress addr, int port) throws IllegalArgumentException, IOException {
        if (addr == null)
            throw new IllegalArgumentException("InetAddress cannot be null");
        return createSocket(new InetSocketAddress(addr, port));
    }
    /**
     * <p>
     * Creates a <code>ChannelStreamSource</code> that uses a
     * <code>SocketChannel</code> to the given socket address/port.
     * Before this method returns, the created <code>SocketChannel</code> is
     * connected and configured for non-blocking operation.</p>
     *
     * <p>
     * The value of <code>sa</code> cannot be <code>null</code>, or an
     * <code>IllegalArgumentException</code> is thrown.</p>
     *
     * @param sa The <code>InetSocketAddress</code> to connect to.
     * @return The <code>ChannelStreamSource</code> for the socket parameters.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws IOException If any I/O errors occur.
     */
    public static ChannelStreamSource createSocket(SocketAddress sa) throws IllegalArgumentException, IOException {
        SocketChannel       sc;
        
        if (sa == null)
            throw new IllegalArgumentException("SocketAddress cannot be null");
        
        sc = SocketChannel.open(sa);
        sc.configureBlocking(false);
        
        return new ChannelStreamSource(sc);
    }
    
    /**
     * <p>
     * Retrieves the <code>ByteChannel</code> owned by this
     * <code>ChannelStreamSource</code>. This method is provided for
     * convenience, since the channel is likely used as a key.</p>
     *
     * <p><b>Do not directly call any of the get or put methods of this
     * <tt>ByteChannel</tt>. Doing so results in undefined behavior</b>.</p>
     *
     * @return The <code>ByteChannel</code> for this
     * <Code>ChannelStreamSource</code>.
     * @since   JSO-0.8
     */ 
    public final ByteChannel getChannel() {
        return _Channel;
    }
    private final void setChannel(ByteChannel sc) {
        if (sc == null)
            throw new IllegalArgumentException("ByteChannel cannot be null");
        _Channel = sc;
    }
    
    public String getHostname() {
        ByteChannel channel = getChannel();
        String      host = "localhost";

        if (channel instanceof SocketChannel) {
            host = ((SocketChannel)channel).socket().getInetAddress().getHostName();
        }
        
        return host;
    }

    public void connect(Stream s) throws IOException {
        if (!getChannel().isOpen())
            throw new IOException("Channel not open");
    }
    
    public void disconnect(Stream s) throws IOException, Exception {
        getChannel().close();
    }
    
    public boolean ready() throws IOException {
        return getChannel().isOpen() || !isEOF();
    }
    
    private boolean isEOF() {
        return _EOF;
    }
    private void setEOF(boolean eof) {
        _EOF = eof;
    }
    
    public int read(byte[] buffer, int offset, int length) throws IllegalArgumentException, IOException {
        ByteChannel     channel;
        ByteBuffer      in;
        int             amt = 0, idx;

        if (isEOF()) {
            amt = -1;
        } else {
            //Validate parameters
            if (buffer == null)
                throw new IllegalArgumentException("buffer cannot be null");
            if ((offset < 0) || (buffer.length > (offset + length)))
                throw new IllegalArgumentException("offset not in valid range");
            if ((length < 0) || (length > (buffer.length - offset)))
                throw new IllegalArgumentException("length not in valid range");

            in = ByteBuffer.allocate(length);
            channel = getChannel();
            amt = channel.read(in);
            
            if (amt > 0) {
                in.flip();
                in.get(buffer, offset, amt);
            } else if (amt == -1) {
                setEOF(true);
            }
        }
        
        return amt;
    }

    public int write(byte[] buffer, int offset, int length) throws IllegalArgumentException, IOException {
        ByteChannel     channel;
        int             amt = 0, idx;
        
        //Validate parameters
        if (buffer == null)
            throw new IllegalArgumentException("buffer cannot be null");
        if ((offset < 0) || (buffer.length < (offset + length)))
            throw new IllegalArgumentException("offset not in valid range");
        if ((length < 0) || (length > (buffer.length - offset)))
            throw new IllegalArgumentException("length not in valid range");
        
        channel = getChannel();

	ByteBuffer out = ByteBuffer.wrap(buffer, offset, length);
	amt = channel.write(out);
	
	if (amt < length) {
	    int waitTime = 0;
	    for (;;) {  
		int w = 0;
		while ((w = channel.write(out)) > 0) {
		    amt += w;
		    waitTime = 10;  // unlikely to loop many times
		}
		
		if (amt >= length) break;
		
		try {
		    Thread.sleep(waitTime);
		    // keep increasing wait time in order to release some cpu
		    waitTime += waitTime;
		} catch (Exception e) {}
	    }
	}
	return amt;
    }

}
