/*
 * SocketStreamSource.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.io.src;


import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.nio.charset.Charset;

import org.jabberstudio.jso.Stream;
import org.jabberstudio.jso.event.StreamStatusEvent;
import org.jabberstudio.jso.event.StreamStatusListener;

/**
 * <p>
 * Implementation of <tt>StreamSource</tt> for using sockets as a
 * <tt>StreamSource</tt>.</p>
 *
 * <p>
 * Starting with version 0.4, instances of this class cannot be reused.  Each
 * connection will require a new <tt>SocketStreamSource</tt>.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public class SocketStreamSource extends IOStreamSource {
    //Inner-classes
    private class StatusWatcher implements StreamStatusListener {
        //Methods
        public void statusChanged(StreamStatusEvent evt) {
            try {
                configureSocket(getSocket(), evt.getStream().getCurrentStatus());
            } catch (Exception e) {
            }
        }
    }
    
    //Fields
    private Socket          _Socket;
    
    private StatusWatcher   _Watch;
    
    //Constructors
    /**
     * <p>
     * Creates a new <tt>SocketStreamSource</tt> with the given host and
     * port. Once created, the underlying <tt>Socket</tt> is connected to
     * the network. Calling {@link #connect} is still necessary to setup the
     * I/O.</p>
     *
     * <p>
     * The value of <tt>host</tt> must be a valid (and resolvable) DNS name
     * or IP address, or an <tt>UnknownHostException</tt> is thown.</p>
     * 
     * <p>
     * The value of <tt>port</tt> must be a positive integer in the proper
     * range (0 &lt; <tt>port</tt> &lt; Short.MAX_VALUE), or an
     * <tt>IllegalArgumentException</tt>. is thrown.</p>
     *
     * @param host The host to connect to.
     * @param port The port to connect to.
     * @throws IllegalArgumeException If any parameters are invalid.
     * @throws UnknownHostException If the host cannot be resolved.
     * @throws IOException If any I/O errors occur.
     * @since   JSO-0.2
     */
    public SocketStreamSource(String host, int port) throws UnknownHostException, IllegalArgumentException, IOException {
        this(new Socket(host, port));
    }
    /**
     * <p>
     * Creates a new <tt>SocketStreamSource</tt> with the given
     * <tt>Socket</tt>.  Once created, the underlying <tt>Socket</tt>
     * is connected to the network.  Calling {@link #connect} is still
     * necessary to setup the I/O.</p>
     *
     * <p>
     * The value of <tt>s</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param s The socket to connect with.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since JSO-0.4
     */
    public SocketStreamSource(Socket s) throws IllegalArgumentException {
        setSocket(s);
    }
    
    //Methods
    /**
     * <p>
     * Retrieves the <tt>Socket</tt> for this <tt>SocketStreamSource</tt>.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The backing <tt>Socket</tt>.
     * @since   JSO-0.4
     */
    protected Socket getSocket() {
        return _Socket;
    }
    /**
     * <p>
     * Sets the <tt>Socket</tt> for this <tt>SocketStreamSource</tt>.</p>
     *
     * <p>
     * The value of <tt>s</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param s The new backing <tt>Socket</tt>.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.4
     */
    protected void setSocket(Socket s) throws IllegalArgumentException {
        if (s == null)
            throw new IllegalArgumentException("Socket cannot be null");
        _Socket = s;
    }
    
    /**
     * <p>
     * Retrieves the <tt>InetAddress</tt> for this <tt>SocketStreamSource</tt>.
     * This method is a convenience for retrieving the address this
     * <tt>SocketStreamSouce</tt> connects to.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return the <tt>InetAddress</tt> for this <tt>SocketStreamSource</tt>.
     * @since   JSO-0.4
     */
    protected InetAddress getInetAddress() {
        return getSocket().getInetAddress();
    }
    
    /**
     * <p>
     * Retreives the port number for this <tt>SocketStreamSource</tt>.</p>
     *
     * <p>
     * The value returned by this method is always a positive number.</p>
     *
     * @return The port to connect to.
     * @since   JSO-0.2
     */
    public int getPort() {
        return getSocket().getPort();
    }
    
    public String getHostname() {
        return getInetAddress().getHostName();
    }
    
    public void connect(Stream s) throws IOException, Exception {
        Socket sock = getSocket();

        s.addStreamStatusListener(_Watch = new StatusWatcher());
        configureSocket(sock, s.getCurrentStatus());
        setInput(sock.getInputStream());
        setOutput(sock.getOutputStream());
        
        super.connect(s);
    }
    public void disconnect(Stream s) throws IOException, Exception {
        super.disconnect(s);
        s.removeStreamStatusListener(_Watch);
        _Socket.close();
        _Watch = null;
    }

    /**
     * <p>
     * Configures the current socket. This method ensures the socket is using
     * TCP keep-alive and sets the timeout to 2 seconds.</p>
     *
     * @throws SocketException If there is any error configuring the socket.
     * @since   JSO-0.12
     */
    protected void configureSocket(Socket sock, Stream.Status stat) throws SocketException {
        if (sock == null)
            sock = getSocket();
        if (stat == null)
            stat = Stream.DISCONNECTED;
        
        if ((sock instanceof javax.net.ssl.SSLSocket) || !stat.isOpened()) {
            sock.setSoTimeout(0);
        } else {
            sock.setSoTimeout(100);
        }
        sock.setKeepAlive(true);
    }
}