/*
 * UnsupportedMechanismException.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.sasl;


import java.security.NoSuchAlgorithmException;
import java.util.Collections;
import java.util.List;

import org.jabberstudio.jso.JSOImplementation;

/**
 * <p>
 * Exception thrown when reported mechanisms do not match the supported
 * mechanisms.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public class NoSupportedMechanismsException extends NoSuchAlgorithmException {
    //Fields
    private List    _Mechs;
    
    //Methods
    /**
     * <p>
     * Creates a new <tt>NoSupportedMechanismsException</tt> with the given
     * list of reported exceptions.</p>
     *
     * <p>
     * The value of <tt>mechs</tt> may be <tt>null</tt>, in which case it is
     * considered to be an empty list.</p>
     *
     * @param mechs The list of reported mechanisms.
     * @since   JSO-0.12
     */
    public NoSupportedMechanismsException(List mechs) {
        this(null, mechs);
    }
    /**
     * <p>
     * Creates a new <tt>NoSupportedMechanismsException</tt> with the given
     * message and list of reported exceptions.</p>
     *
     * <p>
     * The value of <tt>msg</tt> may be <tt>null</tt> or "", in which case it
     * is ignored.</p>
     *
     * <p>
     * The value of <tt>mechs</tt> may be <tt>null</tt>, in which case it is
     * considered to be an empty list.</p>
     *
     * @param msg The message for this.
     * @param mechs The list of reported mechanisms.
     * @since   JSO-0.12
     */
    public NoSupportedMechanismsException(String msg, List mechs) {
        super(msg);
        
        if (mechs == null)
            mechs = Collections.EMPTY_LIST;
        _Mechs = Collections.unmodifiableList(mechs);
    }
    
    //Methods
    /**
     * <p>
     * Retrieves the list of reported mechanisms.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each element
     * in the returned list is the string name of a SASL mechanism. The
     * returned list cannot be modified.</p>
     *
     * @return The list of reported mechanisms.
     * @since   JSO-0.12
     */
    public List getReportedMechanisms() {
        return _Mechs;
    }
}