/*
 * SASLMechanismProvider.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.sasl;


import java.util.List;
import java.util.Map;


/**
 * <p>
 * Interface for creating <tt>SASLMechanism</tt> instances. This is the
 * provider queried for.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public interface SASLMechanismProvider {
    //Methods
    /**
     * <p>
     * Retrieves the list of client mechanism names this
     * <tt>SASLMechanismProvider</tt> supports, within the constraints of the
     * given properties.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each element
     * in the returned list is a string.</p>
     *
     * <p>
     * The value of <tt>props</tt> may be <tt>null</tt>, in which case an
     * empty map is used.</p>
     *
     * @param props The map of authentication properties.
     * @return The list of supported client mechanisms.
     * @since   JSO-0.12
     */
    public List getClientMechanismNames(Map props);
    /**
     * <p>
     * Creates a new client <tt>SASLMechanism</tt> using the given
     * parameters. This method creates a client <tt>SASLMechanism</tt> for the
     * first supported mechanism and within the constraints of the given
     * properties.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if a client
     * <tt>SASLMechanism</tt> cannot be created.</p>
     *
     * <p>
     * The value of <tt>ci</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown. Also, if the call to {@link
     * SASLClientInfo#isValid} on <tt>ci</tt> returns <tt>false</tt>, an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param ci The <tt>SASLClientIInfo</tt> to create by.
     * @return The client <tt>SASLMechanism</tt> appropriate for the given
     * parameters.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @since   JSO-0.12
     */
    public SASLMechanism createClientMechanism(SASLClientInfo ci) throws IllegalArgumentException;
    
    /**
     * <p>
     * Retrieves the list of server mechanism names this
     * <tt>SASLMechanismProvider</tt> supports, within the constraints of the
     * given properties.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each element
     * in the returned list is a string.</p>
     *
     * <p>
     * The value of <tt>props</tt> may be <tt>null</tt>, in which case an
     * empty map is used.</p>
     *
     * @param props The map of authentication properties.
     * @return The list of supported server mechanisms.
     * @since   JSO-0.12
     */
    public List getServerMechanismNames(Map props);
    /**
     * <p>
     * Creates a new server <tt>SASLMechanism</tt> using the given
     * parameters. This method server a client <tt>SASLMechanism</tt> for the
     * first supported mechanism and within the constraints of the given
     * properties.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if a server
     * <tt>SASLMechanism</tt> cannot be created.</p>
     *
     * <p>
     * The value of <tt>si</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown. Also, if the call to {@link
     * SASLServerInfo#isValid} on <tt>si</tt> returns <tt>false</tt>, an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param si The <tt>SASLServerInfo</tt> to create by.
     * @return The server <tt>SASLMechanism</tt> appropriate for the given
     * parameters.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @since   JSO-0.12
     */
    public SASLMechanism createServerMechanism(SASLServerInfo si) throws IllegalArgumentException;
}