/*
 * StartTLSFeatureConsumer.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.tls;


import java.util.Collections;

import org.saxpath.SAXPathException;

import org.jabberstudio.jso.PacketRouter;
import org.jabberstudio.jso.StreamDataFactory;
import org.jabberstudio.jso.StreamException;
import org.jabberstudio.jso.StreamFeature;
import org.jabberstudio.jso.StreamFeatureset;
import org.jabberstudio.jso.event.PacketDispatcher;
import org.jabberstudio.jso.event.PacketEvent;
import org.jabberstudio.jso.features.AbstractFeatureConsumer;
import org.jabberstudio.jso.features.FeatureContext;
import org.jabberstudio.jso.util.XPathListener;
import org.jabberstudio.jso.xpath.XPathSupport;

/**
 * <p>
 * A <tt>FeatureConsumer</tt> implementation for consuming STARTTLS.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public class StartTLSSocketFeatureConsumer extends AbstractFeatureConsumer {
    //Inner-classes
    private class StartTLSWatcher extends XPathListener {
        //Constructors
        public StartTLSWatcher(XPathSupport support) throws SAXPathException {
            super(support, "tls:*", Collections.singletonMap("tls", StartTLSPacket.NAMESPACE));
        }
        
        //Methods
        public void packetMatched(PacketEvent evt) {
            FeatureContext              ctx = getFeatureContext();
            StartTLSSocketStreamSource  ss = getStreamSource();
            PacketDispatcher            dispatch = ctx.getDispatcher();
            
            try {
                ss.negotiateClientTLS();

                dispatch.removePacketListener(PacketEvent.RECEIVED, this);
                completeSucceededConsumption(true);
            } catch (Throwable t) {
                dispatch.removePacketListener(PacketEvent.RECEIVED, this);
                completeFailedConsumption(t);
            }
        }
    }
    
    //Fields
    private StartTLSSocketStreamSource  _Src;
    
    private FeatureContext              _FC;
    
    //Constructors
    /**
     * <p>Creates a new <tt>StartTLSSocketFeatureConsumer</tt>, with no
     * <tt>StartTLSSocketStreamSource</tt>. The method {@link #setStreamSource}
     * must be called before consumption begins.</p>
     *
     * @since   JSO-0.12
     */
    public StartTLSSocketFeatureConsumer() {
    }
    /**
     * <p>Creates a new <tt>StartTLSSocketFeatureConsumer</tt> with the given
     * <tt>StartTLSSocketStreamSource</tt>. This constructor is provided as a
     * convienence for using it in a "single-shot" (e.g. no re-use of the
     * <tt>FeatureConsumer</tt>).</p>
     *
     * <p>
     * The value of <tt>ss</tt> may be <tt>null</tt>. If <tt>null</tt>, the
     * method {@link #setStreamSource} must be called before consumption
     * begins.</p></p>
     *
     * @param ss The <tt>StartTLSSocketStreamSource</tt> to use.
     * @since   JSO-0.12
     */
    public StartTLSSocketFeatureConsumer(StartTLSSocketStreamSource ss) {
        setStreamSource(ss);
    }
    
    //Methods
    /**
     * <p>
     * Retrieves the <tt>StartTLSSocketStreamSource</tt> to use with this
     * <tt>StartTLSSocketFeatureConsumer</tt>.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt>.</p>
     *
     * @return The <tt>StartTLSSocketStreamSource</tt> to use.
     * @since   JSO-0.12
     */
    public StartTLSSocketStreamSource getStreamSource() {
        return _Src;
    }
    /**
     * <p>
     * Sets the <tt>StartTLSSocketStreamSource</tt> to use with this
     * <tt>StartTLSSocketFeatureConsumer</tt>.</p>
     *
     * <p>
     * The value of <tt>ss</tt> may be <tt>null</tt>, in which case any
     * previous value is removed.</p>
     *
     * @param ss The new <tt>StartTLSSocketStreamSource</tt>.
     * @since   JSO-0.12
     */
    public void setStreamSource(StartTLSSocketStreamSource ss) {
        _Src = ss;
    }
    
    /**
     * <p>
     * Finds the <tt>StreamFeature</tt> in the given <tt>StreamFeatureset</tt>
     * that can be consumed. This implementation searches for the first
     * <tt>StartTLSFeature</tt> in <tt>fs</tt>.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if a consumable
     * <tt>StreamFeature</tt> could not be found.</p>
     *
     * <p>
     * The value of <tt>fs</tt> will never be <tt>null</tt>.</p>
     *
     * @param fs The <tt>StreamFeatureset</tt> to search in
     * @return The consumable <tt>StreamFeature</tt>.
     * @since   JSO-0.12
     */
    public StreamFeature findFeature(StreamFeatureset fs) {
        return (StreamFeature)fs.getFirstElement(StartTLSFeature.NAME);
    }
    /**
     * <p>
     * Begins the STARTTLS process. This method registers a
     * <tt>PacketListener</tt> for STARTTLS packets, and sends the initial
     * <tt>StartTLSPacket</tt>.</p>
     *
     * @throws IllegalStateException If {@link #getStreamSource} returns
     * <tt>null</tt>, or the STARTTLS <tt>PacketListener</tt> could not be
     * registered.
     * @throws StreamException If there is a stream-level error
     * @since   JSO-0.12
     */
    protected void startConsumption() throws IllegalStateException, StreamException {
        FeatureContext          ctx = getFeatureContext();
        PacketRouter            router = ctx.getRouter();
        PacketDispatcher        dispatch = ctx.getDispatcher();
        StreamDataFactory       sdf = router.getDataFactory();
        StartTLSPacket          packet;
        
        if (getStreamSource() == null)
            throw new IllegalStateException("StartTLSSocketStreamSource cannot be null");

        try {
            dispatch.addPacketListener(PacketEvent.RECEIVED, new StartTLSWatcher(sdf));
        } catch (SAXPathException spe) {
            throw new IllegalStateException("Could not register STARTTLS packet listener");
        }
        
        packet = (StartTLSPacket)sdf.createPacketNode(StartTLSPacket.NAME_STARTTLS, StartTLSPacket.class);

        router.send(packet);
    }
    public void reset() {
        super.reset();
        
        setStreamSource(null);
    }
}