/*
 * StartTLSSocketStreamSource.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.tls;


import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.security.NoSuchAlgorithmException;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;

import org.jabberstudio.jso.Packet;
import org.jabberstudio.jso.Stream;
import org.jabberstudio.jso.StreamException;
import org.jabberstudio.jso.io.StreamSource;
import org.jabberstudio.jso.io.src.InputStreamQueue;
import org.jabberstudio.jso.io.src.SocketStreamSource;

/**
 * <p>
 * {@link StreamSource} implementation for STARTTLS support using a (blocking)
 * socket connection.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public class StartTLSSocketStreamSource extends SocketStreamSource {
    //Fields
    private Stream          _Conn;
    private SSLContext      _Ctx;
    private boolean         _Secured;
    
    //Constructors
    /**
     * <p>
     * Creates a new <tt>StartTLSSocketStreamSource</tt> using the given
     * host and port. This constructor attempts to obtain a suitable
     * <tt>SSLContext</tt> instance.</p>
     *
     * <p>
     * The value of <tt>host</tt> must be a valid (and resolvable) DNS name
     * or IP address, or an <tt>UnknownHostException</tt> is thown.</p>
     * 
     * <p>
     * The value of <tt>port</tt> must be a positive integer in the proper
     * range (0 &lt; <tt>port</tt> &lt; Short.MAX_VALUE), or an
     * <tt>IllegalArgumentException</tt>. is thrown.</p>
     *
     * @param host The host to connect to.
     * @param port The port to connect to.
     * @throws IllegalArgumeException If any parameters are invalid.
     * @throws NoSuchAlgorithmException If a <tt>SSLContext</tt> could not be
     * obtained for "TLS".
     * @throws UnknownHostException If the host cannot be resolved.
     * @throws IOException If any I/O errors occur.
     * @since   JSO-0.12
     */
    public StartTLSSocketStreamSource(String host, int port) throws IllegalArgumentException, NoSuchAlgorithmException, UnknownHostException, IOException {
        this(new Socket(host, port), SSLContext.getInstance("TLS"));
    }
    /**
     * <p>
     * Creates a new <tt>StartTLSSocketStreamSource</tt> using the given
     * host, port, and <tt>SSLContext</tt>.</p>
     *
     * <p>
     * The value of <tt>host</tt> must be a valid (and resolvable) DNS name
     * or IP address, or an <tt>UnknownHostException</tt> is thown.</p>
     * 
     * <p>
     * The value of <tt>port</tt> must be a positive integer in the proper
     * range (0 &lt; <tt>port</tt> &lt; Short.MAX_VALUE), or an
     * <tt>IllegalArgumentException</tt>. is thrown.</p>
     *
     * <p>
     * The value of <tt>ctx</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param host The host to connect to.
     * @param port The port to connect to.
     * @param ctx The <tt>SSLContext</tt> to use.
     * @throws IllegalArgumeException If any parameters are invalid.
     * @throws UnknownHostException If the host cannot be resolved.
     * @throws IOException If any I/O errors occur.
     * @since   JSO-0.12
     */
    public StartTLSSocketStreamSource(String host, int port, SSLContext ctx) throws IllegalArgumentException, UnknownHostException, IOException {
        this(new Socket(host, port), ctx);
    }
    
    /**
     * <p>
     * Creates a new <tt>StartTLSSocketStreamSource</tt> with the given
     * <tt>Socket</tt>. This constructor attempts to obtain a suitable
     * <tt>SSLContext</tt> instance.</p>
     *
     * <p>
     * The value of <tt>plain</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param plain The socket to connect with.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @throws NoSuchAlgorithmException If a <tt>SSLContext</tt> could not be
     * obtained for "TLS".
     * @since JSO-0.12
     */
    public StartTLSSocketStreamSource(Socket plain) throws IllegalArgumentException, NoSuchAlgorithmException {
        this(plain, SSLContext.getInstance("TLS"));
    }
    /**
     * <p>
     * Creates a new <tt>StartTLSSocketStreamSource</tt> with the given
     * <tt>Socket</tt> and <tt>SSLContext</tt>.</p>
     *
     * <p>
     * The value of <tt>plain</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>ctx</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param plain The socket to connect with.
     * @param ctx The <tt>SSLContext</tt> to use.
     * @throws IllegalArgumeException If any parameters are invalid.
     * @since JSO-0.12
     */
    public StartTLSSocketStreamSource(Socket plain, SSLContext ctx) throws IllegalArgumentException {
        super(plain);
        
        if (ctx == null)
            throw new IllegalArgumentException("SSLContext cannot be null");
        _Ctx = ctx;
    }
    
    //Methods
    /**
     * <p>
     * Determines if this <tt>StartTLSSocketStreamSource</tt> has been
     * secured.</p>
     *
     * @return <tt>true</tt> if the TLS layer has been installed.
     * @since   JSO-0.12
     */
    public boolean isSecured() {
        return _Secured;
    }
    private void setSecured(boolean secured) {
        _Secured = secured;
    }
    
    /**
     * <p>
     * Retrieves the <tt>SSLContext</tt> to use for this
     * <tt>StartTLSSocketStreamSource</tt>. This object is used to create
     * <tt>SSLSocket</tt> objects.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The <tt>SSLContext</tt> to use.
     * @since   JSO-0.12
     */
    public SSLContext getTLSContext() {
        return _Ctx;
    }
    
    /**
     * <p>
     * Retrieves the <tt>Stream</tt> this <tt>StartTLSSocketStreamSource</tt>
     * is connected to.</p>
     *
     * <p>
     * This method may return <tt>null</tt> if {@link #connect} has not yet
     * been called.</p>
     *
     * @return the <tt>Stream</tt> connected to.
     * @since   JSO-0.12
     */
    protected Stream getStream() {
        return _Conn;
    }

    /**
     * <p>
     * Connects this <tt>StreamSource</tt> for input and output. Until this
     * method is called, {@link #read} and {@link #write} will throw
     * <tt>IOException</tt>.</p>
     *
     * <p>
     * This implementation calls its superclass' implementation, then stores
     * <tt>s</tt> for later use.</p>
     *
     * {@inheritDoc}
     * @since   JSO-0.12
     */    
    public void connect(Stream s) throws IOException, Exception {
        super.connect(s);
        
        _Conn = s;
    }
    /**
     * <p>
     * Disconnects this <tt>StreamSource</tt> from input and output. Once this
     * method is called, {@link #read} and {@link #write} will throw
     * <tt>IOException</tt>.</p>
     *
     * <p>
     * This implementation forgets any previously stored <tt>Stream</tt>, then
     * calls its superclass' implementation.</p>
     *
     * {@inheritDoc}
     * @since   JSO-0.12
     */    
    public void disconnect(Stream s) throws IOException, Exception {
        _Conn = null;
        
        super.disconnect(s);
    }

    /**
     * <p>
     * Performs the TLS negotation in "client" mode. When this method completes,
     * this <tt>StartTLSStreamSource</tt> will be using a secured
     * connection.</p>
     *
     * @throws IOException If any I/O errors occur.
     * @since   JSO-0.12
     */
    public synchronized void negotiateClientTLS() throws IOException {
        SSLContext          ctx = getTLSContext();
        Socket              plain = getSocket();
        SSLSocket           secure;
        String              host = plain.getInetAddress().getHostName();
        int                 port = plain.getPort();

        getInputQueue().stopAndWait(false);
        
        //Layer the SSL Socket in "client" mode
        secure = (SSLSocket)ctx.getSocketFactory().createSocket(plain, host, port, true);
        configureSocket(secure, null);

        setSocket(secure);
        setInput(secure.getInputStream());
        setOutput(secure.getOutputStream());
        
        secure.startHandshake();
        setSecured(true);
        
        getInputQueue().start();
    }
    /**
     * <p>
     * Starts the TLS negotation in "server" mode. When this method completes,
     * this <tt>StartTLSStreamSource</tt> will be using a secured
     * connection.</p>
     *
     * <p>
     * This method sends out the &lt;proceed/&gt; <tt>Packet</tt>, to ensure
     * the proper state is maintained.</p>
     *
     * @throws StreamException If any stream-level errors occur.
     * @throws IOException If any I/O errors occur.
     * @since   JSO-0.12
     */
    public synchronized void negotiateServerTLS() throws StreamException, IOException {
        Stream              conn = getStream();
        Packet              proceed;
        SSLContext          ctx = getTLSContext();
        Socket              plain = getSocket();
        SSLSocket           secure;
        String              host = plain.getLocalAddress().getHostName();
        int                 port = plain.getLocalPort();
        
        getInputQueue().stopAndWait(false);
        
        //Send <proceed/>
        proceed = (StartTLSPacket)conn.getDataFactory().createPacketNode(StartTLSPacket.NAME_PROCEED);
        conn.send(proceed);
        
        //Layer the SSLSocket in "server" mode
        secure = (SSLSocket)ctx.getSocketFactory().createSocket(plain, host, port, true);
        secure.setUseClientMode(false);
        secure.setNeedClientAuth(true);
        configureSocket(secure, null);
        
        setSocket(secure);
        setInput(secure.getInputStream());
        setOutput(secure.getOutputStream());
        setSecured(true);
        
        getInputQueue().start();
    }
}