/*
 * BasicStreamObjectVisitor.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.util;


import java.util.Iterator;

import org.jabberstudio.jso.StreamAttribute;
import org.jabberstudio.jso.StreamNamespace;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.StreamNode;
import org.jabberstudio.jso.StreamObjectVisitor;
import org.jabberstudio.jso.StreamText;
import org.jabberstudio.jso.SerializedNodes;

/**
 * <p>
 * Provides a default implementation for <code>StreamNodeVisitor</code>.
 * Subclasses are free to implement only those methods that are necessary.
 * Also, the method <code>{@link #visit(StreamElement)}</code> traverses
 * the <code>StreamAttribute</code> objects and child <code>StreamNode</code>
 * objects, so subclasses simply need to call <code>super.visit(elem)</code>
 * to continue processing the tree.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public class BasicStreamObjectVisitor implements StreamObjectVisitor {
    //Constructors
    /**
     * <p>
     * Creates a new <code>BasicStreamNodeVisitor</code>.</p>
     *
     * @since   JSO-0.2
     */
    public BasicStreamObjectVisitor() {
    }
    
    //Methods
    /** <p>
     * Visits a <code>StreamElement</code>.  This method is called by
     * implementations of <code>StreamElement</code> for specialized
     * processing.  This implementation visits each of the
     * <code>StreamAttribute</code> objects, then each of the child
     * <code>StreamNode</code> objects of <code>elem</code>.</p>
     *
     * <p>
     * The value of <code>elem</code> is assumed to never be <code>null</code>,
     * since this method is not called directly by "users".</p>
     *
     * @param elem The <code>StreamElement</code> being visited.
     * @since   JSO-0.2
     */
    public void visit(StreamElement elem) {
        Iterator    attrItr = elem.getAttributes().values().iterator();
        Iterator    childItr = elem.listNodes().iterator();
        
        //Process Attributes
        while (attrItr.hasNext()) {
            StreamAttribute attr = (StreamAttribute)attrItr.next();
            
            attr.accept(this);
        }
        
        //Process children
        while (childItr.hasNext()) {
            StreamNode  node = (StreamNode)childItr.next();
            
            node.accept(this);
        }
    }
    /**
     * <p>
     * Visits a <code>StreamAttribute</code>.  This method is called by
     * implementations of <code>StreamAttribute</code> for specialized
     * processing.  This implementation does nothing.</p>
     *
     * <p>
     * The value of <code>attr</code> is assumed to never be <code>null</code>,
     * since this method is not called directly by "users".</p>
     *
     * @param attr The <code>StreamAttribute</code> being visited.
     * @since   JSO-0.2
     */
    public void visit(StreamAttribute attr) {
    }
    /**
     * <p>
     * Visits a <code>StreamNamespace</code>.  This method is called by
     * implementations of <code>StreamNamespace</code> for specialized
     * processing.</p>
     *
     * <p>
     * The value of <code>ns</code> is assumed to never be <code>null</code>,
     * since this method is not called directly by "users".</p>
     *
     * @param ns The <code>StreamNamespace</code> being visited.
     * @since   JSO-0.4.1
     */
    public void visit(StreamNamespace ns) {
    }
    /** <p>
     * Visits a <code>StreamText</code>.  This method is called by
     * implementations of <code>StreamText</code> for specialized
     * processing.  This implementation does nothing.</p>
     *
     * <p>
     * The value of <code>text</code> is assumed to never be <code>null</code>,
     * since this method is not called directly by "users".</p>
     *
     * @param text The <code>StreamText</code> being visited.
     * @since   JSO-0.2
     */
    public void visit(StreamText text) {
    }
    
     /**
     * <p>
     * Visits a <code>SerializedNodes</code>.  This method is called by
     * implementations of <code>SerializedNodes</code> for specialized
     * processing.</p>
     *
     * <p>
     * The value of <code>text</code> is assumed to never be <code>null</code>,
     * since this method is not called directly by "users".</p>
     *
     * @param text The <code>SerializedNodes</code> being visited.
     */
    public void visit(SerializedNodes text){
    }
    
    protected void visitAttributes(StreamElement elem) {
        Iterator    itr = elem.getAttributes().values().iterator();
        
        while (itr.hasNext()) {
            visit((StreamAttribute)itr.next());
        }
    }
    protected void visitNamespaces(StreamElement elem) {
        Iterator    itr = elem.getDeclaredNamespaces().iterator();
        
        while (itr.hasNext()) {
            visit((StreamNamespace)itr.next());
        }
    }
    protected void visitChildren(StreamElement elem) {
        Iterator    itr = elem.listNodes().iterator();
        
        while (itr.hasNext()) {
            ((StreamNode)itr.next()).accept(this);
        }
    }
}