/*
 * Utilities.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.util;


import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.Set;


import org.jaxen.JaxenException;
import org.jaxen.XPath;

import org.jabberstudio.jso.StreamNamespace;

/**
 * <p>
 * A class containing "constants" fields and utility methods useful throughout
 * the class library. It also contains a "constant" for the current version of
 * this library.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.2
 */
public final class Utilities {
    //"Constants"
    /** String identifying the current version of this library.
     * @since   JSO-0.2
     */
    public static final String VERSION = "$VERSION$";

    /** Namespace URI for the client "default" namespace.
     * @since   JSO-0.2
     */
    public static final String CLIENT_NAMESPACE = "jabber:client";
    /** Namespace URI for the accept-component "default" namespace..
     * @since   JSO-0.2
     */
    public static final String COMPONENT_ACCEPT_NAMESPACE = "jabber:component:accept";
    /** Namespace URI for the connect-component "default" namespace.
     * @since   JSO-0.2
     */
    public static final String COMPONENT_CONNECT_NAMESPACE = "jabber:component:connect";
    /** Namespace URI for the server "default" namespace.
     * @since   JSO-0.2
     */
    public static final String SERVER_NAMESPACE = "jabber:server";

    /** Namespace URI for XML-Streams.
     * @since   JSO-0.2
     */
    public static final String STREAM_NAMESPACE = "http://etherx.jabber.org/streams";
    /** Namespace prefix for XML-Streams.
     * @since   JSO-0.10
     */
    public static final String STREAM_PREFIX = "stream";
    
    /** Namespace URI for XML.
     * @since   JSO-0.10
     */
    public static final String XML_NAMESPACE = "http://www.w3.org/XML/1998/namespace";
    /** Namespace prefix for XML.
     * @since   JSO-0.10
     */
    public static final String XML_PREFIX = "xml";
    
    //Constructors
    private Utilities() {
    }

    //Methods
    /**
     * <p>
     * Retrieves the current version of JSO.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt> or "".</p>
     *
     * @return The JSO version, as a string.
     * @since   JSO-0.12
     */
    public static String getVersion() {
        return VERSION;
    }
    /**
     * <p>
     * Tests the given string for validity.  A string is considered valid
     * if it is not <tt>null</tt> and is not "".</p>
     *
     * @param test The string to test.
     * @return <tt>true</tt> if the string is valid.
     * @since   JSO-0.2
     */
    public static final boolean isValidString(String test) {
        return (test != null) && (test.length() > 0);
    }
    /**
     * <p>
     * Compares two strings for equality, allowing for <tt>null</tt>
     * values.  This comparison is case-sensitive.  For a case insensitive
     * comparison, use <tt>{@link #equateStringsIgnoreCase}</tt>.</p>
     *
     * @param cmp1 The first string to compare.
     * @param cmp2 The second string to compare.
     * @return The result of the comparison.
     * @since   JSO-0.2
     */
    public static final boolean equateStrings(String cmp1, String cmp2) {
        boolean     results = (cmp1 == cmp2);

        if (!results) {
            if ((cmp1 != null) && (cmp2 != null)) {
                results = cmp1.equals(cmp2);
            } else {
                results = false;
            }
        }

        return results;
    }
    /**
     * <p>
     * Compares two strings for equality, allowing for <tt>null</tt>
     * values.  This comparison is case-insensitive.  For a case-sensitive
     * comparison, use <tt>{@link #equateStrings}</tt>.</p>
     *
     * @param cmp1 The first string to compare.
     * @param cmp2 The second string to compare.
     * @return The result of the comparison.
     * @since   JSO-0.2
     */
    public static final boolean equateStringsIgnoreCase(String cmp1, String cmp2) {
        boolean     results = (cmp1 == cmp2);

        if (!results) {
            if ((cmp1 != null) && (cmp2 != null)) {
                results = cmp1.equalsIgnoreCase(cmp2);
            } else {
                results = false;
            }
        }

        return results;
    }
    
    /**
     * <p>
     * Adds the set of namespaces to the XPath. The namespaces are
     * determined from <tt>nss</tt>, which is a set of {@link
     * StreamNamespace} objects.</p>
     *
     * <p>
     * The value of <tt>xpath</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>nss</tt> may be <tt>null</tt>, in which case it is
     * considered to be an empty Set.</p>
     *
     * @param xpath The XPath to setup.
     * @param nss The namespace set.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @throws RuntimeException If any errors occur adding namespaces to the
     * XPath.
     * @since   JSO-0.8.3
     */
    public static final void setupNamespaces(XPath xpath, Set nss) throws IllegalArgumentException, RuntimeException {
        Iterator    itr;

        if (xpath == null)
            throw new IllegalArgumentException("XPath cannot be null");
        if (nss == null)
            nss = java.util.Collections.EMPTY_SET;

        try {
            itr = nss.iterator();
            while (itr.hasNext()) {
                StreamNamespace     ns = (StreamNamespace)itr.next();

                if (ns != null)
                    xpath.addNamespace(ns.getPrefix(), ns.getURI());
            }
        } catch (JaxenException je) {
            throw new RuntimeException(je);
        }
    }
    /**
     * <p>
     * Adds the map of namespaces to the given XPath. The namespaces are
     * determined from <tt>nss</tt>, using the key as the namespace
     * prefix, and the value as the namespace URI.</p>
     *
     * <p>
     * The value of <tt>xpath</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>nss</tt> may be <tt>null</tt>, in which case it is
     * considered to be an empty Map.</p>
     *
     * @param xpath The XPath to setup.
     * @param nss The namespace map.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @throws RuntimeException If any errors occur adding namespaces to the
     * XPath.
     * @since   JSO-0.8.3
     */
    public static final void setupNamespaces(XPath xpath, Map nss) throws IllegalArgumentException, RuntimeException {
        Iterator    itr;

        if (xpath == null)
            throw new IllegalArgumentException("XPath cannot be null");
        if (nss == null)
            nss = java.util.Collections.EMPTY_MAP;

        try {
            itr = nss.keySet().iterator();
            while (itr.hasNext()) {
                Object      key = itr.next();
                Object      value = nss.get(key);

                if ((key != null) && (value != null))
                    xpath.addNamespace(key.toString(), value.toString());
            }
        } catch (JaxenException je) {
            throw new RuntimeException(je);
        }
    }

    /**
     * <p>
     * Matches the two Locales. The order of matching is "variant", "country",
     * and "language".</p>
     * 
     * <p>
     * The two locales match if the first locale is a valid
     * subset of the second. For example, if <tt>cmp1</tt> is "en" and
     * <tt>cmp2</tt> is "en-US" or "en-UK", then the locales match.</p>
     *
     * <p>
     * The value of <tt>cmp1</tt> cannot be <tt>null</tt>, or a
     * <tt>NullPointerException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>cmp2</tt> cannot be <tt>null</tt>, or a
     * <tt>NullPointerException</tt> is thrown.</p>
     *
     * @param cmp1 The "source" Locale to match.
     * @param cmp2 The "target" Locale to match.
     * @return <tt>true</tt> if <tt>cmp1</tt> is a subset of <tt>cmp2</tt>.
     * @throws NullPointerException If any parameters are <tt>null</tt>.
     * @since   JSO-0.12
     */
    public static final boolean matchLocales(Locale cmp1, Locale cmp2) throws NullPointerException {
        String      test1, test2;
        int         result = (cmp1.equals(cmp2) ? 0 : 1);

        if (result != 0) {
            //Compare variants
            test1 = cmp1.getVariant();
            test2 = cmp2.getVariant();
            
            result = (!isValidString(test1) ? 0 : test1.compareTo(test2));
        }
        if (result == 0) {
            //Compare countries
            test1 = cmp1.getCountry();
            test2 = cmp2.getCountry();

            result = (!isValidString(test1) ? 0 : test1.compareTo(test2));
        }
        if (result == 0) {
            //Compare languages
            test1 = cmp1.getLanguage();
            test2 = cmp2.getLanguage();

            result = (!isValidString(test1) ? 0 : test1.compareTo(test2));
        }
        
        return (result == 0);
    }
}