/*
 * XPathMonitor.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.util;


import java.util.Map;
import java.util.Timer;

import org.jaxen.JaxenException;
import org.jaxen.XPath;

import org.saxpath.SAXPathException;

import org.jabberstudio.jso.Packet;
import org.jabberstudio.jso.PacketRouter;
import org.jabberstudio.jso.Stream;
import org.jabberstudio.jso.StreamContext;
import org.jabberstudio.jso.StreamDataFactory;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.StreamException;
import org.jabberstudio.jso.event.MonitorEvent;
import org.jabberstudio.jso.event.MonitorListener;
import org.jabberstudio.jso.event.PacketDispatcher;
import org.jabberstudio.jso.event.XPathMonitorEvent;
import org.jabberstudio.jso.xpath.XPathSupport;

/**
 * <p>
 * Class for monitoring a router/dispatcher for <tt>Packet</tt>s matching a
 * specific XPath. This class implements both synchronous
 * and asynchronous modes. The asynchronous mode utilitzes the class
 * <tt>java.util.Timer</tt> to perform its monitoring.</p>
 *
 * <p>
 * The matching criteria for a <tt>PacketMonitor</tt> is based on a configured
 * XPath object, using the same methodology as {@link XPathListener}.</p>
 *
 * <p>
 * This class is thread-safe. A single instance may be used by multiple
 * threads. Changing any of the properties for a given <tt>PacketMonitor</tt>
 * does not impact any monitoring requests currently pending.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public class XPathMonitor extends Monitor {
    //Inner-classes
    private class Matcher extends MonitorTask {
        //Fields
        private XPath       _XPath;
        
        //Constructors
        public Matcher(XPath expr, MonitorListener l) {
            super(l);
            
            _XPath = expr;
        }
        
        //Methods
        public final XPath getXPath() {
            return _XPath;
        }
        
        protected boolean matched(Packet target) {
            StreamContext       ctx = getRouter().getInboundContext();
            StreamDataFactory   sdf = ctx.getDataFactory();
            StreamElement       base = sdf.createElementNode(sdf.createNSI("stream", ctx.getNamespaceURI("")));
            StreamElement       parent = target.getParent();
            boolean             result = false;

            target.setParent(null);
            base.add(target);
            try {
                result = (getXPath().selectSingleNode(base) != null);
            } catch (Exception e) {
                obtainLogger().info("XPath could not be evaluated", e);
            }
            target.setParent(parent);
            
            return result;
        }
        
        protected void fireMonitorFound(Packet actual) {
            MonitorEvent    evt = XPathMonitorEvent.createFoundEvent(XPathMonitor.this, getXPath(), actual);
            MonitorListener l = getListener();
            
            stop(evt);
            if (l != null)
                l.monitorFound(evt);
        }
        protected void fireMonitorTimeout() {
            MonitorEvent    evt = XPathMonitorEvent.createTimeoutEvent(XPathMonitor.this, getXPath());
            MonitorListener l = getListener();
            
            stop(evt);
            if (l != null)
                l.monitorTimeout(evt);
        }
        protected void fireMonitorFailed(StreamException cause) {
            MonitorEvent    evt = XPathMonitorEvent.createFailedEvent(XPathMonitor.this, getXPath(), cause);
            MonitorListener l = getListener();
            
            stop(evt);
            if (l != null) 
                l.monitorFailed(evt);
        }
    }
    
    //Fields
    private XPath   _XPath;
    
    //Constructors
    /**
     * <p>
     * Creates a new <tt>XPathMonitor</tt> with all fields set to default
     * values.</p>
     *
     * @since   JSO-0.12
     */
    public XPathMonitor() {
    }
    /**
     * <p>
     * Creates a new <tt>XPathMonitor</tt> with a relative XPath expression.
     * <tt>support</tt> and <tt>expr</tt> are used to create the <tt>XPath</tt>
     * object, and <tt>nss</tt> is used to set up namespace mappings. The
     * expression is modified (if necessary) to be relative to the transferred
     * <tt>Packet</tt>'s (internal) parent.</p>
     *
     * <p>
     * The value of <tt>support</tt> cannot be <tt>null</tt>,
     * or an <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>expr</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown. If <tt>expr</tt>,
     * after formatting, does not evaluate to a valid XPath expression, a
     * <tt>SAXPathException</tt> is thrown.</p>
     *
     * @param support The <tt>XPathSupport</tt> to use for XPath creation.
     * @param expr The relative XPath expression to process.
     * @throws IllegalArgumentException If any of the parameters are invalid.
     * @throws SAXPathException If the expression does not evaluate to a valid
     * XPath.
     * @since   JSO-0.12
     */
    public XPathMonitor(XPathSupport support, String expr) throws IllegalArgumentException, SAXPathException {
        this(support, expr, null);
    }
    /**
     * <p>
     * Creates a new <tt>XPathMonitor</tt> with a relative XPath expression.
     * <tt>support</tt> and <tt>expr</tt> are used to create the <tt>XPath</tt>
     * object, and <tt>nss</tt> is used to set up namespace mappings. The
     * expression is modified (if necessary) to be relative to the transferred
     * <tt>Packet</tt>'s (internal) parent.</p>
     *
     * <p>
     * The value of <tt>support</tt> cannot be <tt>null</tt>,
     * or an <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>expr</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown. If <tt>expr</tt>,
     * after formatting, does not evaluate to a valid XPath expression, a
     * <tt>SAXPathException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>nss</tt> may be <tt>null</tt>, in which case it is
     * considered to be an empty Map. For each entry in the given Map, the
     * key is considered the namespace prefix, and the value is the namespace
     * URI.</p>
     *
     * @param support The <tt>XPathSupport</tt> to use for XPath creation.
     * @param expr The relative XPath expression to process.
     * @param nss The namespace mapping to apply.
     * @throws IllegalArgumentException If any of the parameters are invalid.
     * @throws SAXPathException If the expression does not evaluate to a valid
     * XPath.
     * @since   JSO-0.12
     */
    public XPathMonitor(XPathSupport support, String expr, Map nss) throws IllegalArgumentException, SAXPathException {
        setXPath(setupXPath(support, expr, nss));
    }
    
    //Methods
    /**
     * <p>
     * Retrieves the <tt>XPath</tt> to evaluate against.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if it has not
     * been set.</p>
     *
     * @return The <tt>XPath</tt> to evaluate against.
     * @since   JSO-0.12
     */
    public XPath getXPath() {
        return _XPath;
    }
    /**
     * <p>
     * Sets the <tt>XPath</tt> to evaluate against.</p>
     *
     * <p>
     * The value of <tt>xpath</tt> may be <tt>null</tt>, in which case any
     * previous value removed.</p>
     *
     * @param xpath The new <tt>XPath</tt>.
     * @since   JSO-0.12
     */
    public void setXPath(XPath xpath) {
        _XPath = xpath;
    }
    /**
     * <p>
     * Sets the <tt>XPath</tt> using the given relative XPath expression. The
     * expression is modified (if necessary) to be relative to the transferred
     * <tt>Packet</tt>'s parent.</p>
     *
     * <p>
     * The value of <tt>support</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>expr</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown. If <tt>expr</tt>,
     * after formatting, does not evaluate to a valid XPath expression, a
     * <tt>SAXPathException</tt> is thrown.</p>
     *
     * @param support The <tt>XPathSupport</tt> to use for XPath creation.
     * @param expr The relative XPath expression to process.
     * @throws IllegalArgumentException If any of the parameters are invalid.
     * @throws SAXPathException If the expression does not evaluate to a valid
     * XPath.
     * @since   JSO-0.12
     */
    public XPath setupXPath(XPathSupport support, String expr) throws IllegalArgumentException, SAXPathException {
        return setupXPath(support, expr, null);
    }
    /**
     * <p>
     * Sets the <tt>XPath</tt> using the given relative XPath expression. The
     * expression is modified (if necessary) to be relative to the transferred
     * <tt>Packet</tt>'s parent.</p>
     *
     * <p>
     * The value of <tt>support</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>expr</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown. If <tt>expr</tt>,
     * after formatting, does not evaluate to a valid XPath expression, a
     * <tt>SAXPathException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>nss</tt> may be <tt>null</tt>, in which case it is
     * ignored. Each key of <tt>nss</tt> is treated as the namespace prefix,
     * and each associated value is its namespace URI.</p>
     *
     * @param support The <tt>XPathSupport</tt> to use for XPath creation.
     * @param expr The relative XPath expression to process.
     * @param nss The additional namespace mappings to use.
     * @throws IllegalArgumentException If any of the parameters are invalid.
     * @throws SAXPathException If the expression does not evaluate to a valid
     * XPath.
     * @since   JSO-0.12
     */
    public XPath setupXPath(XPathSupport support, String expr, Map nss) throws IllegalArgumentException, SAXPathException  {
        //Validate input
        if (support == null)
            throw new IllegalArgumentException("XPathSupport cannot be null");
        if (!Utilities.isValidString(expr))
            throw new IllegalArgumentException("XPath expression cannot be null or \"\"");

        //Configure expression
        if (!expr.startsWith("./"))
            expr = (expr.startsWith("/") ? "." : "./") + expr;

        setXPath(support.createXPath(expr, nss));
            
        return getXPath();
    }

    /**
     * <p>
     * Creates a <tt>MonitorTask</tt> using the given <tt>Packet</tt> and
     * <tt>MonitorListener</tt>. The object returned is used for both
     * synchronous and asynchronous monitoring.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * <p>
     * The value of <tt>p</tt> may be <tt>null</tt>.</p>
     *
     * <p>
     * The value of <tt>l</tt> may be <tt>null</tt>.</p>
     *
     * @param p The <tt>Packet</tt> to be sent.
     * @param l The <tt>MonitorListener</tt> to register.
     * @throws IllegalStateException If any required properties are invalid.
     * @see #sendAndWatch(Packet)
     * @see #send(Packet, MonitorListener)
     * @since   JSO-0.12
     */
    protected MonitorTask createMonitorTask(Packet p, MonitorListener l) throws IllegalStateException {
        PacketRouter        router = getRouter();
        PacketDispatcher    dispatch = getDispatcher();
        XPath               xpath = getXPath();
        
        //Validate router and dispatcher
        if (router == null)
            throw new IllegalStateException("PacketRouter is not valid");
        if (dispatch == null)
            throw new IllegalStateException("PacketDispatcher is not valid");
        if (xpath == null)
            throw new IllegalArgumentException("XPath cannot be null");
        
        return new Matcher(xpath, l);
    }

    /**
     * <p>
     * Sends the given <tt>Packet</tt> using the given <tt>Stream</tt>,
     * and waits for its matching result on <tt>PacketDispatcher</tt>. This
     * method creates a new <tt>XPathMonitor</tt>, using <tt>s</tt> for the
     * <tt>PacketRouter</tt> and <tt>PacketDispatcher</tt>, and <tt>x</tt> for
     * the XPath to evaluate against.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if the timeout
     * was reached before any matches were found.</p>
     *
     * <p>
     * The value of <tt>s</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>p</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>x</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param s The <tt>Stream</tt> to monitor.
     * @param x The <tt>XPath</tt> to match against.
     * @param p The <tt>Packet</tt> to send.
     * @return The <tt>Packet</tt> matching against <tt>x</tt>.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws StreamException If the <tt>Stream</tt> fails.
     * @since   JSO-0.12
     */
    public static final Packet sendAndWatch(Stream s, Packet p, XPath x) throws IllegalArgumentException, StreamException {
        return sendAndWatch(s, s, p, x, DEFAULT_TIMEOUT, DEFAULT_PROC_INTERVAL);
    }
    /**
     * <p>
     * Sends the given <tt>Packet</tt> using the given <tt>Stream</tt>,
     * and waits for its matching result on <tt>PacketDispatcher</tt>. This
     * method creates a new <tt>XPathMonitor</tt>, using <tt>s</tt> for the
     * <tt>PacketRouter</tt> and <tt>PacketDispatcher</tt>, <tt>x</tt> for the
     * XPath to evaluate against, and <tt>t</tt> for the timeout.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if the timeout
     * was reached before any matches were found.</p>
     *
     * <p>
     * The value of <tt>s</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>p</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>x</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>t</tt> may be less than 0, in which case it is
     * considered to be {@link #DEFAULT_TIMEOUT}.</p>
     *
     * @param s The <tt>Stream</tt> to monitor.
     * @param x The <tt>XPath</tt> to match against.
     * @param p The <tt>Packet</tt> to send.
     * @param t The timeout period, in milliseconds.
     * @return The <tt>Packet</tt> matching against <tt>x</tt>.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws StreamException If the <tt>Stream</tt> fails.
     * @since   JSO-0.12
     */
    public static final Packet sendAndWatch(Stream s, Packet p, XPath x, long t) throws IllegalArgumentException, StreamException {
        return sendAndWatch(s, s, p, x, t, DEFAULT_PROC_INTERVAL);
    }
    /**
     * <p>
     * Sends the given <tt>Packet</tt> using the given <tt>Stream</tt>,
     * and waits for its matching result on <tt>PacketDispatcher</tt>. This
     * method creates a new <tt>XPathMonitor</tt>, using <tt>s</tt> for the
     * <tt>PacketRouter</tt> and <tt>PacketDispatcher</tt>, <tt>x</tt> for the
     * XPath to evaluate against, <tt>t</tt> for the timeout, and <tt>i</tt>
     * for the process interval.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if the timeout
     * was reached before any matches were found.</p>
     *
     * <p>
     * The value of <tt>s</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>p</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>x</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>t</tt> may be less than 0, in which case it is
     * considered to be {@link #DEFAULT_TIMEOUT}.</p>
     *
     * <p>
     * The value of <tt>i</tt> may be less than 1, in which case it is
     * considered to be {@link #DEFAULT_PROC_INTERVAL}.</p>
     *
     * @param s The <tt>Stream</tt> to monitor.
     * @param x The <tt>XPath</tt> to match against.
     * @param p The <tt>Packet</tt> to send.
     * @param t The timeout period, in milliseconds.
     * @param i The process interval, in milliseconds.
     * @return The <tt>Packet</tt> matching against <tt>x</tt>.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws StreamException If the <tt>Stream</tt> fails.
     * @since   JSO-0.12
     */
    public static final Packet sendAndWatch(Stream s, Packet p, XPath x, long t, long i) throws IllegalArgumentException, StreamException {
        return sendAndWatch(s, s, p, x, t, i);
    }
    
    /**
     * <p>
     * Sends the given <tt>Packet</tt> using the given <tt>PacketRouter</tt>,
     * and waits for its matching result on <tt>PacketDispatcher</tt>. This
     * method creates a new <tt>XPathMonitor</tt>, using <tt>r</tt> for the
     * <tt>PacketRouter</tt>, <tt>d</tt> for the <tt>PacketDispatcher</tt>, and
     * <tt>x</tt> for the XPath to evaluate against.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if the timeout
     * was reached before any matches were found.</p>
     *
     * <p>
     * The value of <tt>r</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>d</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>p</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>x</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param r The <tt>PacketRouter</tt> to monitor.
     * @param d The <tt>PacketDispatcher</tt> to monitor.
     * @param p The <tt>Packet</tt> to send.
     * @param x The <tt>XPath</tt> to match against.
     * @return The <tt>Packet</tt> matching against <tt>x</tt>.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws StreamException If the <tt>PacketRouter</tt> fails.
     * @since   JSO-0.12
     */
    public static final Packet sendAndWatch(PacketRouter r, PacketDispatcher d, Packet p, XPath x) throws IllegalArgumentException, StreamException {
        return sendAndWatch(r, d, p, x, DEFAULT_TIMEOUT, DEFAULT_PROC_INTERVAL);
    }
    /**
     * <p>
     * Sends the given <tt>Packet</tt> using the given <tt>PacketRouter</tt>,
     * and waits for its matching result on <tt>PacketDispatcher</tt>. This
     * method creates a new <tt>XPathMonitor</tt>, using <tt>r</tt> for the
     * <tt>PacketRouter</tt>, <tt>d</tt> for the <tt>PacketDispatcher</tt>,
     * <tt>x</tt> for the XPath to evaluate against, and <tt>t</tt> for the
     * timeout.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if the timeout
     * was reached before any matches were found.</p>
     *
     * <p>
     * The value of <tt>r</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>d</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>p</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>x</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>t</tt> may be less than 0, in which case it is
     * considered to be {@link #DEFAULT_TIMEOUT}.</p>
     *
     * @param r The <tt>PacketRouter</tt> to monitor.
     * @param d The <tt>PacketDispatcher</tt> to monitor.
     * @param p The <tt>Packet</tt> to send.
     * @param x The <tt>XPath</tt> to match against.
     * @param t The timeout period, in milliseconds.
     * @return The <tt>Packet</tt> matching against <tt>x</tt>.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws StreamException If the <tt>PacketRouter</tt> fails.
     * @since   JSO-0.12
     */
    public static final Packet sendAndWatch(PacketRouter r, PacketDispatcher d, Packet p, XPath x, long t) throws IllegalArgumentException, StreamException {
        return sendAndWatch(r, d, p, x, t, DEFAULT_PROC_INTERVAL);
    }
    /**
     * <p>
     * Sends the given <tt>Packet</tt> using the given <tt>PacketRouter</tt>,
     * and waits for its matching result on <tt>PacketDispatcher</tt>. This
     * method creates a new <tt>XPathMonitor</tt>, using <tt>r</tt> for the
     * <tt>PacketRouter</tt>, <tt>d</tt> for the <tt>PacketDispatcher</tt>,
     * <tt>x</tt> for the XPath to evaluate against, <tt>t</tt> for the timeout,
     * and <tt>i</tt> for the process interval.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if the timeout
     * was reached before any matches were found.</p>
     *
     * <p>
     * The value of <tt>r</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>d</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>p</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>x</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>t</tt> may be less than 0, in which case it is
     * considered to be {@link #DEFAULT_TIMEOUT}.</p>
     *
     * <p>
     * The value of <tt>i</tt> may be less than 1, in which case it is
     * considered to be {@link #DEFAULT_PROC_INTERVAL}.</p>
     *
     * @param r The <tt>PacketRouter</tt> to monitor.
     * @param d The <tt>PacketDispatcher</tt> to monitor.
     * @param p The <tt>Packet</tt> to send.
     * @param x The <tt>XPath</tt> to match against.
     * @param t The timeout period, in milliseconds.
     * @param i The process interval, in milliseconds.
     * @return The <tt>Packet</tt> matching against <tt>x</tt>.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @throws StreamException If the <tt>PacketRouter</tt> fails.
     * @since   JSO-0.12
     */
    public static final Packet sendAndWatch(PacketRouter r, PacketDispatcher d, Packet p, XPath x, long t, long i) throws IllegalArgumentException, StreamException {
        XPathMonitor    mon;

        if (r == null)
            throw new IllegalArgumentException("PacketRouter cannot be null");
        if (d == null)
            throw new IllegalArgumentException("PacketDispatcher cannot be null");
        if (x == null)
            throw new IllegalArgumentException("XPath cannot be null");
        if (p == null)
            throw new IllegalArgumentException("Packet cannot be null");
        
        mon = new XPathMonitor();
        mon.setRouter(r);
        mon.setDispatcher(d);
        mon.setXPath(x);
        mon.setTimeout(t);
        mon.setProcessInterval(i);
        
        return mon.sendAndWatch(p);
    }
}