/*
 * CommandNote.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.x.commands;


import org.jabberstudio.jso.NSI;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.util.Enumerator;

/**
 * <p>
 * Defines the interface for specifying notes about a stage in command
 * execution. Each <tt>CommandNote</tt> has a type and a text message.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public interface CommandNote extends StreamElement {
    //Inner-classes
    /**
     * <p>
     * Class to enumerate the <tt>CommandNote</tt> type.</p>
     *
     * @author  linuxwolf
     * @version $VERSION$
     * @since   JSO-0.12
     */
    public static final class Type extends Enumerator {
        //Constructors
        private Type(String name) {
            super(name);
        }
    }

    //"Constants"
    /** The qualified name for <tt>CommandNote</tt>
     * ({http://jabber.org/protocol/commands}note).
     *
     * @since   JSO-0.12
     */
    public static final NSI     NAME    = new NSI("note", CommandQuery.NAMESPACE);

    /** Type to mark a <tt>CommandNote</tt> as an informational message.
     *
     * @since   JSO-0.12
     */
    public static final Type    INFO    = new Type("info");
    /** Type to mark a <tt>CommandNote</tt> as a warning message.
     *
     * @since   JSO-0.12
     */
    public static final Type    WARN    = new Type("warn");
    /** Type to mark a <tt>CommandNote</tt> as an error message.
     *
     * @since   JSO-0.12
     */
    public static final Type    ERROR   = new Type("error");

    //Methods
    /**
     * <p>
     * Retrieves the type for this <tt>CommandNote</tt>.</p>
     *
     * <p>
     * The value returned by this method should not be <tt>null</tt>.</p>
     *
     * @return The type for this <tt>CommandNote</tt>.
     * @since   JSO-0.12
     */
    public Type getType();
    /**
     * <p>
     * Retrieves the textual message for this <tt>CommandNote</tt>. This
     * value should be appropriate for the requested or current locale.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>, and should
     * never be "".</p>
     *
     * @return the textual message for this <tt>CommandNote</tt>.
     * @since   JSO-0.12
     */
    public String getMessage();
}