/*
 * CommandQuery.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.x.commands;


import java.util.List;

import org.jabberstudio.jso.Extension;
import org.jabberstudio.jso.NSI;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.util.Enumerator;

/**
 * <p>
 * Defines the Command query interface. Through this interface, all
 * other information about a command stage request/response is obtained.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public interface CommandQuery extends Extension {
    //Inner-classes
    /**
     * <p>
     * Class to enumerate the <tt>CommandQuery</tt> action.</p>
     *
     * @author  linuxwolf
     * @version $VERSION$
     * @since   JSO-0.12
     */
    public static final class Action extends Enumerator {
        //Constructors
        private Action(String name) {
            super(name);
        }
    }
    /**
     * <p>
     * Class to enumerate the <tt>CommandQuery</tt> status.</p>
     *
     * @author  linuxwolf
     * @version $VERSION$
     * @since   JSO-0.12
     */
    public static final class Status extends Enumerator {
        //Constructors
        private Status(String name) {
            super(name);
        }
    }
    
    //"Constants"
    /** The Namespace URI for <tt>CommandQuery</tt>
     * (http://jabber.org/protocol/commands).
     *
     * @since   JSO-0.12
     */
    public static final String  NAMESPACE   = "http://jabber.org/protocol/commands";
    /** The qualified name for <tt>CommandQuery</tt>
     * ({http://jabber.org/protocol/commands}command).
     *
     * @since   JSO-0.12
     */
    public static final NSI     NAME        = new NSI("command", NAMESPACE);

    /** Action to execute this <tt>CommandQuery</tt>. This is considered the
     * default action.
     *
     * @since   JSO-0.12
     */
    public static final Action  EXECUTE     = new Action("execute");
    /** Action to cancel this <tt>CommandQuery</tt>.
     *
     * @since   JSO-0.12
     */
    public static final Action  CANCEL      = new Action("cancel");
    /** Action to proceed to the next stage for this <tt>CommandQuery</tt>.
     *
     * @since   JSO-0.12
     */
    public static final Action  NEXT        = new Action("next");
    /** Action to preceed to the previous stage for this <tt>CommandQuery</tt>.
     *
     * @since   JSO-0.12
     */
    public static final Action  PREV        = new Action("prev");
    /** Action to complete this <tt>CommandQuery</tt>.
     *
     * @since   JSO-0.12
     */
    public static final Action  COMPLETE    = new Action("complete");
    
    /** Status to mark this <tt>CommandQuery</tt> as executing.
     *
     * @since   JSO-0.12
     */
    public static final Status  EXECUTING   = new Status("executing");
    /** Status to mark this <tt>CommandQuery</tt> as completed.
     *
     * @since   JSO-0.12
     */
    public static final Status  COMPLETED   = new Status("completed");
    /** Status to mark this <tt>CommandQuery</tt> as canceled.
     *
     * @since   JSO-0.12
     */
    public static final Status  CANCELED    = new Status("canceled");
    
    //Methods
    /**
     * <p>
     * Retrieves the node for this <tt>CommandQuery</tt>. The node uniquely
     * identifies a command for a particular JID, and corresponds to the disco
     * node used for gathering information and associations on the command.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * @return The node for this <tt>CommandQuery</tt>.
     * @since   JSO-0.12
     */
    public String getNode();
    /**
     * <p>
     * Sets the node for this <tt>CommandQuery</tt>.</p>
     *
     * <p>
     * The value of <tt>node</tt> may be <tt>null</tt> or "", in which case
     * any previous value is removed.</p>
     *
     * @param node The new node.
     * @since   JSO-0.12
     */
    public void setNode(String node);

    /**
     * <p>
     * Retrieves the action for this <tt>CommandQuery</tt>. This provides the
     * command responder with hints on how to process this stage of a command's
     * execution.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt>.</p>
     *
     * @return The action for this <tt>CommandQuery</tt>.
     * @since   JSO-0.12
     */
    public Action getAction();
    /**
     * <p>
     * Sets the action for this <tt>CommandQuery</tt>.</p>
     *
     * <p>
     * The value of <tt>act</tt> may be <tt>null</tt>, in which case
     * any previous value is removed.</p>
     *
     * @param act The new action.
     * @since   JSO-0.12
     */
    public void setAction(Action act);
    
    /**
     * <p>
     * Retrieves the status for this <tt>CommandQuery</tt>. This provides the
     * command requester with the status from the command's last stage of
     * execution.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt>.</p>
     *
     * @return The status for this <tt>CommandQuery</tt>.
     * @since   JSO-0.12
     */
    public Status getStatus();
    /**
     * <p>
     * Sets the status for this <tt>CommandQuery</tt>.</p>
     *
     * <p>
     * The value of <tt>stat</tt> may be <tt>null</tt>, in which case
     * any previous value is removed.</p>
     *
     * @param stat The new status.
     * @since   JSO-0.12
     */
    public void setStatus(Status stat);
    
    /**
     * <p>
     * Retrieves the session identifier for this <tt>CommandQuery</tt>. This
     * uniquely identifies the command execution session between a specific
     * requester and responder.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt>.</p>
     *
     * @return The session identifier for this <tt>CommandQuery</tt>.
     * @since   JSO-0.12
     */
    public String getSessionID();
    /**
     * <p>
     * Sets the session identifier for this <tt>CommandQuery</tt>.</p>
     *
     * <p>
     * The value of <tt>id</tt> may be <tt>null</tt> or "", in which case
     * any previous value is removed.</p>
     *
     * @param id The new session identifier.
     * @since   JSO-0.12
     */
    public void setSessionID(String id);
    
    /**
     * <p>
     * Retrieves information on the allowed actions for this
     * <tt>CommandQuery</tt>. This helps the requester determine what actions
     * can be specified, and how {@link #EXECUTE} is interpreted.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. If it is not
     * yet present, a <tt>CommandActions</tt> is first created.</p>
     *
     * @return The action information this <tt>CommandQuery</tt>.
     * @since   JSO-0.12
     */
    public CommandActions getActionInfo();
    
    /**
     * <p>
     * Retrieves the list of <tt>CommandNote</tt> objects for this
     * <tt>CommandQuery</tt>.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each element
     * in the returned list is a <tt>CommandNote</tt> instance.</p>
     */
    public List listNotes();
    /**
     * <p>
     * Adds a <tt>CommandNote</tt> with the given type and message.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>.</p>
     *
     * <p>
     * The value of <tt>type</tt> cannot be <tt>null</tt>, or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>msg</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param type The type of <tt>CommandNote</tt>.
     * @param msg The message for the <tt>CommandNote</tt>.
     * @return The newly created <tt>CommandNote</tt>.
     * @throws IllegalArgumentException If any parameters are invalid.
     * @since   JSO-0.12
     */
    public CommandNote addNote(CommandNote.Type type, String msg) throws IllegalArgumentException;
    /**
     * <p>
     * Removes all <tt>CommandNote</tt> objects in this
     * <tt>CommandQuery</tt>.</p>
     *
     * @since   JSO-0.12
     */
    public void clearNotes();
    
    /**
     * <p>
     * Retrieves the list of <tt>CommandNote</tt> objects of the given type for
     * this <tt>CommandQuery</tt>.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each element
     * in the returned list is a <tt>CommandNote</tt> instance.</p>
     *
     * <p>
     * The value of <tt>type</tt> may be <tt>null</tt>, in which case this
     * method behaves exactly like {@link #listNotes()}.</p>
     *
     * @param type The <tt>CommandNote</tt> type to look for.
     * @return The list of <tt>CommandNote</tt> objects.
     * @since   JSO-0.12
     */
    public List listNotes(CommandNote.Type type);
    /**
     * <p>
     * Removes all <tt>CommandNote</tt> objects of the given type in this
     * <tt>CommandQuery</tt>.</p>
     *
     * <p>
     * The value of <tt>type</tt> may be <tt>null</tt>, in which case this
     * method behaves exactly like {@link #clearNotes()}.</p>
     *
     * @param type The <tt>CommandNote</tt> type to remove.
     * @since   JSO-0.12
     */
    public void clearNotes(CommandNote.Type type);

    /**
     * <p>
     * Retrieves the payload for this <tt>CommandQuery</tt>. This value is the
     * first <tt>StreamElement</tt> within the <tt>CommandQuery</tt> that is
     * not in the the namespace {@link #NAMESPACE}.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt>.</p>
     *
     * @return The payload.
     * @since   JSO-0.12
     */
    public StreamElement getPayload();
    /**
     * <p>
     * Sets the payload for this <tt>CommandQuery</tt>.</p>
     *
     * <p>
     * The value of <tt>elem</tt> may be <tt>null</tt>, in which case any
     * previous payload is removed.</p>
     *
     * @param payload The new payload.
     */
    public void setPayload(StreamElement payload);

    /**
     * <p>
     * Lists all of the payload for this <tt>CommandQuery</tt>. This value is
     * the list of all <tt>StreamElement</tt> objects that is not in the
     * namespace {@link #NAMESPACE}.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each element
     * in the returned list is a <tt>StreamElement</tt> instance.</p>
     *
     * @return The list of payload elements.
     * @since   JSO-0.12
     */
    public List listPayload();
    /**
     * <p>
     * Clears all of the payload from thsi <tt>CommandQuery</tt>.</p>
     *
     * @since   JSO-0.12
     */
    public void clearPayload();
}