/*
 * RosterItem.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.x.core;


import java.util.List;

import org.jabberstudio.jso.JID;
import org.jabberstudio.jso.StreamElement;
import org.jabberstudio.jso.util.Enumerator;

/**
 * <p>
 * Represents &lt;item/&gt; elements within a "jabber:iq:roster" query or
 * "jabber:x:roster" extension.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.4
 */
public interface RosterItem extends StreamElement {
    //Inner-classes
    /**
     * <p>
     * Enumerating class for subscription type information.</p>
     *
     * @author  linuxwolf
     * @version $VERSION$
     * @since   JSO-0.4
     */
    public static class SubscriptionType extends Enumerator {
        //Fields
        private boolean     _To;
        private boolean     _From;
        
        //Constructors
        protected SubscriptionType(String name, boolean to, boolean from) {
            super(name);
            _To = to;
            _From = from;
        }
        
        //Methods
        /**
         * <p>
         * Determines if this <code>SubscriptionType</code> indicates presence
         * is sent.</p>
         *
         * @return <code>true</code> if presence is sent.
         * @since   JSO-0.4
         */
        public boolean isSubscribedTo() {
            return _To;
        }
        /**
         * <p>
         * Determines if this <code>SubscriptionType</code> indicates presence
         * is received.</p>
         *
         * @return <code>true</code> if presence is received.
         * @since   JSO-0.4
         */
        public boolean isSubscribedFrom() {
            return _From;
        }
    }
    
    //"Constants"
    /** <code>SubscriptionType</code> to flag "none" subscriptions.
     * @since   JSO-0.4
     */
    public static final SubscriptionType    NONE = new SubscriptionType("none", false, false);
    /** <code>SubscriptionType</code> to flag "to" subscriptions.
     * @since   JSO-0.4
     */
    public static final SubscriptionType    TO = new SubscriptionType("to", true, false);
    /** <code>SubscriptionType</code> to flag "from" subscriptions.
     * @since   JSO-0.4
     */
    public static final SubscriptionType    FROM = new SubscriptionType("from", false, true);
    /** <code>SubscriptionType</code> to flag "both" subscriptions.
     * @since   JSO-0.4
     */
    public static final SubscriptionType    BOTH = new SubscriptionType("both", true, true);
    /** <code>SubscriptionType</code> to flag "remove" subscriptions.
     * @since   JSO-0.4
     */
    public static final SubscriptionType    REMOVE = new SubscriptionType("remove", false, true);
    
    //Methods
    /**
     * <p>
     * Retrieves the JID for this <code>RosterItem</code>.  This method maps
     * to the attribute "jid".</p>
     *
     * <p>
     * This method does not have a complimenting "set" because this attribute
     * cannot be changed.</p>
     *
     * <p>
     * The value returned by this method should not be <code>null</code>.</p>
     *
     * @return The JID for this <code>RosterItem</code>.
     * @since   JSO-0.4
     */
    public JID getJID();
    
    /**
     * <p>
     * Retrieves the "display name" for this <code>RosterItem</code>.  This
     * method maps to the attribute "name".</p>
     *
     * <p>
     * The value returned by this method may be <code>null</code> if there is
     * no display name for this <code>RosterItem</code>.
     *
     * @return The display name for this <code>RosterItem</code>.
     * @since   JSO-0.4
     */
    public String getDisplayName();
    /**
     * <p>
     * Sets the "display name" for this <code>RosterItem</code>.</p>
     *
     * <p>
     * The value of <code>name</code> can be <code>null</code> or "", in which
     * case the attribute is "unset".</p>
     *
     * @param name The new display name for this <code>RosterItem</code>.
     * @since   JSO-0.4
     */
    public void setDisplayName(String name);
    
    /**
     * <p>
     * Retrieves the subscription type for this <Code>RosterItem</code>.  This
     * value indicates if presence information is sent and/or received from
     * this JID.</p>
     *
     * <p>
     * The value returned by this method is never <code>null</code>.</p>
     *
     * @return The type of subscription for this <code>RosterItem</code>.
     * @since   JSO-0.4
     */
    public SubscriptionType getSubscription();
    /**
     * <p>
     * Sets the subscription type for this <Code>RosterItem</code>.<p>
     *
     * <p>
     * The value of <code>type</code> may be <code>null</code>, in which case
     * it is considered to be {@link #NONE}.</p>
     *
     * @param type The new subscription type.
     * @since   JSO-0.4
     */
    public void setSubscription(SubscriptionType type);
    
    /**
     * <p>
     * Determines if this <code>RosterItem</code> is pending a subscription
     * reguest.  This methos maps to the attribute "ask", and the value
     * "subscribe".</p>
     *
     * <p>
     * Note that this method always returns <code>false</code> if {@link
     * #isAskUnsubscribe} returns <code>true</code>.</p>
     *
     * @return <code>true</code> if the value of attribute "ask" is equal to
     * "subscribe".
     * @since   JSO-0.4
     */
    public boolean isAskSubscribe();
    /**
     * <p>
     * Sets or unsets the "pending subscription request" flag for this
     * <Code>RosterItem</code>.</p>
     *
     * <p>
     * If the value of <code>sub</code> is <code>true</code>, the value of the
     * attribute "ask" is set to "subscribe".  Otherwise, the value completely
     * removed.</p>
     *
     * @param sub The new state of the flag.
     * @since   JSO-0.4
     */
    public void setAskSubscribe(boolean sub);

    /**
     * <p>
     * Determines if this <code>RosterItem</code> is pending an unsubscription
     * reguest.  This methos maps to the attribute "ask", and the value
     * "unsubscribe".</p>
     *
     * <p>
     * Note that this method always returns <code>false</code> if {@link
     * #isAskSubscribe} returns <code>true</code>.</p>
     *
     * @return <code>true</code> if the value of attribute "ask" is equal to
     * "unsubscribe".
     * @since   JSO-0.4
     */
    public boolean isAskUnsubscribe();
    /**
     * <p>
     * Sets or unsets the "pending unsubscription request" flag for this
     * <Code>RosterItem</code>.</p>
     *
     * <p>
     * If the value of <code>unsub</code> is <code>true</code>, the value of the
     * attribute "ask" is set to "unsubscribe".  Otherwise, the value completely
     * removed.</p>
     *
     * @param unsub The new state of the flag.
     * @since   JSO-0.4
     */
    public void setAskUnsubscribe(boolean unsub);
    
    /**
     * <p>
     * Retrieves the array of groups for this <code>RosterItem</code>.  The
     * array of groups is determined from the list of contained &lt;group/&gt;
     * elements.  It is determined upon each call to this method, and is not
     * cached.</p>
     *
     * <p>
     * The value returned by this method is never <code>null</code>.</p>
     *
     * @return The array of groups for this <Code>RosterItem</code>.
     * @since   JSO-0.4
     * @deprecated As of JSO-0.12, this method is replaced with {@link
     * #listGroups}.
     */
    public String[] getGroups();
    /**
     * <p>
     * Sets the array of groups for this <Code>RosterItem</code>.  This method
     * removes the current list of contained &lt;group/&gt; elements, and
     * replaces them with values appropriate to the given array.</p>
     *
     * <p>
     * The value of <code>groups</code> can be <code>null</code>, in which
     * case this <code>RosterItem</code> will not be associated to any
     * groups.</p>
     *
     * @param groups The new array of groups.
     * @since   JSO-0.4
     * @deprecated As of JSO-0.12, this method is replaced with {@link
     * #addGroup}, {@link #removeGroup}, and {@link #clearGroups}.
     */
    public void setGroups(String[] groups);
    /**
     * <p>
     * Retrieves the list of groups for this <tt>RosterItem</tt>. The
     * array of groups is determined from the list of contained &lt;group/&gt;
     * elements.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each element
     * of the returned list is a string. The returned list is a snapshot of the
     * groups at the time of this method call, and does not reflect changes to
     * this <tt>RosterItem</tt>.</p>
     *
     * @return The list of groups for this <tt>RosterItem</tt>.
     * @since   JSO-0.12
     */
    public List listGroups();
    /**
     * <p>
     * Adds the given groups to the list of groups for this
     * <tt>RosterItem</tt>.</p>
     *
     * <p>
     * The value of <tt>grp</tt> may be <tt>null</tt> or "", in which ase no
     * actions are performed.</p>
     *
     * @param grp The group to add.
     * @since   JSO-0.12
     */
    public void addGroup(String grp);
    /**
     * <p>
     * Removes the given groups from the list of groups for this
     * <tt>RosterItem</tt>.</p>
     *
     * <p>
     * The value of <tt>grp</tt> may be <tt>null</tt> or "", in which ase no
     * actions are performed.</p>
     *
     * @param grp The group to remove.
     * @since   JSO-0.12
     */
    public void removeGroup(String grp);
    /**
     * <p>
     * Removes all groups from this <tt>RosterItem</tt>.</p>
     *
     * @since   JSO-0.12
     */
    public void clearGroups();
}