/*
 * XDataFieldContainer.java
 *
 * The contents of this file are subject to the GNU Lesser General Public License
 * Version 2.1 (the "License"). You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License. You
 * may obtain a copy of the License at http://www.fsf.org/licenses/lgpl.txt or
 * http://www.opensource.org/. Software distributed under the License is
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
 * or implied. See the License for the specific language governing rights and
 * limitations under the License.
 */

package org.jabberstudio.jso.x.xdata;


import java.util.List;

import org.jabberstudio.jso.StreamElement;

/**
 * <p>
 * Interface for containing <tt>XDataField</tt> objects.</p>
 *
 * @author  linuxwolf
 * @version $VERSION$
 * @since   JSO-0.12
 */
public interface XDataFieldContainer extends StreamElement {
    //Methods
    /**
     * <p>
     * Retrieves the list of contained fields for this
     * <tt>XDataFieldContainer</tt>.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each element
     * in the returned list is a non-<tt>null</tt> instance of
     * <tt>XDataField</tt>.</p>
     *
     * @return The list of contained <tt>XDataField</tt> objects.
     * @since   JSO-0.12
     */
    public List listFields();
    /**
     * <p>
     * Retrieves the field with the given "var" attribute.</p>
     *
     * <p>
     * This method returns <tt>null</tt> if there is no field with the given
     * "var" attribute.</p>
     *
     * <p>
     * The value of <tt>var</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param var The "var" to search for.
     * @return The <tt>XDataField</tt> with attribute "var".
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.12
     */
    public XDataField getField(String var) throws IllegalArgumentException;
    /**
     * <p>
     * Adds a new field with the given "var" and no type.</p>
     *
     * <p>
     * This method never returns <tt>null</tt>. If a <tt>XDataField</tt>
     * already exists with the given "var" and type, it is returned.</p>
     *
     * <p>
     * The value of <tt>var</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * If a <tt>XDataField</tt> already exists for the given "var", but is of a
     * different type, this method throws an <tt>IllegalStateException</tt>.</p>
     *
     * @param var The "var" for the new <tt>XDataField</tt>.
     * @return The new (or existing) <tt>XDataField</tt>.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @throws IllegalStateException If the new type does not match the existing
     * type.
     * @since   JSO-0.12
     */
    public XDataField addField(String var) throws IllegalArgumentException, IllegalStateException;
    /**
     * <p>
     * Adds a new field with the given "var" and type.</p>
     *
     * <p>
     * This method never returns <tt>null</tt>. If a <tt>XDataField</tt>
     * already exists with the given "var" and type, it is returned.</p>
     *
     * <p>
     * The value of <tt>var</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * <p>
     * The value of <tt>type</tt> may be <tt>null</tt>, in which case it is
     * not set.</p>
     *
     * <p>
     * If a <tt>XDataField</tt> already exists for the given "var", but is of a
     * different type, this method throws an <tt>IllegalStateException</tt>.</p>
     *
     * @param var The "var" for the new <tt>XDataField</tt>.
     * @param type The type for the new <tt>XDataField</tt>.
     * @return The new (or existing) <tt>XDataField</tt>.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @throws IllegalStateException If the new type does not match the existing
     * type.
     * @since   JSO-0.12
     */
    public XDataField addField(String var, XDataField.Type type) throws IllegalArgumentException, IllegalStateException;
    /**
     * <p>
     * Removes the field with the given "var" attribute.</p>
     *
     * <p>
     * The value of <tt>var</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param var The "var" of the <tt>XDataField</tt> to remove.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.12
     */
    public void removeField(String var) throws IllegalArgumentException;
    /**
     * <p>
     * Removes all fields from this <tt>XDataFieldContainer</tt>.</p>
     *
     * @since   JSO-0.12
     */
    public void clearFields();
    
    /**
     * <p>
     * Retrieves the list of field values for the given "var" attribute.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if there is
     * no such field, or if the field does not have any values. If the field is
     * a multi-value field, the first such value is retrieved.</p>
     *
     * <p>
     * The value returned by this method is never <tt>null</tt>. Each element
     * in the returned list is a non-<tt>null</tt> string.</p>
     *
     * <p>
     * The value of <tt>var</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param var The "var" of the <tt>XDataField</tt> value to retrieve.
     * @return The list of values of the field.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.12
     */
    public List listFieldValues(String var) throws IllegalArgumentException;
    /**
     * <p>
     * Retrieves the field value for the given "var" attribute.</p>
     *
     * <p>
     * The value returned by this method may be <tt>null</tt> if there is
     * no such field, or if the field does not have any values. If the field is
     * a multi-value field, the first such value is retrieved.</p>
     *
     * <p>
     * The value of <tt>var</tt> cannot be <tt>null</tt> or "", or an
     * <tt>IllegalArgumentException</tt> is thrown.</p>
     *
     * @param var The "var" of the <tt>XDataField</tt> value to retrieve.
     * @return The value of the field.
     * @throws IllegalArgumentException If the parameter is invalid.
     * @since   JSO-0.12
     */
    public String getFieldValue(String var) throws IllegalArgumentException;
}